/** @preserve Eloquence WEBDLG v2
 (C) Copyright Marxmeier Software AG, 2020-2021
 Version: 1.970 2021-11-19
 $Id: 00-base.js,v 28.60 2021/11/19 15:54:03 mike Exp $
*//*======================================================================
   WEBDLG v2 namespace / globals
========================================================================*/

var eq = eq || {};
eq.c = eq.c || {};   // Classes
eq.p = eq.p || {};   // Properties:
eq.p.version = 1970; //  Version 1.970 (major*1,000 + minor)

/*========================================================================
   class Obj
   Generic base class.
========================================================================*/

eq.c.Obj = function() {};

/*------------------------------------------------------------------------
   static Obj.subClass()
   Subclassing convenience helper.
------------------------------------------------------------------------*/

(eq.c.Obj.setClassName = function(name, cls) {
   var pd = Object.getOwnPropertyDescriptor(cls.prototype.constructor, 'name');
   if (pd !== undefined)
      pd.value = name;
   else
      pd = {
         value : name,
         configurable : true,
         enumerable : false,
         writable : false
      };
   Object.defineProperty(cls.prototype.constructor, 'name', pd);
})("Obj", eq.c.Obj);

eq.c.Obj.subClass = function(name, cls) {
   var base = this;
   cls.prototype = Object.create(base.prototype);
   Object.defineProperty(cls.prototype, 'constructor', {
      value : cls,
      configurable : false,
      enumerable : false,
      writable : true
   });
   eq.c.Obj.setClassName(name, cls);
   cls.subClass = function(n, c) { return base.subClass.call(this, n, c) };
   return eq.c[name] = cls;
};

/** @preserve $Id: 05-const.js,v 28.48 2021/11/19 15:46:20 mike Exp $
*//*======================================================================
   static MainOp
   API->Main message opcodes.
========================================================================*/

eq.MainOp = {
   ssnNewOrResume     : 1,  // Data: session id
   applicationStopped : 2,  // Data: restartable true/false
   applicationFailed  : 3,  // Data: none
   layout             : 4,  // Data: layout mode
   typeAhead          : 5,  // Data: enabled
   cancelCurrentCall  : 5,  // Data: none
   updated            : 6,  // Data: dialog id, element id, update tag, arg
   openUrl            : 7,  // Data: URL, window name
   startApp           : 8,  // Data: application name
   activate           : 9,  // Data: dialog title
   notify             : 10, // Data: dialog id, rule
   resume             : 11, // Data: none
   addControl         : 12, // Data: dialog id, DOM element template
   modifyControl      : 13, // Data: dialog id, DOM element template
   deleteControl      : 14, // Data: dialog id, element id
   invalidate         : 15, // Data: dialog id array
   dlg_DRAW           : 16, // Data: dialog id
   dlg_DO             : 17, // Data: dialog id, focus element id
   dlg_POPUP_BOX      : 18  // Data: POPUP BOX descriptor
};

/*========================================================================
   static ApiOp
   Main->API message opcodes.
========================================================================*/

eq.ApiOp = {
   documentComplete : 1, // Data: layout grid
   openWebSocket    : 2, // Data: WebSocket URL, session id
   onEvent          : 3, // Data: element id, event type
   onPopupEvent     : 4, // Data: button number
   onSynced         : 5, // Data: none
   onDo             : 6, // Data: dialog id
   onHelp           : 7, // Data: element id
   invalidate       : 8, // Data: dialog id
   update           : 9  // Data: dialog id, element id, tag, argument
};

/*========================================================================
   static UpdateTag
   Main->API update request/response tags.
========================================================================*/

eq.UpdateTag = {
   mouseButton2 : 1, // Secondary mouse button pressed
   contextMenu  : 2, // Open context menu
   tabOrder     : 3, // Update tab order (TabBox tab switched)
   colResized   : 4, // ListBox column resized
   colMoved     : 5, // ListBox column moved
   sortSequence : 6, // Listbox sort sequence modified
   lineOrder    : 7  // Update ListBox line order (server request)
};

/*========================================================================
   static DlgRq.tag
   DLG request tags.
========================================================================*/

eq.RqTag  = {
   child : 1, // Child control

   // Global properties

   baseUrl     : 2, // Base url
   helpBaseUrl : 3, // Help base url
   layout      : 4, // Dialog layout mode
   typeAhead   : 5, // Type-ahead enabled/disabled
   openUrl     : 6, // Open url
   openHelpUrl : 7, // Open help url
   startApp    : 8, // Start application
   activate    : 9, // Activate window

   // Common attributes.

   cssid       : 10, // CSS id
   cssclsDel   : 11, // Delete CSS class(es)
   cssclsAdd   : 12, // Add CSS class(es)
   raster      : 13, // Raster
   posSize     : 14, // Position/size
   size        : 15, // Size, ToolBar children
   border      : 16, // Border
   rule        : 17, // Rule
   ruleMode    : 18, // When to submit rule (Tree/ListBox.singleclick)
   tabonenter  : 19, // Enter key navigates focus
   visible     : 20, // Visibility
   sensitive   : 21, // Sensitivity
   inactive    : 22, // Inactive, eg. TabBox tab
   tabStop     : 23, // Tab stop
   bgColor     : 24, // Background color
   fgColor     : 25, // Foreground color
   focusColor  : 26, // Focus color
   contextMenu : 27, // Context menu
   url         : 28, // Url
   help        : 29, // Help url

   // Dialog attributes.

   dialogState   : 30, // Resizepolicy, maximized
   dialogPosSize : 31, // Dialog position/size
   fKey          : 32, // Function keys
   dialogTimer   : 33, // Dialog timer rule, delay
   ruleButton2   : 34, // Secondary mouse button rule
   dialogDo      : 35, // Dialog.do

   // GroupBox attributes.

   margin  : 40, // Dialog/GroupBox margin
   vx      : 41, // X scroll position
   vy      : 42, // Y scroll position
   vWidth  : 43, // Visible width
   vHeight : 44, // Visible height

   // Button, Menu, StaticText attributes.

   text       : 50, // Text, also Dialog title
   textPos    : 51, // Text position
   align      : 52, // Align
   active     : 53, // Active
   radioGroup : 54, // RadioButton group
   icon       : 55, // PushButton icon
   iconActive : 56, // PushButton icon when active

   // EditText attributes.

   multiline  : 60, // Multiline
   readonly   : 61, // Read-only
   content    : 62, // Content
   ispassword : 63, // IsPassword

   // ListBox, PopText, Tree attributes.

   vcRule      : 70, // Rule: view has changed
   columns     : 71, // Columns
   colOrder    : 72, // Column order
   colState    : 73, // Column states
   colTitle    : 74, // Column titles
   lineOrder   : 75, // Line order
   activeLine  : 76, // Active line, model order
   vActiveLine : 77, // Active line, view order
   topItem     : 78, // First visible line
   rowHeight   : 79, // Row height
   status      : 80, // Tree status
   bgColor2    : 81, // Alternate background color
   fgColor2    : 82, // Alternate foreground color

   // TabBox attributes.

   tabs        : 90, // Tabs
   selectedTab : 91, // Selected tab

   // Splitter, ProgressBar attributes.

   panes       : 92, // Setup splitter panes
   vertical    : 93, // Split vertical
   position    : 94, // Divider position
   quickExpand : 95, // Quick-expand controls

   // ComboBox attributes.

   items   : 100, // List items
   boxSize : 101, // Box width/height
   open    : 102, // Auto-open when focused

   // ProgressBar attributes.

   value : 110, // Value, empty=indeterminate

   // Image attributes.

   scaling : 111, // Image scaling

   // Plugin attributes.

   plgInt    : 120, // INT attribute value
   plgString : 121  // STRING attribute value
};

/*========================================================================
   static RsValType
   Response changed element value type.
========================================================================*/

eq.RsValType = {
   idFocus       : 1,  // Focus object id, STRING
   ruleOnce      : 2,  // Submit this rule once, INT
   textValue     : 3,  // Text value, STRING
   buttonChecked : 4,  // Button checked, UINT
   caretPos      : 5,  // Caret position, UINT
   activeLine    : 6,  // Active line, UINT
   activeLineCol : 7,  // Active line and column, UINT line/col
   topItem       : 8,  // First visible line, UINT
   selectedTab   : 9,  // Selected tab, UINT
   splitterPos   : 10, // Splitter position, UINT
   scrollPos     : 11, // GroupBox scroll position, UINT x/y
   status        : 12, // ListBox/Tree status, ARRAY
   colOrder      : 13, // ListBox column order, ARRAY
   colWidth      : 14, // ListBox column widths, ARRAY
   lineOrder     : 15, // ListBox line order, ARRAY
   ruleKey       : 16, // Rule submission key code, UINT
   mouseBtns     : 17, //                 mouse buttons, UINT
   mouseCnt      : 18, //                 mouse click count, UINT
   mouseMods     : 19, //                 mouse modifier keys, UINT
   notified      : 20, // Flag: Dialog notified
   dialogMax     : 21, // Dialog maximized, UINT
   dialogPos     : 22, // Dialog position, INT x/y
   dialogSize    : 23, // Dialog size, INT w/h
   plgInt        : 24, // Plugin INT attribute value
   plgString     : 25  // Plugin STRING attribute value
};

/*========================================================================
   static DlgCall
   DLG call constants.
========================================================================*/

eq.DlgCall = {
   set :  6, // DLG SET
   get :  7  // DLG GET
};

/*========================================================================
   static DlgValType
   DLG value type constants.
========================================================================*/

eq.DlgValType = {
   tInt    : 1,
   tString : 2
};

/*========================================================================
   static LayoutMode
   Layout mode constants.
========================================================================*/

eq.LayoutMode = {
   inline    : 1,
   maximized : 2,
   dialog    : 4
};

/*========================================================================
   static DialogMode, DialogState, DialogRaster
   Dialog constants.
========================================================================*/

eq.DialogMode = {
   none       : 0,
   layout     : 1,
   scale      : 2,
   inline     : 3,
   modeMask   : 3, // Mask out mode bits
   stateShift : 2  // Shift down state bits
};

eq.DialogState = {
   maximized : 1
};

eq.DialogRaster = {
   pos  : 1,
   size : 2
};

/*========================================================================
   static Margin
   Dialog/GroupBox margin constants.
========================================================================*/

eq.Margin = {
   top    : 1,
   right  : 2,
   bottom : 4,
   left   : 8
};

/*========================================================================
   static ColState
   ListBox column state constants.
========================================================================*/

eq.ColState = {
   alignMask : 3,   // Mask out align bits
   visible   : 4,   // Column is visible
   raster    : 8,   // If set: column width *= sizeRaster
   movable   : 16,  // Column can be moved
   sizable   : 32,  // Column can be resized
   sortable  : 64,  // Column is sortable
   userHide  : 128  // User can show/hide the column (column context menu)
};

/*========================================================================
   static ColType
   ListBox column type constants.
========================================================================*/

eq.ColType = {
   colAlphaNofold : 0,
   colNumeric     : 1,
   colBoolean     : 2,
   colBarGraph    : 3,
   colAlphaFold   : 5,
   colDate        : 6
};

/*========================================================================
   static RuleType
   CSS rule type constants.
========================================================================*/

eq.RuleType = {
   height     : 1,
   bgColor    : 2,
   fgColor    : 3,
   bgColor2   : 4,
   fgColor2   : 5,
   focusColor : 6
};

/*========================================================================
   static PluginFlags
   Plugin class flags.
========================================================================*/

eq.PluginFlags = {
   focusable  : 4,
   tabonenter : 8
};

/** @preserve $Id: 05-util.js,v 28.50 2021/11/19 15:50:37 mike Exp $
*//*======================================================================
   static Dom
   DOM utilities.
========================================================================*/

eq.Dom = {};

/*------------------------------------------------------------------------
   static Dom.viewRect()
   Obtain current view rectangle.
------------------------------------------------------------------------*/

eq.Dom.viewRect = function() {
   return {
      x : window.pageXOffset,
      y : window.pageYOffset,
      w : document.documentElement.clientWidth,
      h : document.documentElement.clientHeight
   };
};

/*------------------------------------------------------------------------
   static Dom.setContent()
   Set element text or HTML content.
------------------------------------------------------------------------*/

eq.Dom.setContent = function(el, cn, checkAccel) {
   var html = this.reIsHtml.exec(cn);
   if (html) {
      el.textContent = ''; // Delete children.
      this.setHtmlContent(html[1], el);
   }
   else if (checkAccel)
      el.textContent = cn.replace('&', '');
   else
      el.textContent = cn;
};

eq.Dom.setHtmlContent = function(cn, body, head) {
   var
      html = this.ParseHtml.parseFromString(cn, 'text/html'),
      from, child;
   if (head && (from = html.head))
      while (child = from.firstChild)
         head.appendChild(child);
   if (body && (from = html.body))
      while (child = from.firstChild)
         body.appendChild(child);
};

eq.Dom.reIsHtml = new RegExp('^<html>(.*?)(</html>)?$', 'i');
eq.Dom.ParseHtml = new DOMParser();

/*------------------------------------------------------------------------
   static Dom.copyContent()
   Copy element text or HTML content.
------------------------------------------------------------------------*/

eq.Dom.copyContent = function(to, from) {
   to.textContent = ''; // Delete children.
   if (from) {
      var child = from.firstChild;
      while (child) {
         to.appendChild(child.cloneNode(true));
         child = child.nextSibling;
      }
   }
};

/*------------------------------------------------------------------------
   static Dom.updateQueryString()
   Update query string, add/replace or remove parameter.
------------------------------------------------------------------------*/

eq.Dom.updateQueryString = function(url, key, val) {
   var
      r = this.updateQueryString.re.exec(url),
      q = r[2];
   if (q) {
      var re = new RegExp('([?&])' + key + '=[^&#]*', 'i');
      if (q.match(re)) {
         if (val !== undefined)
            q = q.replace(re, '$1' + key + '=' + val);
         else {
            q = q.replace(re, '');
            if (q.length && q.charAt(0) === '&')
               q = q.replace('&', '?');
         }
      }
      else if (val !== undefined)
         q = q + '&' + key + '=' + val;
   }
   else if (val !== undefined)
      q = '?' + key + '=' + val;
   url = r[1];
   if (q)
      url += q;
   if (r[3])
      url += r[3];
   return url;
};

eq.Dom.updateQueryString.re = new RegExp('^([^?#]+)([?][^#]*)?(#.*)?$');

/*------------------------------------------------------------------------
   static Dom.zIndex()
   Return numeric element z-index.
------------------------------------------------------------------------*/

eq.Dom.zIndex = function(el) {
   var z = window.getComputedStyle(el).zIndex;
   return +z === +z ? +z : 0;
};

/*------------------------------------------------------------------------
   static Dom.eqControl()
   Obtain 'eq-control' element.
------------------------------------------------------------------------*/

eq.Dom.eqControl = function(el) {
   var c;
   while (el && el !== document) {
      if (c = el.classList) {
         if (c.contains('eq-control'))
            return el;
         if (c.contains('eq-container'))
            break;
      }
      el = el.parentNode;
   }
   return null;
};

/*------------------------------------------------------------------------
   static Dom.eqOverlay()
   Obtain 'eq-overlay' element.
------------------------------------------------------------------------*/

eq.Dom.eqOverlay = function(el) {
   var c;
   while (el && el !== document) {
      if ((c = el.classList) && c.contains('eq-overlay'))
         return el;
      el = el.parentNode;
   }
   return null;
};

/*------------------------------------------------------------------------
   static Dom.isRootChild()
   Check whether element is child of root element.
------------------------------------------------------------------------*/

eq.Dom.isRootChild = function(el) {
   var cl;
   while (el && el !== document) {
      if (cl = el.classList)
         if (cl.contains('eq-root'))
            return true;
      el = el.parentNode;
   }
   return false;
};

/*------------------------------------------------------------------------
   static Dom.isMenu()
   Check whether element is menu element.
------------------------------------------------------------------------*/

eq.Dom.isMenu = function(el) {
   var cl;
   while (el && el !== document) {
      if (cl = el.classList)
         if (cl.contains('eq-menubar') || cl.contains('eq-menu'))
            return true;
      el = el.parentNode;
   }
   return false;
};

/*------------------------------------------------------------------------
   static Dom.closeMenu()
   Close menu if necessary.
------------------------------------------------------------------------*/

eq.Dom.closeMenu = function() {
   var app = eq.app, id = app.menuOpen;
   if (id) {
      var el = document.getElementById(id);
      if (el && el.classList.contains('eq-open')) {
         el.classList.remove('eq-open');
         var n = el.getElementsByClassName('eq-open');
         for (var i = n.length; --i >= 0;)
            n[i].classList.remove('eq-open');
      }
      app.menuOpen = undefined;
   }
};

/*------------------------------------------------------------------------
   static Dom.parentIndex()
   Return child element parent index.
------------------------------------------------------------------------*/

eq.Dom.parentIndex = function(el) {
   var i = 0;
   while (el = el.previousElementSibling)
      i++;
   return i;
};

/*------------------------------------------------------------------------
   static Dom.inputElement()
   Return descendant input element.
------------------------------------------------------------------------*/

eq.Dom.inputElement = function(el) {
   var n = el.getElementsByTagName('INPUT');
   if (n.length !== 1)
      throw new Error("BUG: InputElement id:" + el.id +
         " elements:" + n.length + ", expected: 1");
   return n[0];
};

/*------------------------------------------------------------------------
   static Dom.consumeEvent()
   Consume event.
------------------------------------------------------------------------*/

eq.Dom.consumeEvent = function(e) {
   e.preventDefault();
   e.stopImmediatePropagation();
};

/*------------------------------------------------------------------------
   static Dom.fireEvent()
   Fire event on element.
------------------------------------------------------------------------*/

eq.Dom.fireEvent = function(t, ty) {
   var e = document.createEvent("Event");
   e.initEvent(ty, true, true);
   t.dispatchEvent(e);
};

/*------------------------------------------------------------------------
   static Dom.fireMouseEvent()
   Fire mouse event on element.
------------------------------------------------------------------------*/

eq.Dom.fireMouseEvent = function(t, ty, ea) {
   var e = document.createEvent("MouseEvent");
   e.initEvent(ty, true, true);
   if (ea !== undefined)
      e.eq = ea;
   t.dispatchEvent(e);
};

/*------------------------------------------------------------------------
   static Dom.fireClick()
   Fire mouse click on element.
------------------------------------------------------------------------*/

eq.Dom.fireClick = function(t, ea) {
   this.fireMouseEvent(t, 'mousedown', ea);
   this.fireMouseEvent(t, 'mouseup', ea);
   this.fireMouseEvent(t, 'click', ea);
};

/*------------------------------------------------------------------------
   static Dom.isEnterOrSpace()
   Check for ENTER or SPACE key.
------------------------------------------------------------------------*/

eq.Dom.isEnterOrSpace = function(e) {
   switch (e.key) {
      case 'Enter':
         return 10;
      case ' ':
      case 'Spacebar': // IE, Firefox <= 37
         return 32;
   }
   return 0;
};

/*------------------------------------------------------------------------
   static Dom.onNextCycle()
   Invoke callback on next event cycle.
------------------------------------------------------------------------*/

eq.Dom.onNextCycle = function(cb, arg) {
   var self = this.onNextCycle, q = self.queue;
   q.push({ cb : cb, arg : arg });
   if (q.length === 1)
      window.setTimeout(self.exec);
};

eq.Dom.onNextCycle.queue = [];

eq.Dom.onNextCycle.exec = function() {
   var self = eq.Dom.onNextCycle, q = self.queue;
   for (var i = 0, l = q.length; i < l; i++) {
      var c = q[i], a = c.arg;
      if (a !== undefined)
         c.cb(a);
      else
         c.cb();
   }
   if (l) {
      q.splice(0, l);
      if (q.length)
         window.setTimeout(self.exec);
   }
};

/*------------------------------------------------------------------------
   static Dom.onNextFrame()
   Invoke callback on next animation frame.
------------------------------------------------------------------------*/

eq.Dom.onNextFrame = function(el, qi) {
   var
      self = this.onNextFrame,
      q = self.queue[self.queue_i],
      qv = q.get(el);
   if (qv === undefined) {
      q.set(el, [ qi ]);
      if (q.size === 1)
         window.requestAnimationFrame(self.exec);
      return;
   }
   var cb = qi.cb;
   for (var i = 0, l = qv.length; i < l; i++)
      if (qv[i].cb === cb) {
         qv[i] = qi;
         return;
      }
   qv.push(qi);
};

eq.Dom.onNextFrame.queue = [ new Map(), new Map() ];
eq.Dom.onNextFrame.queue_i = 0;

eq.Dom.onNextFrame.exec = function() {
   var self = eq.Dom.onNextFrame, q = self.queue[self.queue_i];
   if (++self.queue_i === self.queue.length)
      self.queue_i = 0;
   q.forEach(self.commit);
   q.clear();
};

eq.Dom.onNextFrame.commit = function(qv, el) {
   for (var i = 0, l = qv.length; i < l; i++) {
      var qi = qv[i];
      qi.cb(el, qi);
   }
};

/*========================================================================
   class Timers extends Obj
   Manage multiple timers with common context.
========================================================================*/

eq.c.Obj.subClass('Timers', function() {
   eq.c.Obj.call(this);
   this.m = new Map();
});

/*------------------------------------------------------------------------
   Timers.dispose()
   Stop and dispose timers.
------------------------------------------------------------------------*/

eq.c.Timers.prototype.dispose = function() {
   this.m.forEach(eq.c.Timers.cbDispose);
   this.m.clear();
};

eq.c.Timers.cbDispose = function(t) {
   window.clearTimeout(t.id);
};

/*------------------------------------------------------------------------
   Timers.add()
   Add timer.
------------------------------------------------------------------------*/

eq.c.Timers.prototype.add = function(delay, elapsed, canceled, arg) {
   var t = { elapsed : elapsed, canceled : canceled, arg : arg };
   t.id = window.setTimeout(eq.c.Timers.cbElapsed, delay, this, t);
   this.m.set(t.id, t);
};

/*------------------------------------------------------------------------
   Timers.cancel()
   Stop timers, invoke canceled() callbacks.
------------------------------------------------------------------------*/

eq.c.Timers.prototype.cancel = function() {
   this.m.forEach(eq.c.Timers.cbCancel);
   this.m.clear();
};

eq.c.Timers.cbCancel = function(t) {
   window.clearTimeout(t.id);
   var canceled = t.canceled;
   if (canceled)
      canceled(t.arg);
};

/*------------------------------------------------------------------------
   static Timers.cbElapsed()
   Handle elapsed timer.
------------------------------------------------------------------------*/

eq.c.Timers.cbElapsed = function(self, t) {
   self.m.delete(t.id);
   var elapsed = t.elapsed;
   if (elapsed)
      elapsed(t.arg);
};

/** @preserve $Id: 05-window.js,v 28.8 2021/11/19 15:53:11 mike Exp $
*//*======================================================================
   class Window extends Obj
   Window class.
========================================================================*/

eq.c.Obj.subClass('Window', function(root, parent, dlg) {
   eq.c.Obj.call(this);

   if (root.classList.contains('eq-window')) {
      if (dlg)
         root.classList.add('eq-d' + dlg.id);
   }
   else if (!root.classList.contains('eq-dialog'))
      throw new Error("BUG: Invalid window root");

   // Add to app window list.

   var app = eq.app, wl, wr, i, l, s = root.style;
   if ((wl = app.wl) === undefined)
      app.wl = [];
   else if (wl.length === 0)
      wl = undefined;

   if (!parent)
      parent = document.body;

   if (wl === undefined) {
      wl = app.wl;
      s.zIndex = eq.Dom.zIndex(parent) + 1;
   }
   else {
      for (i = 0, l = wl.length;;) {
         if ((wr = wl[i].root) === undefined)
            throw new Error("BUG: Window root undefined");
         if (++i === l) {
            wr.classList.remove('eq-active');
            s.zIndex = eq.Dom.zIndex(wr) + 1;
            break;
         }
      }
   }

   this.root = root;
   this.dlg = dlg;
   this.modified = {};
   wl.push(this);
   s.visibility = 'hidden';
   parent.appendChild(root);
});

/*------------------------------------------------------------------------
   Window.close()
   Close Window.
------------------------------------------------------------------------*/

eq.c.Window.prototype.close = function() {
   var r, wl, i, l, z, wr;
   if ((r = this.root) === undefined)
      throw new Error("BUG: Window root undefined");

   // Remove from app window list.

   if ((wl = eq.app.wl) === undefined)
      throw new Error("BUG: Window list undefined");

   for (i = wl.length; --i >= 0;)
      if (wl[i] === this) {
         wl.splice(i, 1);
         z = eq.Dom.zIndex(r);
         for (l = wl.length; i < l; i++) {
            if ((wr = wl[i].root) === undefined)
               throw new Error("BUG: Window root undefined");
            wr.style.zIndex = z++;
         }
         break;
      }
   if (i === -1)
      throw new Error("BUG: Window not in list");

   r.parentNode.removeChild(r);
   this.root = undefined;
   this.dlg = undefined;
   this.modified = undefined;
   this.pos = undefined;
   this.whenVisible = undefined;
   this.forcedInline = undefined;

   if (l = wl.length)
      return wl[l - 1];

   return null;
};

/*------------------------------------------------------------------------
   static Window.setupFrame()
   Setup Window frame.
------------------------------------------------------------------------*/

eq.c.Window.setupFrame = function(child) {
   if (   !child.classList.contains('eq-dialog')
       && !child.classList.contains('eq-popup-box'))
      throw new Error("BUG: Invalid window frame child");
   var
      f = document.createElement('DIV'),
      p = document.createElement('DIV'),
      c = document.createElement('DIV'),
      l = document.createElement('DIV'),
      s = document.createElement('SPAN'),
      m = document.createElement('DIV'),
      x = document.createElement('DIV'),
      i, n;
   f.className = 'eq-window';
   p.className = 'eq-pane';
   c.className = 'eq-caption';
   l.className = 'eq-logo eq-move';
   c.appendChild(l);
   s.className = 'eq-move';
   c.appendChild(s);
   m.className = 'eq-max';
   c.appendChild(m);
   x.className = 'eq-close';
   c.appendChild(x);
   p.appendChild(c);
   p.appendChild(child);
   f.appendChild(p);
   for (i = 0; i < 8; i++) {
      switch (i) {
         case 0:
            n = 'eq-top';
            break;
         case 1:
            n = 'eq-top eq-right';
            break;
         case 2:
            n = 'eq-right';
            break;
         case 3:
            n = 'eq-bottom eq-right';
            break;
         case 4:
            n = 'eq-bottom';
            break;
         case 5:
            n = 'eq-bottom eq-left';
            break;
         case 6:
            n = 'eq-left';
            break;
         case 7:
            n = 'eq-top eq-left';
      }
      p = document.createElement('DIV');
      p.className = n + ' eq-resize';
      f.appendChild(p);
   }
   return f;
};

/*------------------------------------------------------------------------
   Window.setPos()
   Set Window position.
------------------------------------------------------------------------*/

eq.c.Window.prototype.setPos = function(x, y) {
   if (x === -1 || y === -1)
      eq.Dom.onNextFrame(this, {
         cb : eq.c.Window.setPos.center,
         x : x,
         y : y
      });
   else {
      var r;
      if ((r = this.root) === undefined)
         throw new Error("BUG: Window root undefined");
      eq.c.Window.setPos(this, r, x, y);
   }
};

eq.c.Window.setPos = function(w, r, x, y) {
   var s = r.style, v;
   if (+x === +x)
      s.left = x > 0 ? x + 'px' : 0;
   else
      s.left = x;
   if (+y === +y)
      s.top = y > 0 ? y + 'px' : 0;
   else
      s.top = y;

   s.visibility = '';
   if ((v = w.whenVisible) !== undefined) {
      w.whenVisible = undefined;
      eq.Dom.onNextFrame(w, v);
   }
};

eq.c.Window.setPos.center = function(w, qi) {
   var
      cls = eq.c.Window,
      r, rrc, x, y, wl, i, p, prc, pw, pwrc;
   if ((r = w.root) === undefined) {
      // Window does no longer exist.
      return;
   }
   rrc = r.getBoundingClientRect();
   x = qi.x;
   y = qi.y;
   if ((wl = eq.app.wl) !== undefined)
      for (i = wl.length; --i >= 0;)
         if (wl[i] === w) {
            if (i > 0) {
               if ((pw = wl[i - 1].root) === undefined)
                  throw new Error("BUG: Window root undefined");
               if ((p = r.offsetParent) !== document.body)
                  prc = p.getBoundingClientRect();
               pwrc = pw.getBoundingClientRect();
               if (x === -1) {
                  x = (pwrc.width - rrc.width) / 2 + pwrc.left;
                  if (prc !== undefined)
                     x -= prc.left;
               }
               if (y === -1) {
                  y = (pwrc.height - rrc.height) / 2 + pwrc.top;
                  if (prc !== undefined)
                     y -= prc.top;
               }
               cls.setPos(w, r, x, y);
               return;
            }
            break;
         }

   if ((p = r.offsetParent) === document.body) {
      var v = eq.Dom.viewRect();
      if (x === -1)
         x = (v.w - rrc.width) / 2 + v.x;
      if (y === -1)
         y = (v.h - rrc.height) / 2 + v.y;
   }
   else {
      prc = p.getBoundingClientRect();
      if (x === -1)
         x = (prc.width - rrc.width) / 2;
      if (y === -1)
         y = (prc.height - rrc.height) / 2;
   }

   cls.setPos(w, r, x, y);
};

/*------------------------------------------------------------------------
   Window.setTitle()
   Set Window title.
------------------------------------------------------------------------*/

eq.c.Window.prototype.setTitle = function(t) {
   var r, c;
   if ((r = this.root) === undefined)
      throw new Error("BUG: Window root undefined");
   if (c = r.querySelector('.eq-window>.eq-pane>.eq-caption>span'))
      c.textContent = t;
};

/*------------------------------------------------------------------------
   Window.activate()
   Activate Window.
------------------------------------------------------------------------*/

eq.c.Window.prototype.activate = function() {
   var r, wl, i, l, z, wr, cl;
   if ((r = this.root) === undefined)
      throw new Error("BUG: Window root undefined");

   if ((wl = eq.app.wl) === undefined)
      throw new Error("BUG: Window list undefined");

   // Make topmost in app window list.

   for (l = wl.length, i = l; --i >= 0;) {
      if (wl[i] === this) {
         if (i !== l - 1) {
            wl.splice(i, 1);
            z = eq.Dom.zIndex(r);
            for (l = wl.length; i < l;) {
               if ((wr = wl[i].root) === undefined)
                  throw new Error("BUG: Window root undefined");
               wr.style.zIndex = z++;
               if (++i === l)
                  wr.classList.remove('eq-active');
            }
            r.style.zIndex = z;
            wl.push(this);
         }
         r.classList.add('eq-active');
         break;
      }
   }
   if (i === -1)
      throw new Error("BUG: Window not in list");
};

/*------------------------------------------------------------------------
   Window.makeInline()
   Convert Window to inline Dialog if necessary.
------------------------------------------------------------------------*/

eq.c.Window.prototype.makeInline = function(save) {
   var r, d, dc, c, s;
   if ((r = this.root) === undefined)
      throw new Error("BUG: Window root undefined");
   if (r.classList.contains('eq-window')) {
      if (!(d = r.querySelector('.eq-window>.eq-pane>.eq-dialog')))
         throw new Error("BUG: no Dialog in Window");
      if (save) {
         dc = d.querySelector(
            '.eq-dialog>.eq-pane>.eq-container:not(.eq-panel)');
         c = r.querySelector('.eq-window>.eq-pane>.eq-caption>span');
         this.forcedInline = {
            x : r.offsetLeft,
            y : r.offsetTop,
            w : dc.offsetWidth,
            h : dc.offsetHeight,
            t : c ? c.textContent : undefined,
            r : r.classList.contains('eq-resize'),
            c : r.classList.contains('eq-close'),
            m : r.classList.contains('eq-max')
         };
      }
      (s = d.style).visibility = 'hidden';
      s.zIndex = eq.Dom.zIndex(r);
      d.classList.add('eq-inline');
      r.parentNode.replaceChild(d, r);
      return this.root = d;
   }
   return null;
};

/*------------------------------------------------------------------------
   Window.makeFramed()
   Convert inline Dialog to Window if necessary.
------------------------------------------------------------------------*/

eq.c.Window.prototype.makeFramed = function(restore) {
   var
      cls = eq.c.Window,
      d, fi = this.forcedInline, parent, z, r, s, cl, dc;
   if ((d = this.root) === undefined)
      throw new Error("BUG: Window root undefined");
   if (d.classList.contains('eq-dialog') && (!restore || fi !== undefined)) {
      z = eq.Dom.zIndex(d);
      r = cls.setupFrame((parent = d.parentNode).removeChild(d));
      (s = r.style).visibility = 'hidden';
      s.zIndex = z;
      d.style.zIndex = '';
      (cl = d.classList).remove('eq-inline');
      cl.remove('eq-max');
      if (restore) {
         this.forcedInline = undefined;
         dc = d.querySelector(
            '.eq-dialog>.eq-pane>.eq-container:not(.eq-panel)');
         s.left = fi.x > 0 ? fi.x + 'px' : 0;
         s.top = fi.y > 0 ? fi.y + 'px' : 0;
         (s = dc.style).width = fi.w > 0 ? fi.w + 'px' : 0;
         s.height = fi.h > 0 ? fi.h + 'px' : 0;
         cl = r.classList;
         cl.add('eq-d' + this.dlg.id);
         if (fi.t !== undefined)
            r.querySelector(
               '.eq-window>.eq-pane>.eq-caption>span').textContent = fi.t;
         if (fi.r)
            cl.add('eq-resize');
         else
            cl.remove('eq-resize');
         if (fi.c)
            cl.add('eq-close');
         else
            cl.remove('eq-close');
         if (fi.m)
            cl.add('eq-max');
         else
            cl.remove('eq-max');
      }
      parent.appendChild(r);
      return this.root = r;
   }
   return null;
};

/*------------------------------------------------------------------------
   static Window.zOrderModified()
   Window list stack order has been modified.
------------------------------------------------------------------------*/

eq.c.Window.zOrderModified = function() {
   var app = eq.app, wl, lm, i, w, r, cl, active, hide;
   if ((wl = app.wl) !== undefined) {

      // Inline base Dialog, forced by layout mode?
      if (   (lm = app.layoutMode) !== undefined
          && (!(lm & eq.LayoutMode.inline) || !(lm & eq.LayoutMode.dialog)))
         lm = undefined;

      // Adjust Windows from top to bottom.

      for (active = true, i = wl.length; --i >= 0;) {
         w = wl[i];
         if ((r = w.root) === undefined)
            throw new Error("BUG: Window root undefined");
         if (lm !== undefined && w.dlg !== undefined) {
            if (i === 0) {
               // Base Dialog, make inline.
               if (r = w.makeInline(true)) {
                  if (lm & eq.LayoutMode.maximized)
                     r.classList.add('eq-max');
                  else
                     r.classList.remove('eq-max');
                  r.style.visibility = '';
               }
            }
            else {
               // Make framed if previously made inline.
               if (r = w.makeFramed(true))
                  r.style.visibility = '';
            }
         }
         if ((r = w.root) === undefined)
            throw new Error("BUG: Window root undefined");
         cl = r.classList;
         if (active === undefined)
            cl.remove('eq-active');
         else {
            active = undefined;
            cl.add('eq-active');
         }
         if (hide)
            cl.add('eq-hidden');
         else {
            cl.remove('eq-hidden');
            // Windows below topmost inline Dialog become invisible.
            if (!cl.contains('eq-window'))
               hide = true;
         }
      }
   }
};

/*------------------------------------------------------------------------
   static Window.fromTarget()
   Obtain Window from target element.
------------------------------------------------------------------------*/

eq.c.Window.fromTarget = function(t) {
   var wl, cl, i, w;

   // Locate target in app window list.

   if ((wl = eq.app.wl) !== undefined) {
      while (t && t !== document) {
         if (cl = t.classList) {
            if (cl.contains('eq-dialog') || cl.contains('eq-popup-box'))
               break;
            if (cl.contains('eq-window')) {
               for (i = wl.length; --i >= 0;)
                  if ((w = wl[i]).root === t)
                     return w;
               break;
            }
         }
         t = t.parentNode;
      }
   }
   return null;
};

/*------------------------------------------------------------------------
   Window.onMouseDownOrClick()
   Window 'mousedown' and 'click' event handler.
------------------------------------------------------------------------*/

eq.c.Window.moveMode = {
   left   : 1,
   top    : 2,
   right  : 4,
   bottom : 8,
   size   : 16
};

eq.c.Window.prototype.onMouseDownOrClick = function(e) {
   if (e.type === 'mousedown') {
      var
         cls = eq.c.Window,
         r = this.root,
         t = e.target,
         p = t.parentNode,
         cl = t.classList,
         app, mm, mode;
      if (p && cl) {
         app = eq.app;
         mm = cls.moveMode;
         if (cl.contains('eq-resize') && p === r) {
            if (cl.contains('eq-top')) {
               mode = mm.top + mm.size;
               if (cl.contains('eq-left'))
                  mode += mm.left;
               else if (cl.contains('eq-right'))
                  mode += mm.right;
            }
            else if (cl.contains('eq-bottom')) {
               mode = mm.bottom;
               if (cl.contains('eq-left'))
                  mode += mm.left + mm.size;
               else if (cl.contains('eq-right'))
                  mode += mm.right;
            }
            else if (cl.contains('eq-left'))
               mode = mm.left + mm.size;
            else if (cl.contains('eq-right'))
               mode = mm.right;
         }
         else if (   p.classList.contains('eq-caption')
                  && (p = p.parentNode) && p.classList.contains('eq-pane')
                  && (p = p.parentNode) && p === r) {
            if (cl.contains('eq-close')) {
               cl.add('eq-click');
               var d;
               if ((d = this.dlg) !== undefined) {
                  if (d === app.currDlg)
                     app.submit(document.body, { type : 'close' });
                  else
                     d.pendingClose = true;
               }
               return true;
            }
            if (cl.contains('eq-max')) {
               if ((cl = p.classList).contains('eq-max')) {
                  cl.remove('eq-max');
                  var pos = this.pos;
                  if (pos !== undefined) {
                     this.pos = undefined;
                     cls.setPos(this, r, pos.x, pos.y);
                  }
               }
               else {
                  this.pos = { x : r.offsetLeft, y : r.offsetTop };
                  cl.add('eq-max');
               }
               this.modified.maximized = true;
               return true;
            }
            if (cl.contains('eq-move'))
               if (!p.classList.contains('eq-max'))
                  mode = mm.left + mm.top;
         }
         if (mode !== undefined) {
            var m = {
               w : this,
               x : e.clientX,
               y : e.clientY,
               mode : mode
            };
            if ((p = r.offsetParent) === document.body)
               m.px = m.py = 0;
            else {
               var prc = p.getBoundingClientRect();
               m.px = prc.left;
               m.py = prc.top;
            }
            app.moveCapture(m, cls.onMove);
            return true;
         }
      }
   }
   return false;
};

/*------------------------------------------------------------------------
   static Window.onMove()
   Move Window.
------------------------------------------------------------------------*/

eq.c.Window.onMove = function(e, m) {
   if (!m.busy) {
      m.busy = true;
      eq.Dom.onNextFrame(m.w, {
         cb : eq.c.Window.onMove.exec,
         x : e.clientX,
         y : e.clientY,
         m : m
      });
   }
};

eq.c.Window.onMove.exec = function(w, qi) {
   var
      cls = eq.c.Window,
      mm = cls.moveMode,
      cx = qi.x,
      cy = qi.y,
      m = qi.m,
      mode = m.mode,
      sz = mode & mm.size,
      r = w.root,
      rc, chk;
   if (r === undefined)
      throw new Error("BUG: Window root undefined");
   rc = r.getBoundingClientRect();
   if (mode & mm.left) {
      // Left position modified.
      if (cx !== m.x && (m.maxX === undefined || cx <= m.maxX)) {
         var
            left = rc.left + window.pageXOffset - m.px,
            x = left + cx - m.x;
         if (x < 0) {
            x = 0;
            cx = m.x - left;
         }
         r.style.left = x ? x + 'px' : 0;
         w.modified.pos = true;
         if (!sz)
            m.x = cx;
         else {
            // Width modified.
            if (chk === undefined) {
               var dc = r.querySelector(
                           '.eq-dialog>.eq-pane>.eq-container:not(.eq-panel)');
               chk = { rc : rc, dc : dc, drc : dc.getBoundingClientRect() };
            }
            chk.dc.style.width = (left + chk.drc.width - x) + 'px';
            chk.cx = cx;
            chk.cw = rc.width;
         }
      }
   }
   if (mode & mm.top) {
      // Top position modified.
      if (cy !== m.y && (m.maxY === undefined || cy <= m.maxY)) {
         var
            top = rc.top + window.pageYOffset - m.py,
            y = top + cy - m.y;
         if (y < 0) {
            y = 0;
            cy = m.y - top;
         }
         r.style.top = y ? y + 'px' : 0;
         w.modified.pos = true;
         if (!sz)
            m.y = cy;
         else {
            // Height modified.
            if (chk === undefined) {
               var dc = r.querySelector(
                           '.eq-dialog>.eq-pane>.eq-container:not(.eq-panel)');
               chk = { rc : rc, dc : dc, drc : dc.getBoundingClientRect() };
            }
            chk.dc.style.height = (top + chk.drc.height - y) + 'px';
            chk.cy = cy;
            chk.cw = rc.height;
         }
      }
   }
   if (mode & mm.right) {
      // Width modified.
      if (cx !== m.x && (m.minX === undefined || cx >= m.minX)) {
         if (chk === undefined) {
            var dc = r.querySelector(
                        '.eq-dialog>.eq-pane>.eq-container:not(.eq-panel)');
            chk = { rc : rc, dc : dc, drc : dc.getBoundingClientRect() };
         }
         chk.dc.style.width = (chk.drc.width + cx - m.x) + 'px';
         chk.cx = cx;
         chk.cw = rc.width;
      }
   }
   if (mode & mm.bottom) {
      if (cy !== m.y && (m.minY === undefined || cy >= m.minY)) {
         // Height modified.
         if (chk === undefined) {
            var dc = r.querySelector(
                        '.eq-dialog>.eq-pane>.eq-container:not(.eq-panel)');
            chk = { rc : rc, dc : dc, drc : dc.getBoundingClientRect() };
         }
         chk.dc.style.height = (chk.drc.height + cy - m.y) + 'px';
         chk.cy = cy;
         chk.ch = rc.height;
      }
   }
   if (chk === undefined)
      m.busy = undefined;
   else
      eq.Dom.onNextFrame(w, {
         cb : cls.onMove.chk,
         m : m,
         chk : chk
      });
};

eq.c.Window.onMove.chk = function(w, qi) {
   var
      m = qi.m,
      r = w.root,
      chk = qi.chk,
      rc = r.getBoundingClientRect();
   // New Window size may need to stabilize, could require multiple frames.
   if (   (chk.cw !== undefined && chk.cw !== rc.width)
       || (chk.ch !== undefined && chk.ch !== rc.height)) {
      if (chk.cw !== undefined)
         chk.cw = rc.width;
      if (chk.ch !== undefined)
         chk.ch = rc.height;
   }
   // When done, wait another frame.
   else if (chk.done) {
      m.busy = undefined;
      return;
   }
   // Check width/height constraints.
   else {
      if (chk.cx !== undefined) {
         if (rc.width !== chk.rc.width) {
            m.x = chk.cx;
            m.minX = m.maxX = undefined;
            w.modified.size = true;
         }
         else {
            // Constrained by min/max width, reset to previous position/size.
            r.style.left = (chk.rc.left + window.pageXOffset - m.px) + 'px';
            chk.dc.style.width = chk.drc.width + 'px';
            if (m.minX === undefined && m.maxX === undefined) {
               if (chk.cx < m.x) {
                  m.minX = m.x;
                  m.maxX = undefined;
               }
               else {
                  m.maxX = m.x;
                  m.minX = undefined;
               }
            }
         }
      }
      if (chk.cy !== undefined) {
         if (rc.height !== chk.rc.height) {
            m.y = chk.cy;
            m.minY = m.maxY = undefined;
            w.modified.size = true;
         }
         else {
            // Constrained by min/max height, reset to previous position/size.
            r.style.top = (chk.rc.top + window.pageYOffset - m.py) + 'px';
            chk.dc.style.height = chk.drc.height + 'px';
            if (m.minY === undefined && m.maxY === undefined) {
               if (chk.cy < m.y) {
                  m.minY = m.y;
                  m.maxY = undefined;
               }
               else {
                  m.maxY = m.y;
                  m.minY = undefined;
               }
            }
         }
      }
      chk.done = true;
   }
   eq.Dom.onNextFrame(w, qi);
};

/** @preserve $Id: 10-dlg.js,v 28.7 2021/11/19 15:53:11 mike Exp $
*//*======================================================================
   WEBDLG v2 namespace / globals
   Map of DLG element classes.
========================================================================*/

eq.d = eq.d || new Map();

/*========================================================================
   class DlgElement extends Obj
   DLG element base class.
========================================================================*/

eq.c.Obj.subClass('DlgElement', function() {
   eq.c.Obj.call(this);
   this.cssRules = [];
});

/*------------------------------------------------------------------------
   static DlgElement.addClass()
   Subclassing convenience helper: Add DLG element class.
------------------------------------------------------------------------*/

eq.c.DlgElement.addClass = function(name, cls) {
   this.subClass(name, cls);
   eq.d.set(name, cls);
   return cls;
};

/*------------------------------------------------------------------------
   DlgElement.dispose()
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.dispose = function(id) {
   var
      rules = this.cssRules,
      l = rules.length;
   if (l) {
      var
         stl = eq.app.stl,
         sheet = stl.sheet,
         owner = stl.owner,
         rl = owner.length,
         i = 0, ri, ci;
      if ((ri = ci = owner.indexOf(id)) === -1)
         throw new Error(this.constructor.name + ".dispose(" + id +
            ") " + l + " stale CSS rules");
      do {
         sheet.deleteRule(ci);
         i++;
      } while (++ri < rl && owner[ri] === id);
      owner.splice(ci, i);
      rules.length = 0;
      if (i !== l)
         throw new Error(this.constructor.name + ".dispose(" + id +
            ") CSS rules inconsistent:" + l + "/" + i);
   }
};

/*------------------------------------------------------------------------
   DlgElement.updateID()
   Element identifier has been modified.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.updateId = function(id, oldId) {
   var
      rules = this.cssRules,
      l = rules.length;
   if (l) {
      var
         cls = eq.c.DlgElement,
         stl = eq.app.stl,
         sheet = stl.sheet,
         owner = stl.owner,
         rl = owner.length,
         i = 0, ri;
      if ((ri = owner.indexOf(oldId)) === -1)
         throw new Error(this.constructor.name + ".updateId(" + oldId +
            ") " + l + " stale CSS rules");
      do {
         sheet.deleteRule(ri);
         cls.insertCssRule(sheet, ri, id, rules[i++].rule);
         owner[ri] = id;
      } while (++ri < rl && owner[ri] === oldId);
      if (i !== l)
         throw new Error(this.constructor.name + ".updateId(" + oldId +
            ") CSS rules inconsistent:" + l + "/" + i);
   }
   else if ((l = owner.indexOf(oldId)) !== -1)
      throw new Error(this.constructor.name + ".updateId(" + oldId +
         ") CSS rules inconsistent:" + l);
};

/*------------------------------------------------------------------------
   DlgElement.setAttr()
   Set element attributes.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.setAttr = function(dlg, id, el, rt) {
   throw new Error(this.constructor.name + ".setAttr(" + id +
      ") unexpected invocation");
};

/*------------------------------------------------------------------------
   DlgElement.onEvent()
   Filter and/or handle action event.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.onEvent = function(dlg, id, el, e) {
   return true;
};

/*------------------------------------------------------------------------
   DlgElement.onKey()
   Filter and/or handle key event.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.onKey = function(dlg, id, el, e) {
   return false;
};

/*------------------------------------------------------------------------
   DlgElement.onTypeAhead()
   Process type-ahead key stroke.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.onTypeAhead = function(dlg, id, el, key) {
   var tab, shiftKey;
   switch (key) {
      case 'ShiftTab':
         shiftKey = true;
         // FALLTHROUGH
      case 'Tab':
         tab = true;
         break;
      case 'ShiftEnter':
         shiftKey = true;
         // FALLTHROUGH
      case 'Enter': {
         var cl = el.classList;
         if (cl.contains('eq-enter')) {
            // Invalid: 'Enter key has special function' not handled.
            return false;
         }
         if (!cl.contains('eq-noenter')) {
            // .tabonenter enabled.
            tab = true;
         }
      }
   }

   if (tab) {
      // Navigate to next/previous element in tab order.
      var
         app = eq.app,
         elNext = dlg.nextTab(id, shiftKey);
      if (elNext)
         app.setFocus(elNext.id, elNext);
      else {
         // No other element found in tab order,
         // submit 'blur' event if registered.
         app.processEvent(dlg, {
            type : 'blur',
            target : this.focusElement(dlg, el)
         }, id, el, this);
      }
      return true;
   }

   return false;
};

/*------------------------------------------------------------------------
   DlgElement.onChange()
   Filter and/or handle change event.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.onChange = function(dlg, id, el, e) {
   return true;
};

/*------------------------------------------------------------------------
   DlgElement.focusElement()
   Obtain focus element.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.focusElement = function(dlg, el) {
   return el;
};

/*------------------------------------------------------------------------
   DlgElement.focusGained()
   Element has gained focus.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.focusGained = function(dlg, ct) {
};

/*------------------------------------------------------------------------
   DlgElement.focusLost()
   Element has lost focus.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.focusLost = function(dlg, el) {
};

/*------------------------------------------------------------------------
   DlgElement.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.changed = function(dlg, id, el, ty) {
   return null;
};

/*------------------------------------------------------------------------
   DlgElement.updated()
   Act upon API thread update response.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.updated = function(dlg, id, el, tag, arg) {
   var ut = eq.UpdateTag;
   switch (tag) {
      case ut.mouseButton2: {
         var
            app = eq.app,
            rule = arg.rule;
         app.onceRule = rule;
         app.onceId = id;
         app.submit(el, {
            id : id,
            type : 'click'
         });
         break;
      }

      case ut.contextMenu: {
         var
            app = eq.app,
            et = arg.et,
            ov = document.createElement('UL'),
            menu, st;
         // Create context menu overlay.
         menu = dlg.addControl(et);
         (st = menu.style).left = arg.x + window.pageXOffset + 'px';
         st.top = arg.y + window.pageYOffset + 'px';
         // Save requesting element id.
         dlg.eds.get(menu.id).contextMenuId = arg.id;
         app.makeOverlay(menu, menu, eq.c.Menu.closeContextMenu);
         break;
      }

      default:
         throw new Error(this.constructor.name + ".updated(" + id +
            ") invalid tag " + tag);
   }
};

/*------------------------------------------------------------------------
   DlgElement.setCssRule()
   Set application CSS rule for element.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.setCssRule = function(id, type, rule) {
   var
      cls = eq.c.DlgElement,
      rules = this.cssRules,
      stl = eq.app.stl,
      sheet = stl.sheet,
      owner = stl.owner,
      rl = owner.length,
      i, l, r, ri, ci;
   for (i = 0, l = rules.length; i < l; i++)
      if (type === (r = rules[i]).type) {
         // Modify or delete existing rule.
         if ((ri = owner.indexOf(id)) !== -1) {
            ci = i;
            while (i-- > 0)
               if (++ri >= rl || owner[ri] !== id) {
                  ri = -1;
                  break;
               }
         }
         if (ri === -1)
            throw new Error(this.constructor.name + ".setCssRule(" + id +
               ") not found, type:" + type);
         sheet.deleteRule(ri);
         if (rule) {
            r.rule = rule;
            cls.insertCssRule(sheet, ri, id, rule);
         }
         else {
            rules.splice(ci, 1);
            owner.splice(ri, 1);
         }
         return;
      }
   // Add new rule.
   if ((ri = owner.indexOf(id)) !== -1) {
      i = 1;
      while (++ri < rl && owner[ri] === id)
         i++;
      if (i !== l)
         throw new Error(this.constructor.name + ".setCssRule(" + id +
            ") inconsistent:" + l + "/" + i + ", type:" + type);
   }
   else {
      if (l)
         throw new Error(this.constructor.name + ".setCssRule(" + id +
            ") inconsistent:" + l + ", type:" + type);
      ri = rl;
   }
   if (rule) {
      rules.push({ rule: rule, type : type });
      cls.insertCssRule(sheet, ri, id, rule);
      owner.splice(ri, 0, id);
   }
};

eq.c.DlgElement.insertCssRule = function(sheet, idx, id, rule) {
   var s;
   if (Array.isArray(rule)) {
      s = '';
      for (var i = 0, l = rule.length; i < l; i++)
         s += '#' + id + rule[i];
   }
   else
      s = '#' + id + rule;
   sheet.insertRule(s, idx);
};

/*------------------------------------------------------------------------
   DlgElement.duplicateCssRules()
   Duplicate application CSS element rules to new ID.
------------------------------------------------------------------------*/

eq.c.DlgElement.prototype.duplicateCssRules = function(id) {
   var
      rules = this.cssRules,
      l = rules.length;
   if (l) {
      var
         cls = eq.c.DlgElement,
         stl = eq.app.stl,
         sheet = stl.sheet,
         owner = stl.owner,
         rl = owner.length,
         i, nid;
      for (i = 1;; i++)
         if (owner.indexOf(nid = id + '-' + i) === -1)
            break;
      for (i = 0; i < l; i++, rl++) {
         cls.insertCssRule(sheet, rl, nid, rules[i].rule);
         owner.push(nid);
      }
      return nid;
   }
   return '';
};

/*========================================================================
   final class Dlg extends Obj
   DLG Dialog instance.
========================================================================*/

eq.c.Obj.subClass('Dlg', function(id) {
   eq.c.Obj.call(this);

   // Dialog id.
   this.id = id;

   // Window.
   this.w = null;

   // Root elements.
   this.root = [];

   // Element descriptors.
   this.eds = new Map();

   // Tab order.
   this.tabOrder = null;
});

delete eq.c.Dlg.subClass; // final

/*------------------------------------------------------------------------
   Dlg.close()
   Close Dialog instance.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.close = function() {
   var cls = this.constructor, app = eq.c.App;

   // Reset current Dialog if necessary.
   this.resetCurrDlg();

   // Restore original page title if necessary.
   if (document.title === this.title) {
      var t = app.title;
      if (t !== document.title) {
         document.title = t;
         eq.app.updateSessionList();
      }
   }

   // Remove root elements.
   for (var i = this.root.length; --i >= 0;) {
      var el = this.root[i];
      app.removeRootListeners(el);
      if (!el.classList.contains('eq-dialog'))
         el.parentNode.removeChild(el);
      else {
         // Close Window.
         if (this.w === null)
            throw new Error("Dlg.close(" + this.id + ") BUG: no Window");
         this.w.close();
         this.w = null;
      }
   }
   this.root.length = 0;
   if (this.w !== null)
      throw new Error("Dlg.close(" + this.id + ") BUG: Window not closed");

   // Clear element descriptors.
   this.eds.forEach(cls.cbDisposeDlgElement);
   this.eds.clear();

   // Clear tab order.
   this.tabOrder = null;

   // Clear pending close request.
   this.pendingClose = undefined;

   // Delete properties.
   this.layout = undefined;
   this.title = undefined;
   this.fkey = undefined;
   this.timer = undefined;
   this.ddo = undefined;
};

eq.c.Dlg.cbDisposeDlgElement = function(ed, id) {
   ed.dispose(id);
};

eq.c.Dlg.prototype.resetCurrDlg = function() {
   var app = eq.app;
   if (this === app.currDlg) {
      app.currDlg = null;
      app.focus = null;
      app.changed.clear();
      app.setIdle(this);
      this.ddo = undefined;
   }
};

/*------------------------------------------------------------------------
   Dlg.addRoot()
   Add Dialog root element.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.addRoot = function(el) {

   // TODO: dlgAnchor

   var cl = el.classList, wc, m, lm, l, r;
   el.style.display = 'none';
   cl.add('eq-root');
   if (cl.contains('eq-dialog')) {
      if (this.w !== null)
         throw new Error("Dlg.addRoot(" + this.id + ") BUG: Window exists");
      wc = eq.c.Window;
      m = eq.LayoutMode;
      if (   (lm = eq.app.layoutMode) !== undefined
          && (lm & m.inline) && !(lm & m.dialog)) {
         // Inline Dialog, forced by layout mode.
         r = el;
      }
      else {
         lm = undefined;
         if (   (l = this.layout) === undefined
             || l.mode === eq.DialogMode.inline) {
            // Inline Dialog.
            r = el;
         }
         else {
            // Dialog in frame.
            r = wc.setupFrame(el);
         }
      }
      this.w = new wc(r, document.getElementById('eq-dialog'), this);
      if (lm !== undefined) {
         // Inline Dialog, forced by layout mode.
         cl.add('eq-inline');
         if (lm & m.maximized)
            cl.add('eq-max');
         else
            cl.remove('eq-max');
      }
   }
   else
      throw new Error("Dlg.addRoot(" + this.id + ") BUG: not a Dialog");

   this.root.push(el);
   eq.c.App.addRootListeners(el);
};

/*------------------------------------------------------------------------
   Dlg.removeElement()
   Remove element from Dialog.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.removeElement = function(el) {
   var r = this.root, eds = this.eds, i, empty, n, id, ed;
   if ((i = r.indexOf(el)) !== -1) {
      // Root element.
      r.splice(i, 1);
      empty = r.length === 0;
   }

   if (!empty) {
      // Remove from element descriptors.
      n = el.querySelectorAll('[id]');
      for (i = n.length; --i >= 0;) {
         if ((ed = eds.get(id = n[i].id)) !== undefined) {
            ed.dispose(id);
            eds.delete(id);
         }
      }
      if ((ed = eds.get(id = el.id)) !== undefined) {
         ed.dispose(id);
         eds.delete(id);
      }
   }

   if (empty === undefined)
      el.parentNode.removeChild(el);
   else {
      // Root element.
      eq.c.App.removeRootListeners(el);
      if (!el.classList.contains('eq-dialog'))
         el.parentNode.removeChild(el);
      else {
         // Close Window.
         if (this.w === null)
            throw new Error("Dlg.removeElement(" + this.id +
               ") BUG: no Window");
         this.w.close();
         this.w = null;
      }
   }

   return empty;
};

/*------------------------------------------------------------------------
   Dlg.visible()
   Make Dialog visible.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.visible = function() {
   var cls = this.constructor;
   for (var ri = this.root.length; --ri >= 0;) {
      var rl = this.root[ri], st;
      if (rl.classList.contains('eq-dialog')) {
         if (this.w === null)
            throw new Error("Dlg.visible(" + this.id + ") BUG: no Window");

         // Validate existing panel elements,
         // relink hidden MenuBar and hidden/misplaced ToolBar(s).

         // Try to avoid flicker while relinking.
         if ((st = rl.style).display === 'none') {
            st.visibility = 'hidden';
            st.display = '';
         }
         if (   rl !== this.w.root
             && (st = this.w.root.style).display === 'none') {
            st.visibility = 'hidden';
            st.display = '';
         }

         var
            dc = rl.querySelector(
               '.eq-dialog>.eq-pane>.eq-container:not(.eq-panel)'),
            dp = dc.parentNode,
            el = rl.firstElementChild,
            tc = 'eq-top',
            cl, ch;
         while (el) {
            if ((cl = el.classList).contains('eq-panel')) {
               if (   cl.contains('eq-hidden')
                   || (cl.contains('eq-toolbar') && !cl.contains(tc))) {
                  var nx = el.nextElementSibling;
                  dc.appendChild(el);
                  cl.remove('eq-rt');
                  cl.remove('eq-panel');
                  cl.remove('eq-horz');
                  el = nx;
                  continue;
               }
            }
            else if (cl.contains('eq-pane')) {
               ch = el.firstElementChild;
               tc = 'eq-left';
               while (ch) {
                  if ((cl = ch.classList).contains('eq-panel')) {
                     if (   cl.contains('eq-hidden')
                         || (cl.contains('eq-toolbar') && !cl.contains(tc))) {
                        var nx = ch.nextElementSibling;
                        dc.appendChild(ch);
                        cl.remove('eq-rt');
                        cl.remove('eq-panel');
                        cl.remove('eq-vert');
                        ch = nx;
                        continue;
                     }
                  }
                  else if (cl.contains('eq-container'))
                     tc = 'eq-right';
                  ch = ch.nextElementSibling;
               }
               tc = 'eq-bottom';
            }
            el = el.nextElementSibling;
         }

         // MenuBar.
         if ((cl = (el = rl.firstElementChild).classList).contains('eq-idle'))
            cl = (el = el.nextElementSibling).classList;
         if (!cl.contains('eq-menubar')) {
            // Put first visible MenuBar into top-most panel.
            if (ch = dc.querySelector('.eq-menubar:not(.eq-hidden)')) {
               (cl = ch.classList).add('eq-rt');
               cl.add('eq-panel');
               rl.insertBefore(ch, el);
            }
         }

         // StatusBar.
         if ((cl = (el = rl.lastElementChild).classList).contains('eq-idle'))
            cl = (el = el.previousElementSibling).classList;
         if (!cl.contains('eq-statusbar')) {
            // Put first visible StatusBar into bottom-most panel.
            if (ch = dc.querySelector('.eq-statusbar:not(.eq-hidden)')) {
               (cl = ch.classList).add('eq-rt');
               cl.add('eq-panel');
               rl.appendChild(ch);
            }
         }

         // ToolBars.
         var n = dc.querySelectorAll('.eq-toolbar:not(.eq-hidden)');
         for (var i = 0, l = n.length; i < l; i++) {
            if ((cl = (el = n[i]).classList).contains('eq-top')) {
               cl.add('eq-rt');
               cl.add('eq-panel');
               cl.add('eq-horz');
               if ((ch = rl.firstElementChild).classList.contains('eq-idle'))
                  ch = ch.nextElementSibling;
               if (ch.classList.contains('eq-menubar'))
                  rl.insertBefore(el, ch.nextElementSibling);
               else
                  rl.insertBefore(el, ch);
            }
            else if (cl.contains('eq-bottom')) {
               cl.add('eq-rt');
               cl.add('eq-panel');
               cl.add('eq-horz');
               if ((ch = rl.lastElementChild).classList.contains('eq-idle'))
                  ch = ch.previousElementSibling;
               if (ch.classList.contains('eq-statusbar'))
                  rl.insertBefore(el, ch);
               else
                  rl.appendChild(el);
            }
            else if (cl.contains('eq-left')) {
               cl.add('eq-rt');
               cl.add('eq-panel');
               cl.add('eq-vert');
               dp.insertBefore(el, dp.firstElementChild);
            }
            else if (cl.contains('eq-right')) {
               cl.add('eq-rt');
               cl.add('eq-panel');
               cl.add('eq-vert');
               dp.appendChild(el);
            }
         }

         // Make Window visible on next frame,
         // set position and title if necessary.
         eq.Dom.onNextFrame(this, { cb : cls.commitVisible, rl : rl });
      }
      else {
         // Make root element visible.
         rl.style.display = '';
      }
   }
};

eq.c.Dlg.commitVisible = function(d, qi) {
   var
      cls = eq.c.Dlg,
      m = eq.LayoutMode,
      dm = eq.DialogMode,
      w = d.w,
      rl = qi.rl,
      r, l, lm, cl, whenVisible;
   if (w === null) {
      // Window does no longer exist.
      return;
   }
   if ((r = w.root) === undefined)
      throw new Error("Dlg.commitVisible(" + d.id +
         ") BUG: Window root undefined");
   whenVisible = { cb : cls.whenVisible };
   if ((l = d.layout) === undefined) {
      rl.style.visibility = '';
      r.style.visibility = '';
   }
   else if (   (lm = eq.app.layoutMode) !== undefined
            && (lm & m.inline) && !(lm & m.dialog)) {
      // Inline Dialog, forced by layout mode.
      d.layout = undefined;
      rl.style.visibility = '';
   }
   else {
      d.layout = undefined;
      if (l.mode !== undefined) {
         if (l.mode === dm.inline) {
            // Inline Dialog.
            if (w.makeInline())
               r = w.root;
         }
         else {
            // Dialog in frame.
            if (w.makeFramed())
               r = w.root;
         }
      }
      if ((cl = r.classList).contains('eq-dialog'))
         rl.style.visibility = '';
      else {
         // Dialog in frame.
         w.setTitle(d.title ? d.title : "Eloquence");
         if (l.mode !== undefined)
            if (l.mode === dm.layout) {
               cl.add('eq-resize');
               if (l.state !== undefined)
                  if (l.state & eq.DialogState.maximized)
                     cl.add('eq-max');
                  else
                     cl.remove('eq-max');
            }
            else {
               cl.remove('eq-resize');
               cl.remove('eq-max');
            }
         if (rl.classList.contains('eq-rule'))
            cl.add('eq-close');
         else {
            cl.remove('eq-close');
            d.pendingClose = undefined;
         }
         rl.style.visibility = '';
         if (l.x !== undefined && l.y !== undefined) {
            w.whenVisible = whenVisible;
            whenVisible = undefined;
            w.setPos(l.x, l.y);
         }
         else
            r.style.visibility = '';
      }
   }
   if (whenVisible !== undefined)
      eq.Dom.onNextFrame(w, whenVisible);
};

eq.c.Dlg.whenVisible = function() {
   // Adjust Window list stack order if necessary.
   eq.c.Window.zOrderModified();
   // Restore focus.
   var app = eq.app, d, f, el, ed;
   if (   (d = app.currDlg)
       && (f = app.focus) !== null
       && (el = document.getElementById(f))
       && el.classList.contains('eq-focused')
       && (ed = d.eds.get(f)) !== undefined) {
      eq.Dom.onNextCycle(eq.c.App.setElementFocus, ed.focusElement(d, el));
   }
};

/*------------------------------------------------------------------------
   Dlg.addControl()
   Add DLG control to document from DOM element template(s).
   See: api-dlg.js Dlg.prototype.addChild()
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.addControl = function(et, elParent, rootId) {
   var root;
   if (rootId === undefined && !(rootId = et.id))
      throw new Error("addControl: root id not set");

   while (true) {
      // Create element.
      var
         tg = et.tg,
         el = document.createElement(tg !== undefined ? tg : 'DIV');
      this.updateControl(rootId, el, et);
      if (root === undefined)
         root = el;

      // Add to DOM.
      if (elParent !== undefined)
         elParent.appendChild(el);
      elParent = el;

      // Nested template.
      if (et.et === undefined)
         break;
      et = et.et;
   }

   // Children.
   var ch = et.ch;
   if (ch !== undefined) {
      for (var i = 0, l = ch.length; i < l; i++) {
         var ct = ch[i], id = ct.id;
         if (id === '*')
            ct.id = (id = rootId) + '-' + (i+1);
         this.addControl(ct, elParent, id || rootId);
      }
   }

   return root;
};

/*------------------------------------------------------------------------
   Dlg.modifyControl()
   Modify DLG control from DOM element template(s).
   See: api-dlg.js Dlg.prototype.addChild()
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.modifyControl = function(rootId, el, et) {
   while (true) {
      var tg = et.tg;
      if (tg !== undefined && tg !== el.tagName) {
         // Tag modified, recreate element, retain 'eq-event' if set.
         var en = document.createElement(tg), cl = el.classList;
         if (cl && cl.contains('eq-event'))
            en.classList.add('eq-event');
         el.parentNode.replaceChild(en, el);
         el = en;
      }

      // Update element.
      this.updateControl(rootId, el, et);

      // Nested template?
      if (et.et === undefined)
         break;
      // Yes, use first child which is not 'runtime-created'.
      var cn = el.children;
      el = undefined;
      et = et.et;
      for (var i = 0, l = cn.length; i < l; i++) {
         var ci = cn[i];
         if (!ci.classList.contains('eq-rt')) {
            el = ci;
            break;
         }
      }
      if (el === undefined)
         throw new Error("modifyControl: id '" + rootId +
            "' nested template, no child");
   }

   // Children.
   var ch = et.ch;
   if (ch !== undefined) {
      var cn = el.children;
      for (var i = 0, l = ch.length; i < l; i++) {
         var ci = ch[i];
         if (i >= cn.length) {
            var tg = ci.tg;
            el.appendChild(document.createElement(
               tg !== undefined ? tg : 'DIV'));
         }
         this.modifyControl(rootId, cn[i], ci);
      }
      while (cn.length > l)
         el.removeChild(cn[l]);
   }
};

/*------------------------------------------------------------------------
   Dlg.updateControl()
   Update DLG control from DOM element template.
   See: api-dlg.js Dlg.prototype.addChild()
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.updateControl = function(rootId, el, et) {
   for (var a in et)
      switch (a) {
         case 'id': {
            // id
            var
               id = el.id,
               nid = et.id,
               eds, ed;
            if (id !== nid) {
               if (id) {
                  if ((ed = (eds = this.eds).get(id)) === undefined)
                     throw new Error("BUG: updateControl id:" + id +
                        " not registered");
                  eds.delete(id);
                  eds.set(nid, ed);
                  ed.updateId(nid, id);
               }
               el.id = nid;
            }
            break;
         }

         case 'dcn': {
            // DLG class name.
            var
               id = el.id,
               eds = this.eds,
               ed, cls;
            if (!id)
               throw new Error("BUG: updateControl '" + et.dcn +
                  "' id undefined");
            if ((ed = eds.get(id)) !== undefined)
               throw new Error("BUG: updateControl id:" + id +
                  "already registered as '" + ed.constructor.name + "'");
            if ((cls = eq.d.get(et.dcn)) === undefined)
               cls = eq.c.DlgElement;
            eds.set(id, new cls());
            break;
         }

         case 'hc':
            // Text or HTML content.
            eq.Dom.setContent(el, et.hc);
            break;

         case 'hca':
            // Text or HTML content w/ optional accelerator.
            eq.Dom.setContent(el, et.hca, true);
            break;

         case 'ec':
            // EditText content.
            el.value = et.ec;
            if (el.tagName === 'TEXTAREA')
               el.textContent = et.ec;
            break;

         case 'at': {
            // Attributes.
            for (var at in et.at) {
               var a = et.at[at];
               if (a !== null)
                  el.setAttribute(at, a);
               else
                  el.removeAttribute(at);
            }
            break;
         }

         case 'clr': {
            // Remove class names.
            var cl = et.clr;
            for (var i = 0, l = cl.length; i < l; i++)
               el.classList.remove(cl[i]);
            break;
         }

         case 'cla': {
            // Add class names.
            var cl = et.cla;
            for (var i = 0, l = cl.length; i < l; i++)
               el.classList.add(cl[i]);
            break;
         }

         case 'st': {
            // Styles.
            for (var st in et.st) {
               var s = et.st[st];
               el.style.setProperty(st, s !== null ? s : '');
            }
            break;
         }

         case 'ds': {
            // Data attributes.
            for (var ds in et.ds) {
               var da = et.ds[ds];
               if (da !== null)
                  el.dataset[ds] = da;
               else
                  delete el.dataset[ds];
            }
            break;
         }

         case 'rt': {
            // Value by request tag.
            var ed = this.eds.get(rootId);
            if (ed === undefined)
               throw new Error("BUG: updateControl id:" + rootId +
                  " not registered");
            ed.setAttr(this, rootId, el, et.rt);
            break;
         }

         case 'cr': {
            // CSS rules.
            var ed = this.eds.get(rootId), cr = et.cr, cri;
            if (ed === undefined)
               throw new Error("BUG: updateControl id:" + rootId +
                  " not registered");
            for (var i = 0, l = cr.length; i < l; i++) {
               cri = cr[i];
               ed.setCssRule(rootId, cri.ty, cri.rl);
            }
            break;
         }

         case 'ev': {
            // Events.
            var
               ed = this.eds.get(rootId),
               ev = et.ev;
            if (ed === undefined)
               throw new Error("BUG: updateControl id:" + rootId +
                  " not registered");
            if (ev.a !== undefined)
               ed.a = ev.a !== null ? ev.a : undefined;
            if (ev.c !== undefined)
               ed.c = ev.c !== null ? ev.c : undefined;
            if (ev.m !== undefined)
               ed.m = ev.m !== null ? ev.m : undefined;
            if (ev.f !== undefined)
               ed.f = ev.f !== null ? ev.f : undefined;
            if (ed.a !== undefined || ed.c !== undefined)
               el.classList.add('eq-event');
            else
               el.classList.remove('eq-event');
            break;
         }

         case 'dlg': {
            var d = et.dlg;
            if (d.l !== undefined)
               this.layout = d.l ? d.l : undefined;
            if (d.t !== undefined)
               this.title = d.t ? d.t : undefined;
            if (d.f !== undefined)
               this.fkey = d.f ? d.f : undefined;
            if (d.tm !== undefined)
               this.timer = d.tm ? d.tm : undefined;
            if (d.d !== undefined) {
               if (d.d)
                  this.ddo = d.d;
               else if (this.ddo !== undefined) {
                  this.ddo = undefined;
                  this.resetCurrDlg();
               }
            }
            break;
         }
      }
};

/*------------------------------------------------------------------------
   Dlg.nextTab()
   Locate next/previous element in tab order.
------------------------------------------------------------------------*/

eq.c.Dlg.prototype.nextTab = function(id, toPrevious) {
   var tb = this.tabOrder;
   if (tb !== null) {
      var i = tb.indexOf(id);
      if (i !== -1) {
         var
            j = toPrevious ? i - 1 : i + 1,
            l = tb.length,
            el;
         while (true) {
            if (toPrevious) {
               if (j === -1)
                  j = l - 1;
            }
            else {
               if (j === l)
                  j = 0;
            }
            if (j === i) {
               // No other element found in tab order.
               break;
            }

            el = document.getElementById(id = tb[j]);
            if (el && !el.classList.contains('eq-notab')) {
               // Found next/previous element in tab order.
               return el;
            }

            if (toPrevious)
               j--;
            else
               j++;
         }
      }
   }
   return null;
};

/** @preserve $Id: 10-popup.js,v 28.20 2021/11/19 15:53:11 mike Exp $
*//*======================================================================
   final class PopupBox extends Window
   POPUP BOX class.
========================================================================*/

eq.c.Window.subClass('PopupBox', function(pb, parent) {
   var
      cls = this.constructor,
      base = eq.c.Window,
      dm = eq.Dom,
      pp, el, msg, btn, btnDefault;

   if (cls.singleton)
      throw new Error("BUG: POPUP BOX singleton exists");

   pp = document.createElement('DIV'),
   pp.className = 'eq-popup-box';

   // Message.
   msg = pb.msg.split("\n");
   for (var i = 0, l = msg.length; i < l; i++) {
      el = document.createElement('P');
      dm.setContent(el, msg[i]);
      pp.appendChild(el);
   }

   // Buttons.
   btn = document.createElement('DIV');
   btn.className = 'eq-pushbutton';
   btnDefault = (btnDefault = pb.btnDefault) > 0 ? btnDefault - 1 : 0;
   for (var i = 0, l = pb.btn.length; i < l; i++) {
      el = document.createElement('BUTTON');
      el.type = 'button';
      dm.setContent(el, pb.btn[i], true);
      if (i === btnDefault)
         el.classList.add('eq-focused');
      btn.appendChild(el);
   }
   pp.appendChild(btn);

   base.call(this, base.setupFrame(pp), parent);
   this.whenVisible = { cb : cls.setFocus };
   this.setTitle(pb.title ? pb.title : "Eloquence");
   this.setPos(-1, -1);
   this.activate();
   cls.singleton = this;
});

delete eq.c.PopupBox.subClass; // final

eq.c.PopupBox.setFocus = function(w) {
   w.root.querySelector('.eq-pushbutton>.eq-focused').focus();
};

/*------------------------------------------------------------------------
   static PopupBox.singleton
   Global POPUP BOX instance.
------------------------------------------------------------------------*/

eq.c.PopupBox.singleton = null;

/*------------------------------------------------------------------------
   PopupBox.close()
   Close POPUP BOX, submit event.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.close = function(e) {
   var cls = this.constructor, base = eq.c.Window;
   if (this !== cls.singleton)
      throw new Error("BUG: unexpected POPUP BOX invocation");
   if (e !== undefined) {
      var t = e.target;
      if (   base.fromTarget(t) === this
          && this.onMouseDownOrClick(e)) {
         e.stopPropagation();
         return false;
      }

      if (e.type !== 'click')
         return false;

      var i = this.buttonIndex(t);
      if (i === -1) {
         // Not a button.
         return false;
      }

      // Notify API thread.
      eq.app.api.postMessage({
         o : eq.ApiOp.onPopupEvent,
         d : i + 1
      });

      e.stopPropagation();
   }

   base.prototype.close.call(this);
   cls.singleton = null;
   return true;
};

/*------------------------------------------------------------------------
   PopupBox.onKey()
   POPUP BOX keyboard event handler.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.onKey = function(e) {
   var cls = this.constructor, t;
   if (this !== cls.singleton)
      throw new Error("BUG: unexpected POPUP BOX invocation");
   if (   this.isEvent(e)
       || (t = e.target).classList.contains('eq-idle')
       || t === document.body) {
      var i = 0;
      switch (e.key) {
         case 'Enter':
         case ' ':
         case 'Spacebar': // IE, Firefox <= 37
            if (this.buttonIndex(t) !== -1)
               eq.Dom.fireClick(t);
            break;
         case 'Tab':
            i = e.shiftKey ? -1 : 1;
            break;
         case 'ArrowUp':
         case 'Up':
         case 'ArrowLeft':
         case 'Left':
            i = -1;
            break;
         case 'ArrowDown':
         case 'Down':
         case 'ArrowRight':
         case 'Right':
            i = 1;
            break;
         default:
            return false;
      }

      if (i !== 0) {
         var
            btn = this.buttonIndex(t),
            el = this.root.querySelector(
               '.eq-popup-box>.eq-pushbutton').children,
            l = el.length;
         if ((btn += i) < 0)
            btn = l - 1;
         else if (btn >= l)
            btn = 0;
         for (i = 0; i < l; i++)
            el[i].classList.remove('eq-focused');
         el[btn].classList.add('eq-focused');
         el[btn].focus();
      }

      return true; // Consume event.
   }

   return false;
};

/*------------------------------------------------------------------------
   PopupBox.isEvent()
   Check if event applies to POPUP BOX.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.isEvent = function(e) {
   var el = e.target, r = this.root;
   while (el && el !== document && el !== r)
      el = el.parentNode;
   return el === r;
};

/*------------------------------------------------------------------------
   PopupBox.buttonIndex()
   Get POPUP BOX button index of specified target.
------------------------------------------------------------------------*/

eq.c.PopupBox.prototype.buttonIndex = function(t) {
   if (t) {
      var pb;
      while ((pb = t.parentNode) && pb !== document && pb !== this.root) {
         var cl = pb.classList;
         if (cl && cl.contains('eq-pushbutton')) {
            var n = this.root.querySelector(
               '.eq-popup-box>.eq-pushbutton').children;
            for (var i = 0, l = n.length; i < l; i++)
               if (t === n[i])
                  return i;
            break;
         }
         t = pb;
      }
   }
   return -1;
};

/** @preserve $Id: 20-checkbox.js,v 28.3 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class CheckBox extends DlgElement
   DLG CheckBox element class.
========================================================================*/

eq.c.DlgElement.addClass('CheckBox', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.CheckBox.subClass; // final

/*------------------------------------------------------------------------
   CheckBox.onKey()
   Filter and/or handle key event.
------------------------------------------------------------------------*/

eq.c.CheckBox.prototype.onKey = function(dlg, id, el, e) {
   var dm = eq.Dom, kc = dm.isEnterOrSpace(e);
   if (kc) {
      dm.fireClick(dm.inputElement(el), { key : kc });
      return true;
   }
   return false;
};

/*------------------------------------------------------------------------
   CheckBox.onTypeAhead()
   Process type-ahead key stroke.
------------------------------------------------------------------------*/

eq.c.CheckBox.prototype.onTypeAhead = function(dlg, id, el, key) {
   var kc;
   switch (key) {
      case 'Enter':
      case 'ShiftEnter':
         if (el.classList.contains('eq-noenter')) {
            // .tabonenter disabled.
            kc = 10;
         }
         break;
      case ' ':
         kc = 32;
         break;
   }

   if (kc !== undefined) {
      var dm = eq.Dom;
      dm.fireClick(dm.inputElement(el), { key : kc });
      return true;
   }

   return eq.c.DlgElement.prototype.onTypeAhead.call(this, dlg, id, el, key);
};

/*------------------------------------------------------------------------
   CheckBox.focusElement()
   Obtain focus element.
------------------------------------------------------------------------*/

eq.c.CheckBox.prototype.focusElement = function(dlg, el) {
   return eq.Dom.inputElement(el);
};

/*------------------------------------------------------------------------
   CheckBox.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.CheckBox.prototype.changed = function(dlg, id, el, ty) {
   var ch = {
      id : id,
      ty : eq.RsValType.buttonChecked,
      iv : el.checked ? 1 : 0
   };
   return ch;
};

/** @preserve $Id: 20-combobox.js,v 28.6 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class ComboBox extends DlgElement
   DLG ComboBox element class.
========================================================================*/

eq.c.DlgElement.addClass('ComboBox', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.ComboBox.subClass; // final

/*------------------------------------------------------------------------
   static ComboBox.line()
   Obtain ComboBox line from item list.
------------------------------------------------------------------------*/

eq.c.ComboBox.line = function(t, co) {
   var c, cl;
   while (t) {
      if (!(c = t.parentNode))
         break;
      if (cl = c.classList) {
         if (cl.contains('eq-column')) {
            if (co !== undefined)
               co.c = c;
            return t;
         }
         if (cl.contains('eq-combobox')) {
            if (co !== undefined)
               co.ct = c;
            break;
         }
      }
      t = t.parentNode;
   }
   return undefined;
};

/*------------------------------------------------------------------------
   static ComboBox.select()
   Select ComboBox line in item list.
------------------------------------------------------------------------*/

eq.c.ComboBox.select = function(el, c, ci) {
   if (c === undefined)
      if (!(c = el.querySelector('.eq-combobox>.eq-column')))
         c = document.body.querySelector('body>.eq-combobox.eq-column');
   // Deselect.
   var n = c.querySelectorAll('.eq-column>.eq-selected');
   for (var i = 0, l = n.length; i < l; i++)
      n[i].classList.remove('eq-selected');
   // Select.
   if (ci && c === ci.parentNode) {
      ci.classList.add('eq-selected');
      if ((n = c.children).length > 1) {
         // Scroll into view if necessary.
         var
            bh = c.getBoundingClientRect().height,
            lh = n[0].getBoundingClientRect().height;
         if (bh > 0 && lh > 0) {
            var bt = c.scrollTop, lt = ci.offsetTop;
            if (lt < bt)
               c.scrollTop = lt;
            else if (lt + lh > bt + bh)
               c.scrollTop = Math.ceil(lt + lh - bh + 2);
         }
      }
   }
};

/*------------------------------------------------------------------------
   static ComboBox.open()
   Open ComboBox item list.
------------------------------------------------------------------------*/

eq.c.ComboBox.open = function(el, c) {
   var cl = el.classList;
   if (!cl.contains('eq-open')) {
      if (c === undefined)
         c = el.querySelector('.eq-combobox>.eq-column');
      var
         cls = eq.c.ComboBox,
         v = el.firstElementChild.value,
         rc = el.getBoundingClientRect(),
         s = c.style,
         ccl = c.classList,
         ci, w, cs;
      if (v.length) {
         // Pre-select current entry.
         var n = c.children;
         for (var i = 0, l = n.length; i < l; i++)
            if (v === n[i].textContent) {
               ci = n[i];
               break;
            }
      }
      // Set overlay position and width.
      s.left = rc.left + window.pageXOffset + 'px';
      s.top = rc.top + rc.height + window.pageYOffset + 'px';
      if (w = s.maxWidth)
         s.width = w;
      else
         s.width = rc.width + 'px';
      // Set colors.
      cs = window.getComputedStyle(el);
      s.color = cs.color;
      s.backgroundColor = cs.backgroundColor;
      // Make overlay.
      ccl.add('eq-combobox');
      if (cl.contains('eq-border'))
         ccl.add('eq-border');
      else
         ccl.remove('eq-border');
      eq.app.makeOverlay(el, c, cls.close);
      cl.add('eq-open');
      cl.add('eq-enter');
      cls.select(el, c, ci);
   }
};

/*------------------------------------------------------------------------
   static ComboBox.close()
   Close ComboBox item list.
------------------------------------------------------------------------*/

eq.c.ComboBox.close = function(el, c, commit) {
   var cl = el.classList, ci;
   if (!cl.contains('eq-open'))
      throw new Error("ComboBox.close BUG: not open, id:" + el.id);
   el.appendChild(c);
   c.classList.remove('eq-combobox');
   cl.remove('eq-open');
   cl.remove('eq-enter');
   if (commit && (ci = c.querySelector('.eq-column>.eq-selected'))) {
      var ct = el.firstElementChild;
      ct.value = ci.textContent;
      ct.select();
      eq.Dom.fireEvent(el, 'input');
   }
};

/*------------------------------------------------------------------------
   static ComboBox.move()
   Handle ComboBox move capture.
------------------------------------------------------------------------*/

eq.c.ComboBox.move = function(e, m) {
   var c = m.c;
   if (c) {
      var
         cls = eq.c.ComboBox,
         t = e.target,
         el = m.el,
         co = {},
         ln = cls.line(t, co);
      if (el !== co.ct) {
         cls.select(el, c, ln);
         if (ln)
            m.close = true;
      }
   }
};

eq.c.ComboBox.move.done = function(m) {
   if (m.close)
      eq.app.closeOverlay(m.el, true);
   var a = m.a;
   if (a !== undefined)
      a.classList.remove('eq-click');
   return true;
};

/*------------------------------------------------------------------------
   ComboBox.setAttr()
   Set ComboBox attributes.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = eq.c.ComboBox, dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.items: {
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;
         }

         case tag.boxSize: {
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               w : rc.w,
               h : rc.h
            });
            break;
         }

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

eq.c.ComboBox.setAttr = function(rt) {
   var tag = eq.RqTag;
   switch (rt.t) {
      case tag.items: {
         var
            dm = eq.Dom,
            v = rt.v,
            l = v.length,
            el = rt.el,
            c, n, nl, ln;
         if (!l)
            el.classList.add('eq-empty');
         if (!(c = el.querySelector('.eq-combobox>.eq-button'))) {
            c = document.createElement('DIV');
            c.className = 'eq-rt eq-button';
            el.appendChild(c);
         }
         if (!(c = el.querySelector('.eq-combobox>.eq-column'))) {
            c = document.createElement('DIV');
            c.className = 'eq-rt eq-column';
            el.appendChild(c);
         }
         n = c.children;
         nl = n.length;
         for (var i = 0; i < l; i++) {
            if (i < nl)
               ln = n[i];
            else {
               ln = document.createElement('DIV');
               c.appendChild(ln);
            }
            dm.setContent(ln, v[i]);
         }
         while (n.length > l)
            c.removeChild(c.lastElementChild);
         break;
      }

      case tag.boxSize: {
         var
            c = rt.el.querySelector('.eq-combobox>.eq-column'),
            w = rt.w,
            h = rt.h,
            s = c.style;
         s.maxWidth = w ? w : '';
         s.maxHeight = h ? h : '';
         break;
      }

      default:
         throw new Error("ComboBox.setAttr: invalid tag " + rt.t);
   }
};

/*------------------------------------------------------------------------
   ComboBox.onKey()
   Filter and/or handle key event.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.onKey = function(dlg, id, el, e) {
   var
      cls = eq.c.ComboBox,
      c = el.querySelector('.eq-combobox>.eq-column'),
      d;
   switch (e.key) {
      case 'Enter':
         if (!c)
            eq.app.closeOverlay(el, true);
         return true;
      case 'Escape':
      case 'Esc': // IE
         if (!c)
            eq.app.closeOverlay(el, false);
         return true;

      // Navigation?
      case 'PageUp':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowDown':
      case 'Down':
         if (c) {
            cls.open(el, c);
            d = 0;
         }
         break;
      default:
         // No.
         return false;
   }

   var n, ci;
   c = document.body.querySelector('body>.eq-combobox.eq-column');
   if (c && (n = c.children).length) {
      if (d === undefined)
         switch (e.key) {
            case 'PageUp':
               ci = n[0];
               break;
            case 'PageDown':
               ci = n[n.length - 1];
               break;
            case 'ArrowUp':
            case 'Up':
               d = -1;
               break;
            case 'ArrowDown':
            case 'Down':
               d = 1;
               break;
         }
      if (d !== undefined) {
         if (ci = c.querySelector('.eq-column>.eq-selected')) {
            var i = eq.Dom.parentIndex(ci), l = n.length;
            if ((i += d) === -1)
               i = 0;
            else if (i >= l)
               i = l - 1;
            ci = n[i];
         }
         else
            ci = n[0];
      }
      if (ci !== undefined)
         cls.select(el, c, ci);
   }

   return true;
};

/*------------------------------------------------------------------------
   ComboBox.onTypeAhead()
   Process type-ahead key stroke.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.onTypeAhead = function(dlg, id, el, key) {
   if (key.length === 1) {
      var
         ct = eq.Dom.inputElement(el),
         v = ct.value,
         s0 = ct.selectionStart,
         s1 = ct.selectionEnd;
      if (s0 > 0) {
         if (s1 < v.length)
            v = v.substring(0, s0) + key + v.substring(s1);
         else
            v = v.substring(0, s0) + key;
      }
      else if (s1 < v.length)
         v = key + v.substring(s1);
      else
         v = key;
      ct.value = v;
      return true;
   }

   return eq.c.DlgElement.prototype.onTypeAhead.call(this, dlg, id, el, key);
};

/*------------------------------------------------------------------------
   ComboBox.onChange()
   Filter and/or handle change event.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.onChange = function(dlg, id, el, e) {
   if (e.type === 'mousedown') {
      var cls = eq.c.ComboBox, app = eq.app, t = e.target;
      if (t && t.classList.contains('eq-button')) {
         if (app.setFocus(id, el)) {
            app.pendingFocus = {
               id : id,
               act : cls.open
            };
            return false;
         }
         var
            cl = el.classList,
            c = el.querySelector('.eq-combobox>.eq-column'),
            m = { el : el, a : t };
         if (cl.contains('eq-disabled') || cl.contains('eq-empty'))
            return false;
         t.classList.add('eq-click');
         if (c)
            cls.open(el, m.c = c);
         else
            m.close = true;
         app.moveCapture(m, cls.move, cls.move.done);
         eq.Dom.consumeEvent(e);
         return false;
      }
      if (el.classList.contains('eq-open')) {
         var
            m = { el : el },
            co = {},
            ln = cls.line(t, co);
         if (co.c) {
            cls.select(el, m.c = co.c, ln);
            m.close = true;
         }
         app.moveCapture(m, cls.move, cls.move.done);
         if (el !== co.ct) {
            eq.Dom.consumeEvent(e);
            return false;
         }
      }
   }
   return true;
};

/*------------------------------------------------------------------------
   ComboBox.focusElement()
   Obtain focus element.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.focusElement = function(dlg, el) {
   return eq.Dom.inputElement(el);
};

/*------------------------------------------------------------------------
   ComboBox.focusGained()
   ComboBox has gained focus.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.focusGained = function(dlg, ct) {
   // Select on focus except on mouse click or if focus doesn't change.
   if (eq.app.pendingMouseDown === undefined && ct !== document.activeElement)
      ct.select();
   var el = ct.parentNode, cl = el.classList;
   if (cl.contains('eq-auto') && !cl.contains('eq-disabled'))
      eq.Dom.onNextCycle(eq.c.ComboBox.open, el);
};

/*------------------------------------------------------------------------
   ComboBox.focusLost()
   ComboBox has lost focus.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.focusLost = function(dlg, el) {
   if (el.classList.contains('eq-open'))
      eq.app.closeOverlay(el);
   var ct = eq.Dom.inputElement(el);
   ct.selectionStart = ct.selectionEnd = 0;
   ct.blur();
};

/*------------------------------------------------------------------------
   ComboBox.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.ComboBox.prototype.changed = function(dlg, id, el, ty) {
   for (var i = 0, l = ty.length; i < l; i++)
      if (ty[i] === 'input') {
         var ch = {
            id : id,
            ty : eq.RsValType.textValue,
            sv : el.value
         };
         return ch;
      }
   return null;
};

/** @preserve $Id: 20-edittext.js,v 28.3 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class EditText extends DlgElement
   DLG EditText element class.
========================================================================*/

eq.c.DlgElement.addClass('EditText', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.EditText.subClass; // final

/*------------------------------------------------------------------------
   EditText.focusElement()
   Obtain focus element.
------------------------------------------------------------------------*/

eq.c.EditText.prototype.focusElement = function(dlg, el) {
   var n = el.getElementsByTagName('INPUT');
   if (n.length === 0)
      n = el.getElementsByTagName('TEXTAREA');
   if (n.length !== 1)
      throw new Error("BUG: EditText.focusElement id:" + el.id +
         " elements:" + n.length + ", expected: 1");
   return n[0];
};

/*------------------------------------------------------------------------
   EditText.focusGained()
   EditText has gained focus.
------------------------------------------------------------------------*/

eq.c.EditText.prototype.focusGained = function(dlg, ct) {
   // Select on focus except on mouse click or if focus doesn't change.
   if (   eq.app.pendingMouseDown === undefined
       && ct !== document.activeElement) {
      if (ct.tagName === 'INPUT')
         ct.select();
      else
         ct.selectionStart = ct.selectionEnd = 0;
   }
};

/*------------------------------------------------------------------------
   EditText.focusLost()
   EditText has lost focus.
------------------------------------------------------------------------*/

eq.c.EditText.prototype.focusLost = function(dlg, el) {
   var ct = el.firstElementChild;
   if (ct.tagName === 'INPUT')
      ct.selectionStart = ct.selectionEnd = 0;
   ct.blur();
};

/*------------------------------------------------------------------------
   EditText.onTypeAhead()
   Process type-ahead key stroke.
------------------------------------------------------------------------*/

eq.c.EditText.prototype.onTypeAhead = function(dlg, id, el, key) {
   var ct = el.firstElementChild, ch;
   switch (key) {
      case 'Enter':
         if (ct.tagName === 'TEXTAREA')
            ch = '\n';
         break;
      default:
         if (key.length === 1)
            ch = key;
   }

   if (ch !== undefined) {
      var
         v = ct.value,
         s0 = ct.selectionStart,
         s1 = ct.selectionEnd;
      if (s0 > 0) {
         if (s1 < v.length)
            v = v.substring(0, s0) + ch + v.substring(s1);
         else
            v = v.substring(0, s0) + ch;
      }
      else if (s1 < v.length)
         v = ch + v.substring(s1);
      else
         v = ch;
      ct.value = v;
      if (ct.tagName === 'TEXTAREA')
         ct.textContent = v;
      return true;
   }

   return eq.c.DlgElement.prototype.onTypeAhead.call(this, dlg, id, el, key);
};

/*------------------------------------------------------------------------
   EditText.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.EditText.prototype.changed = function(dlg, id, el, ty) {
   var v = [];
   for (var i = 0, l = ty.length; i < l; i++)
      switch (ty[i]) {
         case 'input':
            v.push({
               ty : eq.RsValType.textValue,
               sv : el.value
            });
            break;

         case 'blur': {
            var iv = el.selectionStart;
            if (iv >= 0)
               v.push({
                  ty : eq.RsValType.caretPos,
                  iv : iv
               });
            break;
         }
      }
   switch (v.length) {
      case 0:
         return null;
      case 1:
         v[0].id = id;
         return v[0];
      default:
         return {
            id : id,
            v : v
         };
   }
};

/** @preserve $Id: 20-groupbox.js,v 28.3 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class GroupBox extends DlgElement
   DLG GroupBox element class.
========================================================================*/

eq.c.DlgElement.addClass('GroupBox', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.GroupBox.subClass; // final

/*------------------------------------------------------------------------
   static GroupBox.onScroll()
   Handle GroupBox scroll event.
------------------------------------------------------------------------*/

eq.c.GroupBox.onScroll = function(e) {
   var el = e.target, cl = el.classList;
   if (cl && cl.contains('eq-groupbox')) {
      // 'scroll' change event.
      eq.app.setChanged(el.id, 'scroll');
   }
};

/*------------------------------------------------------------------------
   GroupBox.setAttr()
   Set GroupBox attributes.
------------------------------------------------------------------------*/

eq.c.GroupBox.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = eq.c.GroupBox, dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.vx:
         case tag.vy: {
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               i : rc.i
            });
            break;
         }

         case tag.vWidth:
         case tag.vHeight:
            el.onscroll = rc.s ? cls.onScroll : null;
            break;

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

eq.c.GroupBox.setAttr = function(rt) {
   var tag = eq.RqTag;
   switch (rt.t) {
      case tag.vx: {
         rt.el.scrollLeft = rt.i;
         break;
      }

      case tag.vy: {
         rt.el.scrollTop = rt.i;
         break;
      }

      default:
         throw new Error("GroupBox.setAttr: invalid tag " + rt.t);
   }
};

/*------------------------------------------------------------------------
   GroupBox.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.GroupBox.prototype.changed = function(dlg, id, el, ty) {
   var ch = {
      id : id,
      ty : eq.RsValType.scrollPos,
      x  : el.scrollLeft,
      y  : el.scrollTop
   };
   return ch;
};

/** @preserve $Id: 20-htmlview.js,v 28.4 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class HtmlView extends DlgElement
   DLG HtmlView element class.
========================================================================*/

eq.c.DlgElement.addClass('HtmlView', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.HtmlView.subClass; // final

eq.c.HtmlView.re = new RegExp('^eloq:([0-9]*)(/(.*))?$', 'i');
eq.c.HtmlView.url = new Map();
eq.c.HtmlView.content = new Map();
eq.c.HtmlView.timeout = undefined;

/*------------------------------------------------------------------------
   static HtmlView.ready()
------------------------------------------------------------------------*/

eq.c.HtmlView.ready = function(e) {
   var
      cls = eq.c.HtmlView,
      id = this.name,
      url = cls.url.get(id),
      cn = cls.content.get(id);
   if (url || cn)
      cls.startTimeout();
   try {
      var
         w = this.contentWindow,
         d = w.document,
         h = d.head,
         b = d.body,
         n;
      w.name = id;
      if (cn) {
         b.style.margin = "0";
         eq.Dom.setHtmlContent(cn, b, h);
      }
      n = b.querySelectorAll('a[href^="eloq:"]');
      for (var i = 0, l = n.length; i < l; i++) {
         var a = n[i];
         a.dataset.eqLink = a.href;
         a.href = '#';
         a.addEventListener('click', cls.clicked);
      }
   }
   catch (e) {
      console.log("HtmlView:" + id + (url ? " '" + url + "' " : " ") +
         e.name + ": " + e.message);
   }
};

/*------------------------------------------------------------------------
   static HtmlView.clicked()
------------------------------------------------------------------------*/

eq.c.HtmlView.clicked = function(e) {
   var
      cls = eq.c.HtmlView,
      id = e.view.name,
      el, ln;
   if (id && (el = document.getElementById(id))
          && (ln = this.dataset.eqLink)) {
      var
         app = eq.app,
         r = cls.re.exec(ln),
         rule;
      if (r) {
         if (r.length > 0 && r[1].length)
            rule = r[1];
         if (r.length > 2 && r[3])
            el.dataset.eqLink = r[3];
      }
      app.setChanged(id, 'change');
      if (rule !== undefined || el.classList.contains('eq-rule')) {
         if (rule !== undefined) {
            app.onceRule = rule;
            app.onceId = id;
         }
         app.submit(el, {
            id : id,
            type : 'change'
         });
      }
   }
   e.preventDefault();
   e.stopPropagation();
};

/*------------------------------------------------------------------------
   static HtmlView.startTimeout()
------------------------------------------------------------------------*/

eq.c.HtmlView.startTimeout = function() {
   if (this.timeout !== undefined)
      window.clearTimeout(this.timeout);
   this.timeout = window.setTimeout(this.timeoutElapsed, 1000);
};

eq.c.HtmlView.timeoutElapsed = function() {
   var cls = eq.c.HtmlView;
   cls.timeout = undefined;
   cls.url.clear();
   cls.content.clear();
};

/*------------------------------------------------------------------------
   HtmlView.setAttr()
   Set HtmlView attributes.
------------------------------------------------------------------------*/

eq.c.HtmlView.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = eq.c.HtmlView, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.url:
            cls.url.set(id, rc.v);
            cls.content.delete(id);
            cls.startTimeout();

            var ch = el.firstElementChild;
            if (ch && ch.tagName !== 'IFRAME') {
               el.removeChild(ch);
               ch = null;
            }
            if (!ch) {
               ch = document.createElement('IFRAME');
               ch.name = id;
               ch.onload = cls.ready;
               el.appendChild(ch);
            }
            ch.src = rc.v;
            break;

         case tag.content: {
            var ch = el.firstElementChild;
            switch (rc.ty) {
               case 0: // text/html
                  cls.url.delete(id);
                  cls.content.set(id, rc.v);
                  cls.startTimeout();

                  if (ch && ch.tagName !== 'IFRAME') {
                     el.removeChild(ch);
                     ch = null;
                  }
                  if (!ch) {
                     ch = document.createElement('IFRAME');
                     ch.name = id;
                     ch.onload = cls.ready;
                     el.appendChild(ch);
                  }
                  ch.src = 'about:blank';
                  break;

               case 1: // text/plain
                  cls.url.delete(id);
                  cls.content.delete(id);
                  cls.startTimeout();

                  if (ch && ch.tagName !== 'TEXTAREA') {
                     el.removeChild(ch);
                     ch = null;
                  }
                  if (!ch) {
                     ch = document.createElement('TEXTAREA');
                     ch.disabled = true;
                     ch.wrap = 'off'; // IE
                     el.appendChild(ch);
                  }
                  ch.value = ch.textContent = rc.v;
                  break;

               default:
                  throw new Error("HtmlView.setAttr: tag " + rc.t +
                     " invalid type " + rc.ty);
            }
            break;
         }

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

/*------------------------------------------------------------------------
   HtmlView.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.HtmlView.prototype.changed = function(dlg, id, el, ty) {
   var ln = el.dataset.eqLink;
   if (ln)
      delete el.dataset.eqLink;
   var ch = {
      id : id,
      ty : eq.RsValType.textValue,
      sv : ln ? ln : ""
   };
   return ch;
};

/** @preserve $Id: 20-listbox.js,v 28.9 2021/11/19 15:51:12 mike Exp $
*//*======================================================================
   final class ListBox extends DlgElement
   DLG ListBox element class.
========================================================================*/

eq.c.DlgElement.addClass('ListBox', function() {
   eq.c.DlgElement.call(this);
   this.activeCol = 0;
   this.lineOrder = null;
});

delete eq.c.ListBox.subClass; // final

/*------------------------------------------------------------------------
   static ListBox.select()
   Select ListBox line.
------------------------------------------------------------------------*/

eq.c.ListBox.select = function(el, li) {
   // Deselect.
   var n = el.querySelectorAll('.eq-column>.eq-selected');
   for (var i = 0, l = n.length; i < l; i++)
      n[i].classList.remove('eq-selected');
   // Select.
   if (li !== undefined) {
      n = el.querySelectorAll('.eq-column');
      for (var i = 0, l = n.length; i < l; i++) {
         var c = n[i].children;
         if (li < c.length)
            c[li].classList.add('eq-selected');
      }
   }
};

/*------------------------------------------------------------------------
   static ListBox.scrollIntoView()
   Scroll ListBox line into view if necessary.
------------------------------------------------------------------------*/

eq.c.ListBox.scrollIntoView = function(el, li) {
   var
      n = el.querySelector('.eq-column').children,
      bx, bh, lh;
   if (li < n.length) {
      bx = n[0].parentNode.parentNode;
      bh = bx.getBoundingClientRect().height;
      lh = n[0].getBoundingClientRect().height;
      if (bh > 0 && lh > 0) {
         var bt = bx.scrollTop, lt = n[li].offsetTop;
         if (lt < bt)
            bx.scrollTop = lt;
         else if (lt + lh > bt + bh)
            bx.scrollTop = Math.ceil(lt + lh - bh + 2);
      }
   }
};

/*------------------------------------------------------------------------
   static ListBox.onScroll()
   Handle ListBox scroll event.
------------------------------------------------------------------------*/

eq.c.ListBox.onScroll = function(e) {
   if (eq.app.currDlg)
      eq.Dom.onNextFrame(e.target, { cb : eq.c.ListBox.onScroll.exec });
};

eq.c.ListBox.onScroll.exec = function(el) {
   var cl = el.classList, lb;
   if (cl && cl.contains('eq-pane') && (lb = el.parentNode)
                                    && (cl = lb.classList)) {
      if (cl.contains('eq-listbox')) {
         // 'scroll' change event.
         eq.app.setChanged(lb.id, 'scroll');
         if (cl.contains('eq-multi')) {
            // Scroll multi-column ListBox header.
            var h = el.previousElementSibling;
            if (h)
               h.style.marginLeft = '-' + el.scrollLeft + 'px';
         }
      }
   }
};

/*------------------------------------------------------------------------
   static ListBox.setAutoWidth()
   Set ListBox auto-width column widths.
------------------------------------------------------------------------*/

eq.c.ListBox.setAutoWidth = function(el) {
   var
      hn = el.firstElementChild.children,
      pn = el.lastElementChild.children,
      wmin = 2 * eq.app.grid.wr;
   for (var i = 0, l = pn.length; i < l; i++) {
      var c, h, cw, hw;
      if ((c = pn[i]).classList.contains('eq-auto')) {
         h = hn[i];
         cw = Number(window.getComputedStyle(c).width.replace(/px$/,''));
         hw = Number(window.getComputedStyle(h).width.replace(/px$/,''));
         if (hw > cw)
            cw = hw;
         if (cw < wmin)
            cw = wmin;
         h.style.width = c.style.width = cw + 'px';
      }
   }
};

/*------------------------------------------------------------------------
   static ListBox.colResize()
   Handle ListBox column-resize move capture.
------------------------------------------------------------------------*/

eq.c.ListBox.colResize = function(e, m) {
   if (!m.busy) {
      m.busy = true;
      eq.Dom.onNextFrame(m.el, {
         cb : eq.c.ListBox.colResize.exec,
         x : e.clientX,
         m : m
      });
   }
};

eq.c.ListBox.colResize.exec = function(el, qi) {
   var
      x = qi.x,
      m = qi.m;
   if (x !== m.x) {
      var
         h = m.h,
         c = m.c,
         w = m.w + x - m.x,
         wmin = 2 * eq.app.grid.wr;
      if (w >= wmin) {
         h.style.width = c.style.width = w + 'px';
         m.x = x;
         m.w = w;
      }
      else {
         h.style.width = c.style.width = wmin + 'px';
         m.x -= m.w - wmin;
         m.w = wmin;
      }
      m.modified = true;
   }
   m.busy = undefined;
};

eq.c.ListBox.colResize.done = function(m) {
   if (m.modified) {
      var app = eq.app, id = m.el.id;
      app.setChanged(id, 'resize');
      app.apiUpdate(id, eq.UpdateTag.colResized, {
         col : eq.Dom.parentIndex(m.h),
         width : m.w
      });
   }
   return false;
};

/*------------------------------------------------------------------------
   static ListBox.colMove()
   Handle ListBox column move capture.
------------------------------------------------------------------------*/

eq.c.ListBox.colMove = function(e, m) {
   if (!m.busy) {
      m.busy = true;
      eq.Dom.onNextFrame(m.el, {
         cb : eq.c.ListBox.colMove.exec,
         x : e.clientX,
         m : m
      });
   }
};

eq.c.ListBox.colMove.exec = function(el, qi) {
   var
      m = qi.m,
      n = m.el.getElementsByClassName('eq-space'),
      cx = qi.x,
      cw = m.cw,
      mx = m.x,
      lx = m.lx,
      from = m.f,
      to = m.t,
      x, ox, ws, i, j, l, w, nj, p;
   if (n.length !== 2) {
      m.busy = undefined;
      return;
   }
   if ((x = lx + cx - mx) < 0) {
      x = 0;
      cx = mx - lx;
   }
   if (x > lx)
      ox = m.ow;
   else if (x < lx)
      ox = 0;
   else {
      m.busy = undefined;
      return;
   }
   for (ox += x, ws = 0, i = 0, l = cw.length; i < l; i++) {
      if (i === to)
         w = cw[from];
      else if (i >= from && i < to)
         w = cw[i + 1];
      else
         w = cw[i];
      ws += w;
      if (ox < ws - w / 2)
         break;
   }
   for (j = 0; j < 2; j++) {
      p = (nj = n[j]).parentNode;
      if (i < l)
         p.insertBefore(nj, p.children[i]);
      else
         p.appendChild(nj);
   }
   m.t = eq.Dom.parentIndex(nj);
   m.ob.style.left = (m.lx = x) + 'px';
   m.x = cx;
   m.busy = undefined;
};

eq.c.ListBox.colMove.done = function(m) {
   var
      app = eq.app,
      el = m.el,
      from = m.f,
      to = m.t,
      id;
   app.closeOverlay(el);
   if (to !== from) {
      id = el.id;
      app.setChanged(id, 'move');
      app.apiUpdate(id, eq.UpdateTag.colMoved, { from : from, to : to });
   }
   return false;
};

eq.c.ListBox.colMove.closeOverlay = function(el, ov) {
   var
      n = el.getElementsByClassName('eq-space'),
      ob = ov.firstElementChild,
      id;
   if (n.length !== 2)
      throw new Error("ListBox.colMove.closeOverlay BUG: " + n.length +
         " eq-space elements");
   var c = n[0];
   c.parentNode.replaceChild(ob.firstElementChild.firstElementChild, c);
   var c = n[0];
   c.parentNode.replaceChild(ob.lastElementChild.firstElementChild, c);
   if (id = ob.id)
      eq.app.removeStyles(id);
   ov.parentNode.removeChild(ov);
};


/*------------------------------------------------------------------------
   ListBox.setAttr()
   Set ListBox attributes.
------------------------------------------------------------------------*/

eq.c.ListBox.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = this.constructor, dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.columns: {
            this.lineOrder = null;
            var cl = el.classList;
            if (rc.l === 0) {
               // Single-column ListBox.
               if (cl.contains('eq-single'))
                  // Delete children.
                  el.firstElementChild.firstElementChild.textContent = '';
               else {
                  el.textContent = ''; // Delete children.
                  cl.remove('eq-multi');
                  cl.add('eq-single');

                  var p = document.createElement('DIV');
                  p.className = 'eq-pane';
                  p.onscroll = cls.onScroll;

                  var c = document.createElement('DIV');
                  c.className = 'eq-column';

                  p.appendChild(c);
                  el.appendChild(p);
               }
            }
            else {
               // Multi-column ListBox.
               if (cl.contains('eq-multi')) {
                  // Delete children.
                  el.firstElementChild.textContent = '';
                  el.lastElementChild.textContent = '';
               }
               else {
                  el.textContent = ''; // Delete children.
                  cl.remove('eq-single');
                  cl.add('eq-multi');

                  var h = document.createElement('DIV');
                  h.className = 'eq-caption';

                  var p = document.createElement('DIV');
                  p.className = 'eq-pane';
                  p.onscroll = cls.onScroll;

                  el.appendChild(h);
                  el.appendChild(p);
               }
            }
            break;
         }

         case tag.colOrder: {
            var
               colMove = rc.m,
               l = colMove.length,
               h = el.firstElementChild,
               hn = h.children,
               p = el.lastElementChild,
               pn = p.children,
               to, from;
            for (to = 0; to < l; to++)
               if ((from = colMove[to]) !== to) {
                  h.insertBefore(hn[from], hn[to]);
                  p.insertBefore(pn[from], pn[to]);
               }
            break;
         }

         case tag.colState: {
            var
               colShow = rc.s,
               l = colShow.length,
               h = el.firstElementChild,
               hn = h.children,
               p = el.lastElementChild,
               pn = p.children,
               ct = eq.ColType,
               i, hi, pi, cs, c, sp, cl, st, autoWidth;
            for (i = hi = pi = 0; i < l; i++) {
               if ((cs = colShow[i]) === null) {
                  if (hi < hn.length)
                     h.removeChild(hn[hi]);
                  if (pi < pn.length)
                     p.removeChild(pn[pi]);
                  continue;
               }
               // Header.
               if (hi < hn.length)
                  sp = (c = hn[hi]).firstElementChild;
               else {
                  c = document.createElement('DIV');
                  c.appendChild(sp = document.createElement('SPAN'));
                  h.appendChild(c);
               }
               hi++;
               if (cs.w !== null)
                  c.style.width = cs.w;
               else {
                  c.style.width = '';
                  if (!autoWidth) {
                     autoWidth = true;
                     dm.onNextFrame(el, { cb : cls.setAutoWidth });
                  }
               }
               st = sp.style;
               if (cs.bc !== null)
                  st.backgroundColor = cs.bc;
               if (cs.fc !== null)
                  st.color = cs.fc;
               cl = c.classList;
               if (cs.movable)
                  cl.add('eq-move');
               else
                  cl.remove('eq-move');
               if (cs.sizable)
                  cl.add('eq-resize');
               else
                  cl.remove('eq-resize');
               if (cs.sortable)
                  cl.add('eq-sort');
               else
                  cl.remove('eq-sort');
               // Column.
               if (pi < pn.length)
                  c = pn[pi];
               else {
                  c = document.createElement('DIV');
                  c.className = 'eq-column';
                  p.appendChild(c);
               }
               pi++;
               cl = c.classList;
               switch (cs.type) {
                  case ct.colBoolean:
                     cl.add('eq-bool');
                     cl.remove('eq-bar');
                     break;
                  case ct.colBarGraph:
                     cl.add('eq-bar');
                     cl.remove('eq-bool');
                     break;
                  default:
                     cl.remove('eq-bool');
                     cl.remove('eq-bar');
               }
               st = c.style;
               if (cs.w !== null) {
                  cl.remove('eq-auto');
                  st.width = cs.w;
               }
               else {
                  cl.add('eq-auto');
                  st.width = '';
               }
               switch (cs.align) {
                  case 0:
                     st.textAlign = '';
                     break;
                  case 1:
                     st.textAlign = 'center';
                     break;
                  case 2:
                     st.textAlign = 'right';
               }
            }
            break;
         }

         case tag.colTitle: {
            var
               colTitle = rc.ct,
               l = colTitle.length,
               hn = el.firstElementChild.children,
               i, s;
            for (i = 0; i < l; i++)
               if ((s = colTitle[i]) !== undefined && s.length)
                  dm.setContent(hn[i].firstElementChild, s);
               else
                  hn[i].firstElementChild.textContent = ' ';
            break;
         }

         case tag.lineOrder: {
            var
               sl = el.querySelector('.eq-selected'),
               lo = this.lineOrder,
               ls = rc.s,
               si, vi, h;
            if (sl) {
               si = dm.parentIndex(sl);
               if (lo !== null) {
                  // selection view->model.
                  if (si >= lo.length)
                     throw new Error("ListBox.setAttr lineOrder: selection " +
                        si + " >= lineOrder " + lo.length);
                  si = lo[si];
               }
            }
            if ((lo = rc.o) !== null && lo.length === 0)
               lo = null;
            this.lineOrder = lo;
            if (si !== undefined) {
               if (lo !== null) {
                  // selection model->view.
                  if ((vi = lo.indexOf(si)) === -1)
                     throw new Error("ListBox.setAttr lineOrder: selection " +
                        si + " not in lineOrder, count:" + lo.length);
                  si = vi;
               }
               cls.select(el, si);
               cls.scrollIntoView(el, si);
            }
            if ((h = el.firstElementChild)
                  .classList.contains('eq-caption')) {
               var
                  n = h.querySelectorAll(
                     '.eq-caption>.eq-up,.eq-caption>.eq-dn'),
                  l = n.length, cl;
               for (var i = 0; i < l; i++) {
                  cl = n[i].classList;
                  cl.remove('eq-up');
                  cl.remove('eq-dn');
                  cl.remove('eq-first');
                  cl.remove('eq-second');
               }
               if (l = ls.length) {
                  var
                     hc = h.children,
                     hl = hc.length;
                  for (var i = 0; i < l; i++) {
                     var
                        li = ls[i],
                        ci = li.i;
                     if (ci < hl) {
                        cl = hc[ci].classList;
                        if (li.d)
                           cl.add('eq-dn');
                        else
                           cl.add('eq-up');
                        if (i === 0)
                           cl.add('eq-first');
                        else if (i === 1)
                           cl.add('eq-second');
                     }
                  }
               }
            }
            break;
         }

         case tag.content: {
            var
               n = el.querySelectorAll('.eq-column'),
               nl = n.length,
               c = rc.c,
               lo = this.lineOrder,
               ni, np, nc, nn, ln, ll, cl, i, cell, ccl, s, autoWidth;
            if (nl !== c.length)
               throw new Error("ListBox.setAttr content: " +
                  "column count " + c.length + ", expected: " + nl);
            for (ni = 0; ni < nl; ni++) {
               nn = (nc = (np = n[ni]).children).length;
               ll = (ln = c[ni]).length;
               if (lo !== null && lo.length !== ll)
                  throw new Error("ListBox.setAttr content: " +
                     "lineOrder " + lo.length + ", expected: " + ll);
               cl = np.classList;
               if (!autoWidth && cl.contains('eq-auto')) {
                  autoWidth = true;
                  dm.onNextFrame(el, { cb : cls.setAutoWidth });
               }
               for (i = 0; i < ll; i++) {
                  if (i < nn)
                     cell = nc[i];
                  else {
                     cell = document.createElement('DIV');
                     np.appendChild(cell);
                  }
                  if (lo !== null)
                     s = ln[lo[i]];
                  else
                     s = ln[i];
                  if (cl.contains('eq-bool')) {
                     cell.textContent = '';
                     ccl = cell.classList;
                     if (s === '1')
                        ccl.add('eq-active');
                     else
                        ccl.remove('eq-active');
                  }
                  else if (cl.contains('eq-bar')) {
                     cell.textContent = '';
                     if (s !== undefined && s.length)
                        cell.style.backgroundSize = s + '% 100%';
                     else
                        cell.style.backgroundSize = '0';
                  }
                  else if (s !== undefined && s.length)
                     dm.setContent(cell, s);
                  else
                     cell.textContent = ' ';
               }
               while (nc.length > ll)
                  np.removeChild(np.lastElementChild);
            }
            break;
         }

         case tag.activeLine:
         case tag.vActiveLine:
         case tag.topItem:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               self : this,
               el : el,
               t  : rc.t,
               i  : rc.i
            });
            break;

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

eq.c.ListBox.setAttr = function(rt) {
   var self = rt.self, tag = eq.RqTag;
   switch (rt.t) {
      case tag.activeLine: {
         var
            cls = self.constructor,
            el = rt.el,
            li = rt.i,
            lo, vi;
         if (li > 0) {
            li--;
            if ((lo = self.lineOrder) !== null) {
               // activeLine model->view.
               if ((vi = lo.indexOf(li)) === -1)
                  throw new Error("ListBox.setAttr: activeLine " + li +
                     " not in lineOrder, count:" + lo.length);
               li = vi;
            }
            cls.select(el, li);
            cls.scrollIntoView(el, li);
         }
         else
            cls.select(rt.el);
         break;
      }

      case tag.vActiveLine: {
         var
            cls = self.constructor,
            el = rt.el,
            li = rt.i;
         if (li > 0) {
            cls.select(el, --li);
            cls.scrollIntoView(el, li);
         }
         else
            cls.select(rt.el);
         break;
      }

      case tag.topItem: {
         var
            el = rt.el.querySelector('.eq-column'),
            n = el.children,
            bx = el.parentNode,
            st = 0;
         if (n.length > 1) {
            var smax = el.getBoundingClientRect().height
                     - bx.getBoundingClientRect().height;
            if (smax > 0) {
               var
                  lh = n[0].getBoundingClientRect().height,
                  li = rt.i, lo, vi;
               if ((lo = self.lineOrder) !== null) {
                  // topItem model->view.
                  if ((vi = lo.indexOf(li)) === -1)
                     throw new Error("ListBox.setAttr: topItem " + li +
                        " not in lineOrder, count:" + lo.length);
                  li = vi;
               }
               if (lh > 0 && (st = li * lh) > smax)
                  st = Math.ceil(smax + 2);
            }
         }
         bx.scrollTop = st;
         break;
      }

      default:
         throw new Error("ListBox.setAttr: invalid tag " + rt.t);
   }
};

/*------------------------------------------------------------------------
   ListBox.onEvent()
   Filter and/or handle action event.
------------------------------------------------------------------------*/

eq.c.ListBox.prototype.onEvent = function(dlg, id, el, e) {
   if (e.type === 'keydown')
      return true;
   var t = e.target, cl, tp;
   while (t && (!(cl = t.classList) || !cl.contains('eq-listbox'))) {
      if (cl && cl.contains('eq-column'))
         break;
      if ((tp = t.parentNode) && (cl = tp.classList)
                              && cl.contains('eq-column'))
         return true;
      t = tp;
   }
   return false;
};

/*------------------------------------------------------------------------
   ListBox.onKey()
   Filter and/or handle key event.
------------------------------------------------------------------------*/

eq.c.ListBox.prototype.onKey = function(dlg, id, el, e) {
   var
      cls = this.constructor,
      app = eq.app,
      dm = eq.Dom,
      kc = dm.isEnterOrSpace(e);
   if (kc) {
      var s = el.querySelector('.eq-column>.eq-selected');
      if (s) {
         e.eq = { key : kc };
         app.processEvent(dlg, e, id, el, this);
      }
      return true;
   }

   switch (e.key) {
      case 'Escape':
      case 'Esc': // IE
         cls.select(el);
         return true;

      // Navigation?
      case 'Home':
      case 'End':
      case 'PageUp':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowDown':
      case 'Down':
         break;

      default:
         // No.
         return false;
   }

   var
      n = el.querySelector('.eq-listbox .eq-column').children,
      d = 0,
      p = -1;
   if (n.length) {
      switch (e.key) {
         case 'Home':
            p = 0;
            kc = 36;
            break;
         case 'End':
            p = n.length - 1;
            kc = 35;
            break;
         case 'PageUp':
            d = -10;
            kc = 33;
            break;
         case 'PageDown':
            d = 10;
            kc = 34;
            break;
         case 'ArrowUp':
         case 'Up':
            d = -1;
            kc = 38;
            break;
         case 'ArrowDown':
         case 'Down':
            d = 1;
            kc = 40;
            break;
      }
      var ln = null, bx, bh, lh, li;
      if (p !== -1)
         ln = n[p];
      else if (d !== 0) {
         if ((d < -1 || d > 1) && n.length > 1) {
            // Page up/down, more than one line,
            // calculate number of visible lines.
            bx = n[0].parentNode.parentNode;
            bh = bx.getBoundingClientRect().height;
            lh = n[0].getBoundingClientRect().height;
            if (bh > 0 && lh > 0)
               d = d > 0 ? Math.floor(bh / lh) : Math.ceil(-(bh / lh));
         }
         var s = el.querySelector('.eq-column>.eq-selected');
         if (s) {
            p = dm.parentIndex(s);
            if ((p += d) < 0)
               p = 0;
            else if (p >= n.length)
               p = n.length - 1;

            ln = n[p];
         }
         else if (d > 0)
            ln = n[0];
      }

      if (ln !== null) {
         // Select line if necessary.
         if (!ln.classList.contains('eq-selected')) {
            app.setChanged(id, 'mousedown');
            cls.select(el, li = dm.parentIndex(ln));
            this.activeCol = 0;
            cls.scrollIntoView(el, li);
            // Submit on selection if enabled.
            if (this.f) {
               e.eq = { key : kc };
               app.processEvent(dlg, e, id, el, this);
               return true;
            }
         }
      }
   }

   return true;
};

/*------------------------------------------------------------------------
   ListBox.onChange()
   Filter and/or handle change event.
------------------------------------------------------------------------*/

eq.c.ListBox.prototype.onChange = function(dlg, id, el, e) {
   var
      cls = this.constructor,
      dm = eq.Dom,
      t = e.target,
      md = e.type === 'mousedown',
      cl, tp;
   while (t && (!(cl = t.classList) || !cl.contains('eq-listbox'))) {
      if (cl) {
         if (cl.contains('eq-column')) {
            cls.select(el);
            break;
         }
         if (md && cl.contains('eq-resize') && t === e.target) {
            var
               i = dm.parentIndex(t),
               c = t.parentNode.nextElementSibling.firstElementChild;
            while (i--)
               c = c.nextElementSibling;
            eq.app.moveCapture({
               el : el,
               h : t,
               c : c,
               x : e.clientX,
               w : Number(window.getComputedStyle(t).width.replace(/px$/,''))
            }, cls.colResize, cls.colResize.done);
            dm.consumeEvent(e);
            return false;
         }
      }
      if ((tp = t.parentNode) && (cl = tp.classList)) {
         if (cl.contains('eq-column')) {
            cls.select(el, dm.parentIndex(t));
            this.activeCol = dm.parentIndex(tp) + 1;
            break;
         }
         if (md && cl.contains('eq-caption')) {
            if (t === e.target)
               break;
            eq.app.mouseDownTimers.add(150,
               cls.beginColMove, cls.colHeaderClicked, {
                  self : this,
                  id   : id,
                  el   : el,
                  e    : e,
                  hd   : t
               });
            dm.consumeEvent(e);
            return false;
         }
      }
      t = tp;
   }
   return true;
};

eq.c.ListBox.colHeaderClicked = function(arg) {
   var hd = arg.hd;
   if (hd.classList.contains('eq-sort')) {
      var
         app = eq.app,
         id = arg.id,
         e = arg.e;
      app.setChanged(id, 'sort');
      app.apiUpdate(id, eq.UpdateTag.sortSequence, {
         col : eq.Dom.parentIndex(hd),
         add : e.shiftKey
      });
   }
};

eq.c.ListBox.beginColMove = function(arg) {
   var
      self = arg.self,
      id = arg.id,
      el = arg.el,
      e = arg.e,
      hd = arg.hd;
   if (hd.classList.contains('eq-move')) {
      var
         cls = self.constructor,
         dm = eq.Dom,
         app = eq.app,
         from = dm.parentIndex(hd),
         cw = [],
         ov, ovs, ob, obs, oc, op, ow, pa, rc, c, n, x, h, sp;
      // Overlay clipping box.
      ov = document.createElement('DIV');
      (ovs = ov.style).position = 'absolute';
      ovs.overflow = 'hidden';
      // Container.
      ob = document.createElement('DIV');
      ob.className = 'eq-listbox eq-multi';
      if (el.classList.contains('eq-focused'))
         ob.className += ' eq-focused';
      ov.appendChild(ob);
      // Header.
      oc = document.createElement('DIV');
      oc.className = 'eq-caption';
      ob.appendChild(oc);
      // Column.
      op = document.createElement('DIV');
      op.className = 'eq-pane';
      ob.appendChild(op);
      // Caption left, top, height.
      rc = (c = (pa = hd.parentNode)).getBoundingClientRect();
      ovs.left = (x = rc.left) + window.pageXOffset + 'px';
      ovs.top = rc.top + window.pageYOffset + 'px';
      ovs.width = rc.width + 'px';
      h = rc.height;
      // Pane height.
      rc = (c = c.nextElementSibling).getBoundingClientRect();
      ovs.height = (obs = ob.style).height = rc.height + h + 'px';
      // Column left, width.
      n = c.children;
      for (var i = 0, l = cw.length = n.length; i < l; i++) {
         rc = n[i].getBoundingClientRect();
         cw[i] = rc.width;
         if (i === from) {
            obs.left = (x = rc.left - x) + 'px';
            obs.width = (ow = rc.width) + 'px';
            c = n[i];
         }
      }
      // Spacer elements.
      sp = document.createElement('DIV');
      sp.className = 'eq-space';
      sp.style.width = hd.style.width;
      sp.appendChild(document.createElement('DIV'));
      oc.appendChild(pa.replaceChild(sp, hd));
      sp = document.createElement('DIV');
      sp.className = 'eq-column eq-space';
      sp.style.width = c.style.width;
      sp.appendChild(document.createElement('DIV'));
      op.appendChild((pa = pa.nextElementSibling).replaceChild(sp, c));
      // Activate overlay, capture column move.
      ob.id = self.duplicateCssRules(id);
      app.makeOverlay(el, ov, cls.colMove.closeOverlay);
      op.scrollTop = pa.scrollTop;
      app.moveCapture({
         el : el,
         ob : ob,
         ow : ow,
         cw : cw,
         lx : x,
         x : e.clientX,
         f : from,
         t : from
      }, cls.colMove, cls.colMove.done);
   }
};


/*------------------------------------------------------------------------
   ListBox.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.ListBox.prototype.changed = function(dlg, id, el, ty) {
   var v = [];
   for (var i = 0, l = ty.length; i < l; i++)
      switch (ty[i]) {
         case 'mousedown': {
            var
               ln = el.querySelector('.eq-selected'),
               lo, li;
            if (ln) {
               lo = this.lineOrder;
               li = eq.Dom.parentIndex(ln);
               if (lo !== null) {
                  // activeLine view->model.
                  if (li >= lo.length)
                     throw new Error("ListBox.changed: activeLine " + li +
                        " >= lineOrder " + lo.length);
                  li = lo[li];
               }
               v.push({
                  ty : eq.RsValType.activeLineCol,
                  i1 : li + 1,
                  i2 : this.activeCol
               });
            }
            else
               v.push({
                  ty : eq.RsValType.activeLineCol,
                  i1 : 0,
                  i2 : 0
               });
            break;
         }

         case 'scroll': {
            var
               n = el.querySelector('.eq-column').children,
               lo = this.lineOrder,
               li = 0, lh;
            if (n.length > 1)
               if ((lh = n[0].getBoundingClientRect().height) > 0)
                  li = Math.round(n[0].parentNode.parentNode.scrollTop / lh);
            if (lo !== null) {
               // topItem view->model.
               if (li >= lo.length)
                  throw new Error("ListBox.changed: topItem " + li +
                     " >= lineOrder " + lo.length);
               li = lo[li];
            }
            v.push({
               ty : eq.RsValType.topItem,
               iv : li
            });
            break;
         }

         case 'resize': {
            v.push({ ty : eq.RsValType.colWidth });
            break;
         }

         case 'move': {
            v.push({ ty : eq.RsValType.colOrder });
            break;
         }

         case 'sort': {
            v.push({ ty : eq.RsValType.lineOrder });
            break;
         }
      }
   switch (v.length) {
      case 0:
         return null;
      case 1:
         v[0].id = id;
         return v[0];
      default:
         return {
            id : id,
            v : v
         };
   }
};

/*------------------------------------------------------------------------
   ListBox.updated()
   Act upon API thread update response.
------------------------------------------------------------------------*/

eq.c.ListBox.prototype.updated = function(dlg, id, el, tag, arg) {
   var ut = eq.UpdateTag;
   switch (tag) {
      case ut.sortSequence:
         this.setAttr(dlg, id, el, arg.rt);
         // FALLTHROUGH
      case ut.colResized:
      case ut.colMoved:
         if ('rule' in el.dataset) {
            var
               app = eq.app,
               rule = el.dataset.rule;
            if (+rule === +rule) {
               app.onceRule = rule;
               app.onceId = id;
               app.submit(el, {
                  id : id,
                  type : 'change'
               });
            }
         }
         break;

      default:
         eq.c.DlgElement.prototype.updated.call(this, dlg, id, el, tag, arg);
   }
};

/** @preserve $Id: 20-menuitem.js,v 28.4 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class MenuItem extends DlgElement
   DLG MenuItem element class.
========================================================================*/

eq.c.DlgElement.addClass('MenuItem', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.MenuItem.subClass; // final

/*------------------------------------------------------------------------
   MenuItem.setAttr()
   Set MenuItem attributes.
------------------------------------------------------------------------*/

eq.c.MenuItem.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = eq.c.Menu, dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.text:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

/*------------------------------------------------------------------------
   MenuItem.onEvent()
   Filter and/or handle action event.
------------------------------------------------------------------------*/

eq.c.MenuItem.prototype.onEvent = function(dlg, id, el, e) {
   var dm = eq.Dom, app = eq.app, ov;
   if (ov = dm.eqOverlay(el)) {
      // Context menu item.
      dm.consumeEvent(e);
      if ('rule' in el.dataset) {
         var
            rule = el.dataset.rule,
            ed = dlg.eds.get(ov.id);
         // Obtain requesting element.
         id = ed.contextMenuId;
         el = id ? document.getElementById(id) : null;
         if (el && +rule === +rule) {
            app.onceRule = rule;
            app.onceId = id;
            app.submit(el, {
               id : id,
               type : 'contextmenu'
            });
            return false;
         }
      }
      app.closeOverlay(ov);
      return false;
   }
   return true;
};

/*------------------------------------------------------------------------
   MenuItem.onChange()
   Filter and/or handle change event.
------------------------------------------------------------------------*/

eq.c.MenuItem.prototype.onChange = function(dlg, id, el, e) {
   var app, ov;
   if (ov = eq.Dom.eqOverlay(el)) {
      // Context menu item, remove from changed elements.
      (app = eq.app).changed.delete(id);
      if (!('rule' in el.dataset))
         app.closeOverlay(ov);
   }
   return true;
};

/** @preserve $Id: 20-menu.js,v 28.4 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class Menu extends DlgElement
   DLG Menu element class.
========================================================================*/

eq.c.DlgElement.addClass('Menu', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.Menu.subClass; // final

/*------------------------------------------------------------------------
   Menu.closeContextMenu()
   Close context menu overlay.
------------------------------------------------------------------------*/

eq.c.Menu.closeContextMenu = function(el) {
   var dlg = eq.app.currDlg;
   if (!dlg)
      throw new Error("BUG: Menu.closeContextMenu no current dialog");
   dlg.removeElement(el);
};

/*------------------------------------------------------------------------
   Menu.setAttr()
   Set Menu attributes.
------------------------------------------------------------------------*/

eq.c.Menu.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = eq.c.Menu, dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.text:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         case tag.align:
            if (rc.v)
               el.classList.add('eq-right');
            else
               el.classList.remove('eq-right');
            break;

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

eq.c.Menu.setAttr = function(rt) {
   var cls = eq.c.Menu, tag = eq.RqTag;
   switch (rt.t) {
      case tag.text: {
         var
            el = rt.el,
            s = el.firstElementChild;
         if (!s || s.tagName !== 'SPAN') {
            var n = document.createElement('SPAN');
            s = el.insertBefore(n, s);
         }
         eq.Dom.setContent(s, rt.v, true);
         break;
      }

      default:
         throw new Error("Menu.setAttr: invalid tag " + rt.t);
   }
};

/*------------------------------------------------------------------------
   Menu.onChange()
   Filter and/or handle change event.
------------------------------------------------------------------------*/

eq.c.Menu.prototype.onChange = function(dlg, id, el, e) {
   var
      app = eq.app,
      dm = eq.Dom,
      p = el.parentNode,
      cl,
      t = p && (cl = p.classList) && cl.contains('eq-menubar'),
      o = (cl = el.classList).contains('eq-open');
   if (dm.eqOverlay(el)) {
      // Context menu item, remove from changed elements.
      app.changed.delete(id);
      // Not a MenuBar top-level menu.
      t = false;
   }
   if (t) {
      // MenuBar top-level menu.
      dm.closeMenu();
      if (!o) {
         cl.add('eq-open');
         app.menuOpen = id;
      }
   }
   else if (o)
      cl.remove('eq-open');
   else {
      var mo = [ el ], tn;
      while (p) {
         if (   ((tn = (el = p).tagName) !== 'UL' || tn !== 'LI')
             || ((p = p.parentNode) && (cl = p.classList)
                                    && cl.contains('eq-menubar'))) {
            var n = el.getElementsByClassName('eq-open');
            for (var i = n.length; --i >= 0;)
               n[i].classList.remove('eq-open');
            break;
         }
         if ((cl = p.classList) && cl.contains('eq-menu'))
            mo.push(p);
      }
      for (var i = mo.length; --i >= 0;)
         mo[i].classList.add('eq-open');
   }
   return false;
};

/** @preserve $Id: 20-plugin.js,v 28.3 2021/11/19 15:51:27 mike Exp $
*//*======================================================================
   class DlgPlugin extends DlgElement
   DLG plugin base class.
========================================================================*/

eq.c.DlgElement.subClass('DlgPlugin', function() {
   eq.c.DlgElement.call(this);
   var init = this.constructor.init;
   if (init)
      init.call(this);
});

/*------------------------------------------------------------------------
   static DlgPlugin.addClass()
   Subclassing convenience helper: Add DLG plugin class.
------------------------------------------------------------------------*/

eq.c.DlgPlugin.addClass = function(name, cls) {
   this.subClass(name, cls);
   eq.d.set(name, cls);
   return cls;
};

/*========================================================================
   Install plugins from declarations.
========================================================================*/

eq.plugin.install = function() {
   var m = this.m, pa = [];
   if (m !== undefined) {
      // Iterate plugin declarations, install plugins.
      m.forEach(this.cb, pa);
      m.clear();
   }
   // Global plugin object is now obsolete.
   delete eq.plugin;
   return pa;
};

eq.plugin.cb = function(decl, name) {
   var
      pa = { n : name },
      plg = eq.c.DlgPlugin.addClass('plg' + name, function() {
         eq.c.DlgPlugin.call(this);
      });
   delete plg.subClass; // final
   for (var p in decl)
      switch (p) {
         case 'atr':
            pa.a = decl.atr;
            break;
         case 'prp':
            pa.p = decl.prp;
            break;
         default:
            plg[p] = decl[p];
      }
   this.push(pa);
};

/** @preserve $Id: 20-poptext.js,v 28.5 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class PopText extends DlgElement
   DLG PopText element class.
========================================================================*/

eq.c.DlgElement.addClass('PopText', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.PopText.subClass; // final

/*------------------------------------------------------------------------
   static PopText.line()
   Obtain PopText line from item list.
------------------------------------------------------------------------*/

eq.c.PopText.line = function(t, co) {
   var c, cl;
   while (t) {
      if (!(c = t.parentNode))
         break;
      if (cl = c.classList) {
         if (cl.contains('eq-column')) {
            if (co !== undefined)
               co.c = c;
            return t;
         }
         if (cl.contains('eq-poptext')) {
            if (co !== undefined)
               co.ct = c;
            break;
         }
      }
      t = t.parentNode;
   }
   return undefined;
};

/*------------------------------------------------------------------------
   static PopText.select()
   Select PopText line in item list.
------------------------------------------------------------------------*/

eq.c.PopText.select = function(el, c, ci) {
   if (c === undefined)
      if (!(c = el.querySelector('.eq-poptext>.eq-column')))
         c = document.body.querySelector('body>.eq-poptext.eq-column');
   // Deselect.
   var n = c.querySelectorAll('.eq-column>.eq-selected');
   for (var i = 0, l = n.length; i < l; i++)
      n[i].classList.remove('eq-selected');
   // Select.
   if (ci && c === ci.parentNode) {
      ci.classList.add('eq-selected');
      if ((n = c.children).length > 1) {
         // Scroll into view if necessary.
         var
            bh = c.getBoundingClientRect().height,
            lh = n[0].getBoundingClientRect().height;
         if (bh > 0 && lh > 0) {
            var bt = c.scrollTop, lt = ci.offsetTop;
            if (lt < bt)
               c.scrollTop = lt;
            else if (lt + lh > bt + bh)
               c.scrollTop = Math.ceil(lt + lh - bh + 2);
         }
      }
   }
};

/*------------------------------------------------------------------------
   static PopText.open()
   Open PopText item list.
------------------------------------------------------------------------*/

eq.c.PopText.open = function(el, c) {
   var cl = el.classList;
   if (!cl.contains('eq-open')) {
      if (c === undefined)
         c = el.querySelector('.eq-poptext>.eq-column');
      var
         cls = eq.c.PopText,
         v = el.firstElementChild.textContent,
         rc = el.getBoundingClientRect(),
         s = c.style,
         ccl = c.classList,
         ci, w, cs;
      if (v.length) {
         // Pre-select current entry.
         var n = c.children;
         for (var i = 0, l = n.length; i < l; i++)
            if (v === n[i].textContent) {
               ci = n[i];
               break;
            }
      }
      // Set overlay position and width.
      s.left = rc.left + window.pageXOffset + 'px';
      s.top = rc.top + rc.height + window.pageYOffset + 'px';
      if (w = s.maxWidth)
         s.width = w;
      else
         s.width = rc.width + 'px';
      // Set colors.
      cs = window.getComputedStyle(el);
      s.color = cs.color;
      s.backgroundColor = cs.backgroundColor;
      // Make overlay.
      ccl.add('eq-poptext');
      if (cl.contains('eq-border'))
         ccl.add('eq-border');
      else
         ccl.remove('eq-border');
      eq.app.makeOverlay(el, c, cls.close);
      cl.add('eq-open');
      cl.add('eq-enter');
      cls.select(el, c, ci);
   }
};

/*------------------------------------------------------------------------
   static PopText.close()
   Close PopText item list.
------------------------------------------------------------------------*/

eq.c.PopText.close = function(el, c, commit) {
   var cl = el.classList, ci;
   if (!cl.contains('eq-open'))
      throw new Error("PopText.close BUG: not open, id:" + el.id);
   el.appendChild(c);
   c.classList.remove('eq-poptext');
   cl.remove('eq-open');
   cl.remove('eq-enter');
   if (commit && (ci = c.querySelector('.eq-column>.eq-selected'))) {
      var dm = eq.Dom;
      dm.copyContent(el.firstElementChild, ci);
      dm.fireEvent(el, 'input');
   }
};

/*------------------------------------------------------------------------
   static PopText.move()
   Handle PopText move capture.
------------------------------------------------------------------------*/

eq.c.PopText.move = function(e, m) {
   var c = m.c;
   if (c) {
      var
         cls = eq.c.PopText,
         t = e.target,
         el = m.el,
         co = {},
         ln = cls.line(t, co);
      if (el !== co.ct) {
         cls.select(el, c, ln);
         if (ln)
            m.close = m.commit = true;
      }
   }
};

eq.c.PopText.move.done = function(m) {
   if (m.close)
      eq.app.closeOverlay(m.el, m.commit);
   var a = m.a;
   if (a !== undefined)
      a.classList.remove('eq-click');
   return true;
};

/*------------------------------------------------------------------------
   PopText.setAttr()
   Set PopText attributes.
------------------------------------------------------------------------*/

eq.c.PopText.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = eq.c.PopText, dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.content:
         case tag.activeLine: {
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;
         }

         case tag.boxSize: {
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               w : rc.w,
               h : rc.h
            });
            break;
         }

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

eq.c.PopText.setAttr = function(rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   switch (rt.t) {
      case tag.content: {
         var
            v = rt.v,
            l = v.length,
            el = rt.el,
            c, n, nl, ln;
         if (!l)
            el.classList.add('eq-empty');
         if (!(c = el.querySelector('.eq-poptext>.eq-button'))) {
            c = document.createElement('DIV');
            c.className = 'eq-rt eq-button';
            el.appendChild(c);
         }
         if (!(c = el.querySelector('.eq-poptext>.eq-column'))) {
            c = document.createElement('DIV');
            c.className = 'eq-rt eq-column';
            el.appendChild(c);
         }
         n = c.children;
         nl = n.length;
         for (var i = 0; i < l; i++) {
            if (i < nl)
               ln = n[i];
            else {
               ln = document.createElement('DIV');
               c.appendChild(ln);
            }
            dm.setContent(ln, v[i]);
         }
         while (n.length > l)
            c.removeChild(c.lastElementChild);
         break;
      }

      case tag.activeLine: {
         var
            el = rt.el,
            c = el.querySelector('.eq-poptext>.eq-column'),
            n = c.children,
            v = rt.v,
            ci;
         if (v > 0 && v <= n.length)
            ci = n[v - 1];
         dm.copyContent(el.firstElementChild, ci);
         break;
      }

      case tag.boxSize: {
         var
            c = rt.el.querySelector('.eq-poptext>.eq-column'),
            w = rt.w,
            h = rt.h,
            s = c.style;
         s.maxWidth = w ? w : '';
         s.maxHeight = h ? h : '';
         break;
      }

      default:
         throw new Error("PopText.setAttr: invalid tag " + rt.t);
   }
};

/*------------------------------------------------------------------------
   PopText.onKey()
   Filter and/or handle key event.
------------------------------------------------------------------------*/

eq.c.PopText.prototype.onKey = function(dlg, id, el, e) {
   var
      cls = eq.c.PopText,
      c = el.querySelector('.eq-poptext>.eq-column'),
      d;
   switch (e.key) {
      case 'Enter':
         if (!c)
            eq.app.closeOverlay(el, true);
         return true;
      case 'Escape':
      case 'Esc': // IE
         if (!c)
            eq.app.closeOverlay(el, false);
         return true;

      // Navigation?
      case 'PageUp':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowDown':
      case 'Down':
         if (c) {
            cls.open(el, c);
            d = 0;
         }
         break;
      default:
         // No.
         return false;
   }

   var n, ci;
   c = document.body.querySelector('body>.eq-poptext.eq-column');
   if (c && (n = c.children).length) {
      if (d === undefined)
         switch (e.key) {
            case 'PageUp':
               ci = n[0];
               break;
            case 'PageDown':
               ci = n[n.length - 1];
               break;
            case 'ArrowUp':
            case 'Up':
               d = -1;
               break;
            case 'ArrowDown':
            case 'Down':
               d = 1;
               break;
         }
      if (d !== undefined) {
         if (ci = c.querySelector('.eq-column>.eq-selected')) {
            var i = eq.Dom.parentIndex(ci), l = n.length;
            if ((i += d) === -1)
               i = 0;
            else if (i >= l)
               i = l - 1;
            ci = n[i];
         }
         else
            ci = n[0];
      }
      if (ci !== undefined)
         cls.select(el, c, ci);
   }

   return true;
};

/*------------------------------------------------------------------------
   PopText.onChange()
   Filter and/or handle change event.
------------------------------------------------------------------------*/

eq.c.PopText.prototype.onChange = function(dlg, id, el, e) {
   if (e.type === 'mousedown') {
      var
         cls = eq.c.PopText,
         app = eq.app,
         dm = eq.Dom,
         t = e.target,
         ib = t && t.classList.contains('eq-button');
      if (!ib) {
         ib = el === dm.eqControl(t)
           && (t = el.querySelector('.eq-poptext>.eq-button'));
         if (el.classList.contains('eq-open')) {
            var
               m = { el : el },
               co = {},
               ln = cls.line(t, co);
            if (co.c) {
               cls.select(el, m.c = co.c, ln);
               m.close = m.commit = true;
            }
            else if (el === co.ct) {
               m.a = t;
               m.close = true;
               t.classList.add('eq-click');
            }
            app.moveCapture(m, cls.move, cls.move.done);
            if (el !== co.ct) {
               dm.consumeEvent(e);
               return false;
            }
            return true;
         }
      }

      if (ib) {
         if (app.setFocus(id, el)) {
            app.pendingFocus = {
               id : id,
               act : cls.open
            };
            return false;
         }
         var
            cl = el.classList,
            c = el.querySelector('.eq-poptext>.eq-column'),
            m = { el : el, a : t };
         if (cl.contains('eq-disabled') || cl.contains('eq-empty'))
            return false;
         t.classList.add('eq-click');
         if (c)
            cls.open(el, m.c = c);
         else
            m.close = true;
         app.moveCapture(m, cls.move, cls.move.done);
         dm.consumeEvent(e);
         return false;
      }
   }
   return true;
};

/*------------------------------------------------------------------------
   PopText.focusLost()
   PopText has lost focus.
------------------------------------------------------------------------*/

eq.c.PopText.prototype.focusLost = function(dlg, el) {
   if (el.classList.contains('eq-open'))
      eq.app.closeOverlay(el);
};

/*------------------------------------------------------------------------
   PopText.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.PopText.prototype.changed = function(dlg, id, el, ty) {
   for (var i = 0, l = ty.length; i < l; i++)
      if (ty[i] === 'input') {
         var
            ci = el.querySelector('.eq-poptext>.eq-column>.eq-selected'),
            ch = {
               id : id,
               ty : eq.RsValType.activeLine,
               iv : ci ? eq.Dom.parentIndex(ci) + 1 : 0
            };
         return ch;
      }
   return null;
};

/** @preserve $Id: 20-progressbar.js,v 28.3 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class ProgressBar extends DlgElement
   DLG ProgressBar element class.
========================================================================*/

eq.c.DlgElement.addClass('ProgressBar', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.ProgressBar.subClass; // final

/*------------------------------------------------------------------------
   ProgressBar.setAttr()
   Set ProgressBar attributes.
------------------------------------------------------------------------*/

eq.c.ProgressBar.prototype.setAttr = function(dlg, id, el, rt) {
   var dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.vertical: {
            if (!el.firstElementChild) {
               var
                  d = document.createElement('DIV'),
                  s1 = document.createElement('SPAN'),
                  s2 = document.createElement('SPAN');
               d.appendChild(s2);
               el.appendChild(s1);
               el.appendChild(d);
            }
            var cl = el.classList;
            if (rc.v) {
               cl.remove('eq-horz');
               cl.add('eq-vert');
            }
            else {
               cl.remove('eq-vert');
               cl.add('eq-horz');
            }
            break;
         }

         case tag.value: {
            var
               cl = el.classList,
               d = el.querySelector('.eq-progressbar>div'),
               v = rc.v;
            if (v) {
               cl.remove('eq-noval');
               cl.add('eq-val');
               if (cl.contains('eq-horz'))
                  d.style.width = v;
               else
                  d.style.height = v;
            }
            else {
               cl.remove('eq-val');
               cl.add('eq-noval');
               d.style.width = '';
            }
            break;
         }

         case tag.text: {
            var s = el.getElementsByTagName('SPAN');
            dm.setContent(s[0], rc.v);
            dm.setContent(s[1], rc.v);
            break;
         }

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

/** @preserve $Id: 20-pushbutton.js,v 28.3 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class PushButton extends DlgElement
   DLG PushButton element class.
========================================================================*/

eq.c.DlgElement.addClass('PushButton', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.PushButton.subClass; // final

/*------------------------------------------------------------------------
   PushButton.onKey()
   Filter and/or handle key event.
------------------------------------------------------------------------*/

eq.c.PushButton.prototype.onKey = function(dlg, id, el, e) {
   var dm = eq.Dom, kc = dm.isEnterOrSpace(e);
   if (kc) {
      dm.fireClick(el, { key : kc });
      return true;
   }
   return false;
};

/*------------------------------------------------------------------------
   PushButton.onTypeAhead()
   Process type-ahead key stroke.
------------------------------------------------------------------------*/

eq.c.PushButton.prototype.onTypeAhead = function(dlg, id, el, key) {
   var kc;
   switch (key) {
      case 'Enter':
      case 'ShiftEnter':
         if (el.classList.contains('eq-noenter')) {
            // .tabonenter disabled.
            kc = 10;
         }
         break;
      case ' ':
         kc = 32;
         break;
   }

   if (kc !== undefined) {
      var dm = eq.Dom;
      dm.fireClick(el, { key : kc });
      return true;
   }

   return eq.c.DlgElement.prototype.onTypeAhead.call(this, dlg, id, el, key);
};

/*------------------------------------------------------------------------
   PushButton.focusElement()
   Obtain focus element.
------------------------------------------------------------------------*/

eq.c.PushButton.prototype.focusElement = function(dlg, el) {
   var n = el.getElementsByTagName('BUTTON');
   if (n.length !== 1)
      throw new Error("BUG: PushButton.focusElement id:" + el.id +
         " elements:" + n.length + ", expected: 1");
   return n[0];
};

/** @preserve $Id: 20-radiobutton.js,v 28.3 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class RadioButton extends DlgElement
   DLG RadioButton element class.
========================================================================*/

eq.c.DlgElement.addClass('RadioButton', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.RadioButton.subClass; // final

/*------------------------------------------------------------------------
   RadioButton.onEvent()
   Filter and/or handle action event.
------------------------------------------------------------------------*/

eq.c.RadioButton.prototype.onEvent = function(dlg, id, el, e) {
   return !eq.Dom.inputElement(el).checked;
};

/*------------------------------------------------------------------------
   RadioButton.onKey()
   Filter and/or handle key event.
------------------------------------------------------------------------*/

eq.c.RadioButton.prototype.onKey = function(dlg, id, el, e) {
   var dm = eq.Dom, kc = dm.isEnterOrSpace(e);
   if (kc) {
      dm.fireClick(dm.inputElement(el), { key : kc });
      return true;
   }

   // Navigation?
   switch (e.key) {
      case 'Home':
      case 'PageUp':
      case 'End':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowLeft':
      case 'Left':
      case 'ArrowDown':
      case 'Down':
      case 'ArrowRight':
      case 'Right':
         break;
      default:
         // No.
         return false;
   }

   var
      t = e.target,
      n = el.parentNode.querySelectorAll('input[name="' + t.name + '"]'),
      d = 0;
   if (n.length) {
      switch (e.key) {
         case 'Home':
            el = n[0];
            kc = 36;
            break;
         case 'End':
            el = n[n.length - 1];
            kc = 35;
            break;
         case 'PageUp':
            el = n[0];
            kc = 33;
            break;
         case 'PageDown':
            el = n[n.length - 1];
            kc = 34;
            break;
         case 'ArrowUp':
         case 'Up':
            d = -1;
            kc = 38;
            break;
         case 'ArrowDown':
         case 'Down':
            d = 1;
            kc = 40;
            break;
         case 'ArrowLeft':
         case 'Left':
            d = -1;
            kc = 37;
            break;
         case 'ArrowRight':
         case 'Right':
            d = 1;
            kc = 39;
            break;
      }
      if (d !== 0) {
         el = null;
         for (var i = 0, l = n.length; i < l; i++)
            if (t === n[i]) {
               if ((i += d) === -1)
                  i = l - 1;
               else if (i >= l)
                  i = 0;
               el = n[i];
               break;
            }
      }
      if (el !== null)
         dm.fireClick(el, { key : kc });
   }

   return true;
};

/*------------------------------------------------------------------------
   RadioButton.onTypeAhead()
   Process type-ahead key stroke.
------------------------------------------------------------------------*/

eq.c.RadioButton.prototype.onTypeAhead = function(dlg, id, el, key) {
   var kc;
   switch (key) {
      case 'Enter':
      case 'ShiftEnter':
         if (el.classList.contains('eq-noenter')) {
            // .tabonenter disabled.
            kc = 10;
         }
         break;
      case ' ':
         kc = 32;
         break;
   }

   if (kc !== undefined) {
      var dm = eq.Dom;
      dm.fireClick(dm.inputElement(el), { key : kc });
      return true;
   }

   return eq.c.DlgElement.prototype.onTypeAhead.call(this, dlg, id, el, key);
};

/*------------------------------------------------------------------------
   RadioButton.focusElement()
   Obtain focus element.
------------------------------------------------------------------------*/

eq.c.RadioButton.prototype.focusElement = function(dlg, el) {
   return eq.Dom.inputElement(el);
};

/*------------------------------------------------------------------------
   RadioButton.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.RadioButton.prototype.changed = function(dlg, id, el, ty) {
   if (!el.checked)
      return null;
   var ch = {
      id : id,
      ty : eq.RsValType.buttonChecked,
      iv : 1
   };
   return ch;
};

/** @preserve $Id: 20-splitter.js,v 28.3 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class Splitter extends DlgElement
   DLG Splitter element class.
========================================================================*/

eq.c.DlgElement.addClass('Splitter', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.Splitter.subClass; // final

/*------------------------------------------------------------------------
   static Splitter.move()
   Handle Splitter move capture.
------------------------------------------------------------------------*/

eq.c.Splitter.move = function(e, m) {
   if (!m.busy) {
      m.busy = true;
      eq.Dom.onNextFrame(m.el, {
         cb : eq.c.Splitter.move.exec,
         x : e.clientX,
         y : e.clientY,
         m : m
      });
   }
};

eq.c.Splitter.move.exec = function(el, qi) {
   var
      x = qi.x,
      y = qi.y,
      m = qi.m,
      pa = m.pa,
      rc = pa.getBoundingClientRect();
   if (m.x !== undefined) {
      // Horizontal.
      var w = rc.width + x - m.x;
      if (w >= 0) {
         pa.style.width = w + 'px';
         if (pa.getBoundingClientRect().width !== rc.width) {
            m.x = x;
            m.modified = true;
         }
      }
   }
   else {
      // Vertical.
      var h = rc.height + y - m.y;
      if (h >= 0) {
         pa.style.height = h + 'px';
         if (pa.getBoundingClientRect().height !== rc.height) {
            m.y = y;
            m.modified = true;
         }
      }
   }

   m.busy = undefined;
};

eq.c.Splitter.move.done = function(m) {
   eq.app.submit(m.el, {
      id : m.id,
      type : 'change'
   });
   return true;
};

/*------------------------------------------------------------------------
   Splitter.setAttr()
   Set Splitter attributes.
------------------------------------------------------------------------*/

eq.c.Splitter.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = eq.c.Splitter, dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.panes:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t
            });
            break;

         case tag.position:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

eq.c.Splitter.setAttr = function(rt) {
   var tag = eq.RqTag;
   switch (rt.t) {
      case tag.panes: {
         var
            el = rt.el,
            n = el.children;
         for (var i = 0, l = n.length; i < l; i++) {
            var
               c = n[i],
               cl = c.classList;
            if (cl.contains('eq-splitter') || cl.contains('eq-groupbox')) {
               var
                  pa = document.createElement('DIV'),
                  ca = document.createElement('DIV');
               el.insertBefore(pa, c);
               pa.appendChild(c);
               ca.classList.add('eq-caption');
               pa.appendChild(ca);
            }
         }
         break;
      }

      case tag.position: {
         var
            el = rt.el,
            n = el.children;
         if (n.length === 2) {
            if (el.classList.contains('eq-vert'))
               n[0].style.height = rt.v + 'px';
            else
               n[0].style.width = rt.v + 'px';
         }
         break;
      }

      default:
         throw new Error("Splitter.setAttr: invalid tag " + rt.t);
   }
};

/*------------------------------------------------------------------------
   Splitter.onChange()
   Filter and/or handle change event.
------------------------------------------------------------------------*/

eq.c.Splitter.prototype.onChange = function(dlg, id, el, e) {
   var t = e.target;
   if (t && t.classList.contains('eq-caption')) {
      var
         cls = eq.c.Splitter,
         cl = el.classList,
         m = {
            id : id,
            el : el,
            pa : t.parentNode
         };
      if (cl.contains('eq-vert'))
         m.y = e.clientY;
      else
         m.x = e.clientX;
      eq.app.moveCapture(m,
         cls.move, cl.contains('eq-rule') ? cls.move.done : null);
      e.stopPropagation();
   }
   return false;
};

/*------------------------------------------------------------------------
   Splitter.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.Splitter.prototype.changed = function(dlg, id, el, ty) {
   var n = el.children;
   if (n.length === 2) {
      var
         rc = n[0].getBoundingClientRect(),
         ch = {
            id : id,
            ty : eq.RsValType.splitterPos,
            iv : el.classList.contains('eq-vert') ? rc.height : rc.width
         };
      return ch;
   }
   return null;
};

/** @preserve $Id: 20-tabbox.js,v 28.5 2021/11/19 15:52:53 mike Exp $
*//*======================================================================
   final class TabBox extends DlgElement
   DLG TabBox element class.
========================================================================*/

eq.c.DlgElement.addClass('TabBox', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.TabBox.subClass; // final

/*------------------------------------------------------------------------
   static TabBox.select()
   Select TabBox tab.
------------------------------------------------------------------------*/

eq.c.TabBox.select = function(el, ta) {
   var tp;
   if (ta !== undefined) {
      if ((tp = ta.parentNode) && tp.classList.contains('eq-tabs')) {
         if (ta.classList.contains('eq-selected')) {
            // Already selected.
            return null;
         }
      }
      else {
         // Not a tab.
         tp = undefined;
      }
   }
   // Deselect.
   var n = el.children;
   for (var ni = 0, nl = n.length; ni < nl; ni++) {
      var tn = n[ni].children;
      for (var i = 0, l = tn.length; i < l; i++)
         tn[i].classList.remove('eq-selected');
   }
   // Select.
   if (tp !== undefined) {
      var
         i = eq.Dom.parentIndex(ta),
         cl;
      ta.classList.add('eq-selected');
      if (tp = tp.nextElementSibling) {
         for (ta = tp.firstElementChild; ta; ta = ta.nextElementSibling) {
            if (   (cl = ta.classList)
                && (cl.contains('eq-hidden') || cl.contains('eq-disabled')))
               continue;
            if (i-- === 0) {
               ta.classList.add('eq-selected');
               return ta;
            }
         }
      }
   }
   return null;
};

/*------------------------------------------------------------------------
   TabBox.setAttr()
   Set TabBox attributes.
------------------------------------------------------------------------*/

eq.c.TabBox.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = eq.c.TabBox, dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.tabs: {
            var
               v = rc.v,
               tb = el.firstElementChild,
               ta, n, nl, s;
            if (!tb || !tb.classList.contains('eq-tabs')) {
               ta = document.createElement('DIV');
               ta.className = 'eq-rt eq-tabs';
               tb = el.insertBefore(ta, tb);
               dlg.updateControl(id, tb, {
                  ev : {
                     // Element is focusable and can submit,
                     // use 'mousedown', not 'click'.
                     c : [ 'mousedown' ]
                  }
               });
            }
            n = tb.children;
            nl = n.length;
            for (var i = 0, l = v.length; i < l; i++) {
               if (i < nl)
                  ta = n[i];
               else {
                  ta = document.createElement('DIV');
                  tb.appendChild(ta);
               }
               if ((s = v[i]).length)
                  dm.setContent(ta, s, true);
               else
                  ta.textContent = ' ';
            }
            while (n.length > l)
               tb.removeChild(tb.lastElementChild);
            // Deselect, selectedTab always set afterwards.
            cls.select(el);
            break;
         }

         case tag.selectedTab:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         case tag.rule:
            dlg.updateControl(id, el.firstElementChild, {
               ev : {
                  // Element is focusable and can submit,
                  // use 'mousedown', not 'click'.
                  a : rc.v ? [ 'mousedown' ] : null
               }
            });
            break;

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

eq.c.TabBox.setAttr = function(rt) {
   var tag = eq.RqTag;
   switch (rt.t) {
      case tag.selectedTab: {
         var
            el = rt.el,
            v = rt.v,
            cl, tb, ta;
         if (v.length) {
            var n = el.children[1].children;
            for (var i = 0, j = 0, l = n.length; i < l; i++) {
               if (   (cl = (tb = n[i]).classList)
                   && (cl.contains('eq-hidden') || cl.contains('eq-disabled')))
                  continue;
               if (v === tb.id) {
                  ta = el.firstElementChild.children[j];
                  break;
               }
               j++;
            }
         }
         eq.c.TabBox.select(el, ta);
         break;
      }

      default:
         throw new Error("TabBox.setAttr: invalid tag " + rt.t);
   }
};

/*------------------------------------------------------------------------
   TabBox.onKey()
   Filter and/or handle key event.
------------------------------------------------------------------------*/

eq.c.TabBox.prototype.onKey = function(dlg, id, el, e) {
   // Navigation?
   switch (e.key) {
      case 'Home':
      case 'PageUp':
      case 'End':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowLeft':
      case 'Left':
      case 'ArrowDown':
      case 'Down':
      case 'ArrowRight':
      case 'Right':
         break;
      default:
         // No.
         return false;
   }

   var
      n = el.parentNode.querySelectorAll('.eq-tabs > *'),
      d = 0,
      kc;
   if (n.length) {
      switch (e.key) {
         case 'Home':
            el = n[0];
            kc = 36;
            break;
         case 'End':
            el = n[n.length - 1];
            kc = 35;
            break;
         case 'PageUp':
            el = n[0];
            kc = 33;
            break;
         case 'PageDown':
            el = n[n.length - 1];
            kc = 34;
            break;
         case 'ArrowUp':
         case 'Up':
            d = -1;
            kc = 38;
            break;
         case 'ArrowDown':
         case 'Down':
            d = 1;
            kc = 40;
            break;
         case 'ArrowLeft':
         case 'Left':
            d = -1;
            kc = 37;
            break;
         case 'ArrowRight':
         case 'Right':
            d = 1;
            kc = 39;
            break;
      }
      if (d !== 0) {
         el = null;
         for (var i = 0, l = n.length; i < l; i++)
            if (n[i].classList.contains('eq-selected')) {
               if ((i += d) === -1)
                  i = l - 1;
               else if (i >= l)
                  i = 0;
               el = n[i];
               break;
            }
      }
      if (el !== null)
         eq.Dom.fireClick(el, { key : kc });
   }

   return true;
};

/*------------------------------------------------------------------------
   TabBox.onChange()
   Filter and/or handle change event.
------------------------------------------------------------------------*/

eq.c.TabBox.prototype.onChange = function(dlg, id, el, e) {
   if (e.type === 'mousedown') {
      var dm = eq.Dom, t = e.target, cl, tp, ta;
      while (t && (!(cl = t.classList) || !cl.contains('eq-tabbox'))) {
         if ((tp = t.parentNode) && (cl = tp.classList)
                                 && cl.contains('eq-tabs')) {
            if (ta = eq.c.TabBox.select(el, t)) {
               var app = eq.app;
               if (ta.classList.contains('eq-rule'))
                  app.submit(ta, {
                     id : ta.id,
                     type : 'change'
                  });
               else if (el.classList.contains('eq-rule'))
                  app.submit(el, {
                     id : id,
                     type : 'change'
                  });
               else
                  app.apiUpdate(id,
                     eq.UpdateTag.tabOrder, eq.Dom.parentIndex(t));
            }
            break;
         }
         t = tp;
      }
   }
   return false;
};

/*------------------------------------------------------------------------
   TabBox.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.TabBox.prototype.changed = function(dlg, id, el, ty) {
   if (   el.classList.contains('eq-tabs')
       && (el = el.querySelector('.eq-tabs>.eq-selected'))) {
      var ch = {
         id : id,
         ty : eq.RsValType.selectedTab,
         iv : eq.Dom.parentIndex(el)
      };
      return ch;
   }

   return null;
};

/*------------------------------------------------------------------------
   TabBox.updated()
   Act upon API thread update response.
------------------------------------------------------------------------*/

eq.c.TabBox.prototype.updated = function(dlg, id, el, tag, arg) {
   var ut = eq.UpdateTag;
   switch (tag) {
      case ut.tabOrder:
         dlg.tabOrder = arg;
         break;

      default:
         eq.c.DlgElement.prototype.updated.call(this, dlg, id, el, tag, arg);
   }
};

/** @preserve $Id: 20-tree.js,v 28.6 2021/11/19 15:52:19 mike Exp $
*//*======================================================================
   final class Tree extends DlgElement
   DLG Tree element class.
========================================================================*/

eq.c.DlgElement.addClass('Tree', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.Tree.subClass; // final

eq.c.Tree.itemFlags = {
   node : 1,
   open : 2
};

/*------------------------------------------------------------------------
   static Tree.select()
   Select Tree item.
------------------------------------------------------------------------*/

eq.c.Tree.select = function(el, i) {
   var ti, cl;
   if (i !== undefined) {
      var n = el.querySelectorAll('.eq-tree .eq-item');
      if (i >= 0 && i < n.length)
         cl = (ti = n[i]).classList;
   }
   if (ti) {
      // Expand parent nodes if necessary.
      var p = ti, pl;
      while ((p = p.parentNode) && (pl = p.classList)
                                && !pl.contains('eq-tree'))
         if (pl.contains('eq-node'))
            pl.add('eq-open');
      // Scroll into view if necessary.
      var
         bh = el.getBoundingClientRect().height,
         ih = ti.querySelector('.eq-item>span').getBoundingClientRect().height;
      if (bh > 0 && ih > 0) {
         var pt = el.scrollTop, it = ti.offsetTop;
         if (it < pt)
            el.scrollTop = Math.floor(it - ih / 2);
         else if (it + ih > pt + bh)
            el.scrollTop = Math.ceil(it + ih * 1.5 - bh);
      }
   }
   // Deselect.
   var n = el.querySelectorAll('.eq-tree .eq-selected');
   for (var ni = 0, nl = n.length; ni < nl; ni++)
      n[ni].classList.remove('eq-selected');
   // Select.
   if (ti) {
      cl.add('eq-selected');
      return true;
   }
   return false;
};

/*------------------------------------------------------------------------
   static Tree.selectedIndex()
   Obtain index of selected Tree item if any.
------------------------------------------------------------------------*/

eq.c.Tree.selectedIndex = function(el, sl) {
   if (sl === undefined)
      sl = el.querySelector('.eq-tree .eq-selected');
   if (sl) {
      var n = el.querySelectorAll('.eq-tree .eq-item');
      for (var i = 0, l = n.length; i < l; i++)
         if (sl === n[i])
            return i;
   }
   return -1;
};

/*------------------------------------------------------------------------
   static Tree.itemExpanded()
   Check whether Tree item is node and expanded.
------------------------------------------------------------------------*/

eq.c.Tree.itemExpanded = function(ti) {
   var cl;
   while (ti && (!(cl = ti.classList) || !cl.contains('eq-tree'))) {
      if (cl && (cl.contains('eq-node') || cl.contains('eq-leaf'))) {
         while ((ti = ti.parentNode) && (cl = ti.classList)
                                     && !cl.contains('eq-tree')) {
            if (cl.contains('eq-node') && !cl.contains('eq-open'))
               return false;
         }
         return true;
      }
      ti = ti.parentNode;
   }
   return false;
};

/*------------------------------------------------------------------------
   static Tree.moveDone()
   Handle Tree move capture.
------------------------------------------------------------------------*/

eq.c.Tree.moveDone = function(m) {
   m.ti.classList.remove('eq-click');
   eq.app.setChanged(m.id, 'status');
   return true;
};

/*------------------------------------------------------------------------
   Tree.setAttr()
   Set Tree attributes.
------------------------------------------------------------------------*/

eq.c.Tree.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = eq.c.Tree, dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.content:
         case tag.activeLine:
         case tag.status:
            // Depend on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

eq.c.Tree.setAttr = function(rt) {
   var cls = eq.c.Tree, tag = eq.RqTag;
   switch (rt.t) {
      case tag.content:
         cls.setContent(rt.el.querySelector('.eq-tree>.eq-pane'), rt.v);
         break;

      case tag.activeLine: {
         var v = rt.v;
         if (v > 0)
            cls.select(rt.el, v - 1);
         else
            cls.select(rt.el);
         break;
      }

      case tag.status: {
         var
            el = rt.el,
            n = el.querySelectorAll('.eq-tree .eq-node'),
            v = rt.v,
            vi = 0,
            vl = v.length;
         for (var i = 0, l = n.length; i < l; i++) {
            var cl = n[i].classList;
            if (vi < vl && i === v[vi] && !cl.contains('eq-empty')) {
               cl.add('eq-open');
               vi++;
            }
            else
               cl.remove('eq-open');
         }
         break;
      }

      default:
         throw new Error("Tree.setAttr: invalid tag " + rt.t);
   }
};

eq.c.Tree.setContent = function(p, v, pn) {
   var
      dm = eq.Dom,
      l = v.length;
   p.textContent = ''; // Delete children.
   if (!l && pn !== undefined) {
      var cl = pn.classList;
      cl.add('eq-empty');
      cl.remove('eq-open');
   }
   for (var i = 0; i < l; i++) {
      var
         tf = this.itemFlags,
         vv = v[i],
         ti = document.createElement('DIV'),
         sp = document.createElement('SPAN');
      if (typeof vv === 'object') {
         // Node.
         var
            bt = document.createElement('DIV'),
            ch = document.createElement('DIV');
         ti.className = 'eq-item eq-node';
         if (vv.f & tf.open)
            ti.classList.add('eq-open');
         bt.className = 'eq-button';
         ti.appendChild(bt);
         dm.setContent(sp, vv.hc);
         ti.appendChild(sp);
         ch.className = 'eq-pane';
         this.setContent(ch, vv.ch, ti);
         ti.appendChild(ch);
      }
      else {
         // Leaf.
         ti.className = 'eq-item eq-leaf';
         dm.setContent(sp, vv);
         ti.appendChild(sp);
      }
      p.appendChild(ti);
   }
};

/*------------------------------------------------------------------------
   Tree.onEvent()
   Filter and/or handle action event.
------------------------------------------------------------------------*/

eq.c.Tree.prototype.onEvent = function(dlg, id, el, e) {
   if (e.type === 'keydown')
      return true;
   var t = e.target, cl;
   while (t && (!(cl = t.classList) || !cl.contains('eq-tree'))) {
      if (t.tagName === 'SPAN') {
         while (   (t = t.parentNode)
                && (!(cl = t.classList) || !cl.contains('eq-tree'))) {
            if (cl && (cl.contains('eq-node') || cl.contains('eq-leaf')))
               return true;
         }
         return false;
      }
      t = t.parentNode;
   }
   return false;
};

/*------------------------------------------------------------------------
   Tree.onKey()
   Filter and/or handle key event.
------------------------------------------------------------------------*/

eq.c.Tree.prototype.onKey = function(dlg, id, el, e) {
   var
      cls = eq.c.Tree,
      app = eq.app,
      kc = eq.Dom.isEnterOrSpace(e);
   if (kc) {
      var sl = el.querySelector('.eq-tree .eq-selected>span');
      if (sl) {
         e.eq = { key : kc };
         app.processEvent(dlg, e, id, el, this);
      }
      return true;
   }

   var d;
   switch (e.key) {
      case 'Escape':
      case 'Esc': // IE
         cls.select(el);
         return true;

      // Navigation?
      case 'Home':
      case 'PageUp':
      case 'End':
      case 'PageDown':
      case 'ArrowUp':
      case 'Up':
      case 'ArrowDown':
      case 'Down':
         break;

      // Collapse?
      case 'ArrowLeft':
      case 'Left':
         d = -1;
         break;

      // Expand?
      case 'ArrowRight':
      case 'Right':
         d = 1;
         break;
      default:
         // No.
         return false;
   }

   if (d !== undefined) {
      var sl = el.querySelector('.eq-tree .eq-node.eq-selected');
      if (sl) {
         var cl = sl.classList;
         if (d === 1) {
            if (!cl.contains('eq-open') && !cl.contains('eq-empty')) {
               cl.add('eq-open');
               app.setChanged(id, 'status');
            }
         }
         else {
            if (cl.contains('eq-open')) {
               cl.remove('eq-open');
               app.setChanged(id, 'status');
            }
         }
      }
      return true;
   }

   var
      n = el.querySelectorAll('.eq-tree .eq-item'),
      l = n.length, i, si;
   if (l) {
      switch (e.key) {
         case 'Home':
            i = 0;
            kc = 36;
            break;
         case 'End':
            i = l - 1;
            kc = 35;
            break;
         case 'PageUp':
            i = 0;
            kc = 33;
            break;
         case 'PageDown':
            i = l - 1;
            kc = 34;
            break;
         case 'ArrowUp':
         case 'Up':
            d = -1;
            kc = 38;
            break;
         case 'ArrowDown':
         case 'Down':
            d = 1;
            kc = 40;
            break;
      }
      si = cls.selectedIndex(el);
      if (d !== undefined) {
         if ((i = si) === -1)
            i = 0;
         else if ((i += d) === -1)
            i = 0;
         else if (i >= l)
            i = l - 1;
      }
      var j = i;
      while (!cls.itemExpanded(n[j])) {
         if (d < 0) {
            if (--j === -1) {
               j = i;
               while (!cls.itemExpanded(n[++j])) {}
               break;
            }
         }
         else {
            if (++j === l) {
               j = i;
               while (!cls.itemExpanded(n[--j])) {}
               break;
            }
         }
      }
      if (j !== si) {
         if (cls.select(el, j))
            app.setChanged(id, 'change');
         // Submit on selection if enabled.
         if (this.f) {
            e.eq = { key : kc };
            app.processEvent(dlg, e, id, el, this);
            return true;
         }
      }
   }

   return true;
};

/*------------------------------------------------------------------------
   Tree.onChange()
   Filter and/or handle change event.
------------------------------------------------------------------------*/

eq.c.Tree.prototype.onChange = function(dlg, id, el, e) {
   if (e.type === 'mousedown') {
      var cls = eq.c.Tree, app = eq.app, t = e.target;
      if (t) {
         if (app.setFocus(id, el))
            return;
         if (t.classList.contains('eq-button')) {
            var ti = t.parentNode, cl;
            if (ti && (cl = ti.classList) && cl.contains('eq-node')) {
               var ck = true;
               if (cl.contains('eq-open'))
                  cl.remove('eq-open');
               else if (!cl.contains('eq-empty'))
                  cl.add('eq-open');
               else
                  ck = false;
               if (ck) {
                  cl.add('eq-click');
                  app.moveCapture({ id : id, ti : ti }, null, cls.moveDone);
               }
            }
         }
         else {
            var cl;
            while (t && (!(cl = t.classList) || !cl.contains('eq-tree'))) {
               if (cl && cl.contains('eq-item')) {
                  var i = cls.selectedIndex(el, t);
                  if (cls.select(el, i)) {
                     app.setChanged(id, 'change');
                     return true;
                  }
                  break;
               }
               t = t.parentNode;
            }
         }
      }
   }
   return false;
};

/*------------------------------------------------------------------------
   Tree.changed()
   Collect element changes.
------------------------------------------------------------------------*/

eq.c.Tree.prototype.changed = function(dlg, id, el, ty) {
   var v = [];
   for (var i = 0, l = ty.length; i < l; i++)
      switch (ty[i]) {
         case 'change':
            v.push({
               ty : eq.RsValType.activeLine,
               iv : eq.c.Tree.selectedIndex(el) + 1
            });
            break;

         case 'status': {
            var
               n = el.querySelectorAll('.eq-tree .eq-node'),
               vv = [];
            for (var ni = 0, nl = n.length; ni < nl; ni++)
               if (n[ni].classList.contains('eq-open'))
                  vv.push(ni);
            v.push({
               ty : eq.RsValType.status,
               vv : vv
            });
            break;
         }
      }
   switch (v.length) {
      case 0:
         return null;
      case 1:
         v[0].id = id;
         return v[0];
      default:
         return {
            id : id,
            v : v
         };
   }
};

/** @preserve $Id: 20-windowmenu.js,v 28.3 2021/11/19 15:52:19 mike Exp $
*//*======================================================================
   final class WindowMenu extends DlgElement
   DLG WindowMenu element class.
========================================================================*/

eq.c.DlgElement.addClass('WindowMenu', function() {
   eq.c.DlgElement.call(this);
});

delete eq.c.WindowMenu.subClass; // final

/*------------------------------------------------------------------------
   WindowMenu.setAttr()
   Set WindowMenu attributes.
------------------------------------------------------------------------*/

eq.c.WindowMenu.prototype.setAttr = function(dlg, id, el, rt) {
   var cls = eq.c.Menu, dm = eq.Dom, tag = eq.RqTag;
   for (var ri = 0, rl = rt.length; ri < rl; ri++) {
      var rc = rt[ri];
      switch (rc.t) {
         case tag.text:
            // Depends on child elements,
            // delay until next event cycle.
            dm.onNextCycle(cls.setAttr, {
               el : el,
               t : rc.t,
               v : rc.v
            });
            break;

         default:
            eq.c.DlgElement.prototype.setAttr.call(this, dlg, id, el, rt);
      }
   }
};

/*------------------------------------------------------------------------
   WindowMenu.onChange()
   Filter and/or handle change event.
------------------------------------------------------------------------*/

eq.c.WindowMenu.prototype.onChange = function(dlg, id, el, e) {
   var
      dm = eq.Dom,
      o = el.classList.contains('eq-open');
   dm.closeMenu();
   if (o) {
      // WindowMenu clicked when open.
      var t = e.target;
      while ((t = t.parentNode) && t !== el)
         if (t.classList.contains('eq-item')) {
            eq.c.App.activateSession(eq.app.sessionByIndex(dm.parentIndex(t)));
            break;
         }
   }
   else {
      // Open WindowMenu.
      var
         app = eq.app,
         ml = el.lastElementChild,
         tl = app.sessionTitles();
      ml.textContent = ''; // Delete children.
      for (var i = 0, l = tl.length; i < l; i++) {
         var
            mi = document.createElement('LI'),
            mt = document.createElement('SPAN');
         mi.className = 'eq-item';
         dm.setContent(mt, tl[i]);
         mi.appendChild(mt);
         ml.appendChild(mi);
      }
      el.classList.add('eq-open');
      app.menuOpen = id;
   }
   return true;
};

/** @preserve $Id: 99-main.js,v 28.66 2021/11/19 15:50:55 mike Exp $
*//*======================================================================
   final class App extends Obj
   Main thread application class.
========================================================================*/

eq.c.Obj.subClass('App', function() {
   eq.c.Obj.call(this);

/* NOTYET

   // DLG anchor elements.
   // If not empty, dialog is anchored on first element.
   // If first element is null, dialog is anchored on body element.
   // Additional anchors are used to anchor dialog children by 'cssid'.
   this.dlgAnchor = [];
*/

   // Dialog map.
   this.dlg = new Map();

   // Current dialog.
   this.currDlg = null;

   // Focus.
   this.focus = null;
   this.pendingFocus = null;
   this.dlgFocus = false;

   // Changed elements.
   this.changed = new Map();

   // Timers.
   this.dialogTimers = new eq.c.Timers();
   this.mouseDownTimers = new eq.c.Timers();

   // Application is idle (not waiting for user input).
   this.idle = true;

   // Layout grid.
   this.grid = {
      // wr : Width reference 1ch
      // wp : Width pixels
      // wf : Width ch factor
      // hr : Height reference 1em
      // hp : Height pixels
      // hf : Height em factor
   };

   // Decode script query string.
   var
      uri, host, port,
      re = new RegExp('^(http[s]?://)([^/]+)(/[^?#]+/)[^?#]*[?#]?.*$', 'i');
   try {
      // Obtain script base URI from script URI.
      uri = re.exec(document.querySelector('script#eq-webdlg').src);
      this.scriptBase = uri[1] + uri[2] + uri[3];
   }
   catch (e) {
      throw new Error("WEBDLG: <script id=\"eq-webdlg\"> not found");
   }

   try {
      // Optional query string app= parameter.
      this.app = /[?&]app=([^&#]+)/i.exec(uri[0])[1];
   } catch (e) {}

   try {
      // Optional query string host= parameter.
      host = /[?&]host=([^&#]+)/i.exec(uri[0])[1];
   } catch (e) {}

   try {
      // Optional query string port= parameter.
      port = /[?&]port=([^&#]+)/i.exec(uri[0])[1];
   } catch (e) {}

   // Compose base and WebSocket URIs.
   uri = re.exec(location.href);
   host = host ? host : uri[2];
   if (port) {
      try {
         host = /^(.+:)[^:]*$/.exec(host)[1] + port;
      }
      catch (e) {
         host += ':' + port;
      }
   }
   this.appBase = uri[1] + host + '/';
   this.appWS = 'ws' + uri[1].substring(4).toLowerCase() + host + '/';

   // Obtain/create application style sheet.
   var el = document.getElementById('eq-style');
   if (!el) {
      el = document.createElement('STYLE');
      el.id = 'eq-style';
      document.head.appendChild(el);
   }
   this.stl = {
      sheet : el.sheet,
      owner : []
   };
   if (!this.stl.sheet)
      throw new Error("WEBDLG: <style id=\"eq-style\"> StyleSheet undefined");
   this.cleanupStyles();
});

delete eq.c.App.subClass; // final

/*------------------------------------------------------------------------
   static App.doubleClickTimeout
   Max. time in millseconds after 'mouseup'
   before next 'mousedown' on same target.
------------------------------------------------------------------------*/

eq.c.App.doubleClickTimeout = 150;

/*------------------------------------------------------------------------
   static App.start()
   Start application.
------------------------------------------------------------------------*/

eq.c.App.start = function(e) {
   var self = eq.app, app = self.constructor, appOverride;

   // Deactivate 'eq-start' elements.
   var el = document.body.getElementsByClassName('eq-start');
   for (var i = 0, l = el.length; i < l; i++) {
      // Check for 'data-app' application override.
      if (e !== undefined && e.target === el[i] && ('app' in el[i].dataset))
         appOverride = el[i].dataset.app;

      // Remove 'click' handler.
      el[i].removeEventListener('click', app.start);

      // Make disabled.
      el[i].classList.add('eq-disabled');
      if (el[i].disabled !== undefined)
         el[i].disabled = true;
   }

/* NOTYET

   // Find 'eq-anchor' elements if any.
   var
      el = document.body.getElementsByClassName('eq-anchor'),
      l = el.length;
   if (l) {
      // First, find 'eq-dialog' anchor if present.
      for (var i = 0, iDialog; i < l; i++)
         if (el[i].id === 'eq-dialog') {
            iDialog = i;
            break;
         }

      // Set first anchor.
      self.dlgAnchor.push(iDialog !== undefined ? 'eq-dialog' : null);

      // Next, find additional anchors if any.
      for (var i = 0; i < l; i++)
         if (i !== iDialog && el[i].id)
            self.dlgAnchor.push(el[i].id);
   }
*/

   // Add 'eq-webdlg' document class.
   document.documentElement.classList.add('eq-webdlg');

   // Notify API thread to start application.

   if (self.currApp === undefined)
   {
      if (appOverride !== undefined)
         self.currApp = appOverride;
      else {
         if (!self.app)
            throw new Error("Cannot start, application name undefined");
         self.currApp = self.app;
      }
   }

   // Persistently store current application
   // so it can be restarted if document is reloaded.
   window.sessionStorage.setItem('eq-webdlg.app', self.currApp);

   self.api.postMessage({
      o   : eq.ApiOp.openWebSocket,
      app : self.currApp,
      url : self.appWS,
      sid : window.sessionStorage.getItem('eq-webdlg.sid')
   });
};

/*------------------------------------------------------------------------
   static App.prepareStart()
   return: true if caller may immediately start application
------------------------------------------------------------------------*/

eq.c.App.prepareStart = function(restart) {
   var self = eq.app;

   // Restart application?
   if (restart === undefined) {
      // Invoked when DOM is ready.
      if (self.currApp = window.sessionStorage.getItem('eq-webdlg.app')) {
         // Yes, done.
         this.start();
         return false;
      }
   }
   else {
      // Application has stopped, clear session storage.
      self.updateSessionList(true);
      var sid, opn;
      if (   (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
          && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))
          && sid !== opn) {
         // Secondary session, close window.
         window.close();
      }
      window.sessionStorage.removeItem('eq-webdlg.app');
      window.sessionStorage.removeItem('eq-webdlg.uat');
      window.sessionStorage.removeItem('eq-webdlg.sid');
      window.sessionStorage.removeItem('eq-webdlg.opn');
   }
   if (restart) {
      // Invoked when application has stopped and may restart.
      if (   self.currApp !== undefined
          && document.documentElement.classList.contains('eq-restart')) {
         // Yes, done.
         this.start();
         return false;
      }
   }
   // No, invalidate current application.
   self.currApp = undefined;

   // Find 'eq-start' elements if any.
   var
      el = document.body.getElementsByClassName('eq-start'),
      l = el.length;
   if (!l) {
      // No 'eq-start' elements,
      // caller may immediately start application.
      return true;
   }

   // Setup 'eq-start' click handlers.
   for (var i = 0; i < l; i++) {
      el[i].addEventListener('click', this.start);
      el[i].classList.remove('eq-disabled');
      if (el[i].disabled !== undefined)
         el[i].disabled = false;
   }

   return false;
};

/*------------------------------------------------------------------------
   static App.stopped()
   Application stopped.
------------------------------------------------------------------------*/

eq.c.App.stopped = function() {
   var self = eq.app, app = self.constructor;

   // Session still alive? Wait for subsequent 'pagehide' event.
   self.stopped = window.setTimeout(app.onStopped, 500);

   // Remove 'eq-webdlg' document class.
   document.documentElement.classList.remove('eq-webdlg');

   // Dispose timers.
   self.mouseDownTimers.dispose();
   self.dialogTimers.dispose();

   // Dispose type-ahead if necessary.
   if (self.typeAhead !== undefined) {
      app.cancelTypeAhead();
      self.typeAhead = undefined;
   }

   // Cancel current user action if any.
   self.cancelUserAction();

   // Close POPUP BOX if currently open.
   if (self.popupBox !== undefined) {
      self.popupBox.close();
      self.popupBox = undefined;
   }

   // Close Dialogs, clear current dialog.
   self.dlg.forEach(app.cbCloseDlg);
   self.dlg.clear();
   self.currDlg = null;

   // Cleanup application styles.
   self.cleanupStyles();

   // Clear focus.
   self.focus = null
   self.pendingFocus = null;
   self.dlgFocus = false;

   // Clear changed elements.
   self.changed.clear();

/* NOTYET

   // Clear 'eq-anchor' elements.
   self.dlgAnchor.length = 0;
*/

   // Clear pending 'mousedown' and 'click' events.
   self.pendingMouseDown = undefined;
   self.pendingClick = undefined;

   // Clear pending 'once' rule.
   self.onceRule = undefined;
   self.onceId = undefined;

   // Clear layout mode.
   self.layoutMode = undefined;

   // Clear pending 'notified' status.
   self.notified = undefined;

   // Application is idle.
   self.idle = true;
};

eq.c.App.onStopped = function() {
   var self = eq.app;
   self.stopped = undefined;
   // Application stopped, no subsequent 'pagehide' event.
   // Clear session storage, session probably closed by server.
   self.updateSessionList(true);
   window.sessionStorage.removeItem('eq-webdlg.app');
   window.sessionStorage.removeItem('eq-webdlg.uat');
   window.sessionStorage.removeItem('eq-webdlg.sid');
   window.sessionStorage.removeItem('eq-webdlg.opn');
};

eq.c.App.cbCloseDlg = function(d) {
   d.close();
};

/*------------------------------------------------------------------------
   static App.domReady()
   Continue main thread initialization when DOM is ready.
------------------------------------------------------------------------*/

eq.c.App.domReady = function(e) {
   var app = eq.c.App;

   // Save original page title.
   app.title = document.title;

   // IE: suppress default help action on F1 key.
   if (window.onhelp !== undefined) {
      window.addEventListener('help', app.onHelp, true);
      app.onHelpInstalled = true;
   }
   if (document.onhelp !== undefined) {
      document.addEventListener('help', app.onHelp, true);
      app.onHelpInstalled = true;
   }

   // Notify API thread when Document is complete.
   if (document.readyState === 'complete')
      app.documentComplete();
   else
      window.addEventListener('load', app.documentComplete);

   // Prepare starting application.
   if (app.prepareStart()) {
      // Immediately start application.
      app.start();
   }

   // This method/listener is now obsolete.
   if (e !== undefined)
      window.removeEventListener('DOMContentLoaded', app.domReady);
   delete app.domReady;
};

/*------------------------------------------------------------------------
   static App.documentComplete()
   Notify API thread when document is complete.
------------------------------------------------------------------------*/

eq.c.App.documentComplete = function(e) {
   var self = eq.app, app = self.constructor;

   // Calculate grid.
   var et = document.createElement('P');
   et.style.position = 'absolute';
   et.style.visibility = 'hidden';
   et.style.zIndex = eq.Dom.zIndex(document.body) - 1;
   et.style.top = et.style.left = '0';

   // Width/height reference 1ch/1em.
   var el = et.cloneNode(false);
   el.style.width = '1ch';
   el.style.height = '1em';
   document.body.appendChild(el);
   var rc = el.getBoundingClientRect(), w, h;
   self.grid.wr = w = rc.width;
   self.grid.hr = h = rc.height;
   document.body.removeChild(el);

   // Letter 'X' width/height.
   el = et.cloneNode(false);
   el.textContent = "X";
   document.body.appendChild(el);
   rc = el.getBoundingClientRect();
   self.grid.wp = rc.width;
   self.grid.wf = rc.width / w;
   self.grid.hp = rc.height + h * 0.125;
   self.grid.hf = rc.height / h + 0.125;
   document.body.removeChild(el);

   var plugin;
   if (eq.plugin !== undefined) {
      // Install plugins from declarations,
      // transfer plugin descriptors to API thread.
      plugin = eq.plugin.install();
   }

   // Notify API thread.
   self.api.postMessage({
      o : eq.ApiOp.documentComplete,
      p : plugin,
      d : self.grid
   });

   // This method/listener is now obsolete.
   if (e !== undefined)
      window.removeEventListener('load', app.documentComplete);
   delete app.documentComplete;
};

/*------------------------------------------------------------------------
   static App.apiStart()
   Start API thread.
------------------------------------------------------------------------*/

eq.c.App.apiStart = function() {
   var self = eq.app, app = self.constructor;
   self.api = new Worker(self.scriptBase + "eq-webdlg-api.js");
   self.api.onmessage = app.apiReceived;
   self.api.onerror = app.apiError;
};

/*------------------------------------------------------------------------
   static App.apiReceived()
   Process message from API thread.
------------------------------------------------------------------------*/

eq.c.App.apiReceived = function(m) {
   var self = eq.app, app = self.constructor, op = eq.MainOp, resume = false;
   for (var di = 0, dl = m.data.length; di < dl; di++) {
      var o = m.data[di].o, d = m.data[di].d;
      switch (o) {
         case op.ssnNewOrResume:
            if (!d.uat)
               throw new Error("apiReceived ssnNewOrResume:" +
                  " invalid, user agent tag undefined");
            window.sessionStorage.setItem('eq-webdlg.uat', d.uat);
            if (!d.sid)
               throw new Error("apiReceived ssnNewOrResume:" +
                  " invalid, session id undefined");
            window.sessionStorage.setItem('eq-webdlg.sid', d.sid);
            if (!window.sessionStorage.getItem('eq-webdlg.opn'))
               window.sessionStorage.setItem('eq-webdlg.opn', d.sid);
            window.name = 'eq-webdlg.' + d.sid;
            self.updateSessionList();
            continue;

         case op.applicationStopped: {
            // Application has stopped.
            var sid, opn;
            if (   (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
                && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))
                && sid !== opn) {
               // Secondary session, close window.
               window.close();
            }
            else {
               app.stopped();
               if (d) {
                  // Restart or prepare restarting.
                  app.prepareStart(true);
               }
            }
            continue;
         }

         case op.applicationFailed:
            // Prepare restarting if necessary.
            app.stopped();
            self.currApp = undefined;
            app.prepareStart(false);
            continue;

         case op.layout: {
            self.layoutMode = d;
            continue;
         }

         case op.typeAhead: {
            if (d) {
               if (self.typeAhead === undefined)
                  self.typeAhead = {
                     active : false,
                     dlg    : null,
                     focus  : null,
                     buf    : [],
                     timer  : 0
                  };
            }
            else if (self.typeAhead !== undefined) {
               app.cancelTypeAhead();
               self.typeAhead = undefined;
            }
            continue;
         }

         case op.cancelCurrentCall:
            if (self.currDlg !== null) {
               self.mouseDownTimers.dispose();
               self.dialogTimers.dispose();
               self.cancelUserAction();
               self.startTypeAhead();
               self.setIdle(self.currDlg);
               self.currDlg.ddo = undefined;
               self.currDlg = null;
               self.focus = null;
               self.changed.clear();
            }
            else if (self.popupBox !== undefined) {
               self.popupBox.close();
               self.popupBox = undefined;
            }
            continue;

         case op.updated: {
            var
               dlg = self.currDlg,
               id, el, ed;
            if (   dlg !== null && d === dlg.id
                && (el = document.getElementById(id = m.data[di].i))
                && (ed = dlg.eds.get(id)) !== undefined)
               ed.updated(dlg, id, el, m.data[di].t, m.data[di].a);
            continue;
         }

         case op.openUrl: {
            var sid = window.sessionStorage.getItem('eq-webdlg.sid');
            if (!sid)
               throw new Error("apiReceived openUrl: invalid, sid undefined");
            if (!d.name)
               throw new Error("apiReceived openUrl: invalid, name undefined");
            if (!window.open(d.url, 'eq-webdlg.' + sid + '.' + d.name))
               console.error("Failed to open window for URL '"
                  + d.url + "'");
            continue;
         }

         case op.startApp: {
            // Open new window, set query string app= parameter.
            var url = eq.Dom.updateQueryString(window.location.href, 'app', d);
            if (!window.open(url))
               console.error("Failed to open window for application '"
                  + d + "'");
            continue;
         }

         case op.activate: {
            app.activateSession(self.sessionByTitle(d));
            continue;
         }

         case op.notify:
            if (self.currDlg !== null && m.data[di].i === self.currDlg.id
                                      && self.currDlg.ddo === undefined) {
               self.notified = true;
               self.submit(document.body, { type : 'notify' });
            }
            continue;
      }

      if (self.currDlg === null || self.currDlg.ddo === undefined) {
         // Sanity check except when Dialog.do currently active.
         if (!self.idle)
            throw new Error("apiReceived op " + o + ": invalid, not idle");
         if (self.currDlg !== null)
            throw new Error("apiReceived op " + o +
               ": invalid, current dialog pending");
         if (self.focus !== null)
            throw new Error("apiReceived op " + o + ": invalid, focus pending");
         if (self.changed.size)
            throw new Error("apiReceived op " + o + ": invalid, " +
               self.changed.size + " changed elements pending");
      }
      if (self.popupBox !== undefined)
         throw new Error("apiReceived op " + o + ": invalid, POPUP BOX open");

      switch (o) {
         case op.resume:
            resume = true;
            break;

         case op.addControl: {
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (d.pa !== undefined) {
               // Add to specified parent.
               if (dlg === undefined)
                  throw new Error("apiReceived addControl: DLG " +
                     id + " does not exist");
               var
                  p = document.getElementById(d.pa),
                  el;
               if (!p)
                  throw new Error("apiReceived addControl: parent '" +
                     d.pa + "' does not exist");
               p = p.querySelector('.eq-container');
               if (!p)
                  throw new Error("apiReceived addControl: parent '" +
                     d.pa + "' container not found");
               el = dlg.addControl(d, p);
/* NOTYET
               // TODO: dlgAnchor
               if (self.dlgAnchor.indexOf(d.pa) !== -1) {
                  el.classList.add('eq-root');
                  dlg.root.push(el);
                  app.addRootListeners(el);
                  self.setIdle(dlg);
               }
*/
            }
            else {
               // Add to root.
               if (dlg !== undefined)
                  dlg.close();
               else
                  dlg = self.newDialog(id);
               dlg.addRoot(dlg.addControl(d));
               self.setIdle(dlg);
            }
            break;
         }

         case op.modifyControl: {
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (dlg === undefined)
               throw new Error("apiReceived modifyControl: DLG " +
                  id + " does not exist");
            if (!d.id)
               throw new Error("apiReceived modifyControl: id not set");
            if (!(el = document.getElementById(d.id)))
               throw new Error("apiReceived modifyControl: id '" +
                  d.id + "' does not exist");
            dlg.modifyControl(d.id, el, d);
            break;
         }

         case op.deleteControl: {
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id),
               el = document.getElementById(d),
               i;
            if (dlg === undefined)
               throw new Error("apiReceived deleteControl: DLG " +
                  id + " does not exist");
            if (!el)
               throw new Error("apiReceived deleteControl: id '" +
                  d + "' does not exist");
            if (dlg.removeElement(el))
               self.closeDialog(dlg);
            break;
         }

         case op.invalidate: {
            for (var i = d.length; --i >= 0;) {
               var dlg = self.dlg.get(d[i]);
               if (dlg !== undefined)
                  self.closeDialog(dlg);
            }
            break;
         }

         case op.dlg_DRAW: {
            if (resume) {
               resume = false;
               self.dialogsVisible();
            }
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (dlg === undefined)
               throw new Error("apiReceived DLG DRAW: DLG " +
                  id + " does not exist");
            // Make Dialog top and visible.
            self.setCurrDialog(dlg);
            if (dlg.ddo === undefined) {
               // Not Dialog.do, immediately reset currDlg.
               self.currDlg = null;
            }
            // Notify API thread on next event cycle.
            eq.Dom.onNextCycle(app.sendSyncedNotification);
            break;
         }

         case op.dlg_DO: {
            if (resume) {
               resume = false;
               self.dialogsVisible();
            }
            var
               id = m.data[di].i,
               dlg = self.dlg.get(id);
            if (dlg === undefined)
               throw new Error("apiReceived DLG DO: DLG " +
                  id + " does not exist");
            // Make Dialog top and visible.
            self.setCurrDialog(dlg);
            if (d.f)
               self.setFocus(d.f);
            if (d.t !== undefined)
               dlg.tabOrder = d.t;
            // Activate dialog.
            self.setIdle();
            // Execute pending focus action if any.
            if (self.pendingFocus !== null) {
               var f = self.pendingFocus, act, el;
               self.pendingFocus = null;
               if (   f.id === self.focus
                   && (act = f.act)
                   && (el = document.getElementById(f.id)))
                  act(el);
            }
            if (dlg.ddo !== undefined) {
               // Dialog.do implicit DLG DRAW,
               // notify API thread on next event cycle.
               eq.Dom.onNextCycle(app.sendSyncedNotification);
            }
            else {
               if (dlg.w !== null) {
                  // Reset Window close button state if necessary,
                  // act upon pending Window close request.
                  var el = dlg.w.root.querySelector(
                     '.eq-window>.eq-pane>.eq-caption>.eq-close.eq-click');
                  if (el)
                     el.classList.remove('eq-click');
                  if (dlg.pendingClose) {
                     eq.Dom.onNextCycle(app.submitWindowClose);
                     break;
                  }
               }
               // Dialog has focus.
               self.dlgFocus = true;
               // Start Dialog timer if configured.
               if (dlg.timer !== undefined)
                  self.dialogTimers.add(dlg.timer.delay,
                     app.submitDialogTimer, null, dlg.timer.rule);
            }
            break;
         }

         case op.dlg_POPUP_BOX: {
            if (resume) {
               resume = false;
               self.dialogsVisible();
            }
            self.popupBox = new eq.c.PopupBox(d,
               document.getElementById('eq-dialog'));
            // POPUP BOX has focus.
            self.dlgFocus = true;
            break;
         }

         default:
            throw new Error("apiReceived invalid op: " + o);
      }
   }

   if (resume) {
      self.dialogsVisible();
      var dlg = self.topDialog();
      if (dlg && dlg.ddo !== undefined) {
         // Resuming Dialog.do, activate dialog.
         if (self.currDlg !== null)
            throw new Error("apiReceived resume: " +
               "invalid, current dialog pending");
         self.setCurrDialog(dlg);
         if (dlg.ddo.f)
            self.setFocus(dlg.ddo.f);
         if (dlg.ddo.t)
            dlg.tabOrder = dlg.ddo.t;
         self.setIdle();
      }
   }
};

eq.c.App.sendSyncedNotification = function() {
   eq.app.api.postMessage({ o : eq.ApiOp.onSynced });
};

eq.c.App.submitWindowClose = function() {
   var
      self = eq.app,
      d = self.currDlg;
   if (d !== null && d.pendingClose) {
      d.pendingClose = undefined;
      self.submit(document.body, { type : 'close' });
   }
};

eq.c.App.submitDialogTimer = function(rule) {
   var self = eq.app;
   self.onceRule = rule;
   self.submit(document.body, { type : 'timer' });
};

/*------------------------------------------------------------------------
   App.apiUpdate()
   Post update request to API thread.
------------------------------------------------------------------------*/

eq.c.App.prototype.apiUpdate = function(id, tag, arg) {
   var dlg = this.currDlg;
   if (!dlg)
      throw new Error("BUG: apiUpdate no current dialog");
   this.api.postMessage({
      o : eq.ApiOp.update,
      d : dlg.id,
      i : id,
      t : tag,
      a : arg
   });
};

/*------------------------------------------------------------------------
   static App.apiError()
   Handle API thread error.
------------------------------------------------------------------------*/

eq.c.App.apiError = function(e) {
/* NOTYET

   var self = eq.app, app = self.constructor;
   // Terminate and restart API thread.
   self.api.terminate();
   app.apiStart();
   // Stop application, prepare restarting if necessary.
   app.stopped();
   self.currApp = undefined;
   app.prepareStart(false);
*/
   // Issue error.
   throw new Error("WEBDLG API thread failed in line " +
      e.lineno + ": " + e.message);
};

/*------------------------------------------------------------------------
   static App.onEvent()
   Root element event handler.
------------------------------------------------------------------------*/

eq.c.App.onEvent = function(e) {
   var
      self = eq.app,
      dm = eq.Dom,
      t = e.target,
      focusIn, mouseDown;

   // Dialog has focus.
   self.dlgFocus = true;

   switch (e.type) {
      case 'focusin':
         focusIn = true;
         break;
      case 'mousedown':
         mouseDown = true;
         break;
   }

   if (t.classList.contains('eq-idle')) {
      // Event on idle pane, set wait cursor.
      t.style.cursor = 'wait';
      dm.consumeEvent(e);
      return;
   }
   if (self.idle)
      return;

   var
      el = dm.eqControl(t), // Obtain 'eq-control' element.
      mdFocus, moEvent;
   if (el) {
      if (el.classList.contains('eq-disabled')) {
         dm.consumeEvent(e);
         if (self.menuOpen)
            dm.onNextCycle(dm.closeMenu);
         return;
      }

      var id = el.id, dlg = self.currDlg, ed;
      if (!dlg)
         throw new Error("BUG: onEvent no current dialog");
      if ((ed = dlg.eds.get(id)) !== undefined) {
         if (mouseDown) {
            mdFocus = el.classList.contains('eq-focus');
            self.pendingMouseDown = { id : id, t : t };
         }
         else if (focusIn) {
            // About to receive focus.
            if (el.classList.contains('eq-focus')) {
               // Set focus if different from current focus.
               if (id !== self.focus)
                  self.setFocus(id, el);
            }
            else {
               // Target not focusable, ignore event.
               dm.consumeEvent(e);
            }
            return;
         }

         self.processEvent(dlg, e, id, el, ed);
      }

      moEvent = self.menuOpen && dm.isMenu(el);
   }

   if (mouseDown) {
      if (!mdFocus) {
         // target not focusable.
         dm.consumeEvent(e);
      }
      if (self.menuOpen && !moEvent) {
         // Menu open, not a menu event, close menu on next event cycle.
         dm.onNextCycle(dm.closeMenu);
      }
   }
};

/*------------------------------------------------------------------------
   App.processEvent()
   Process change and action event.
------------------------------------------------------------------------*/

eq.c.App.prototype.processEvent = function(dlg, e, id, el, ed) {
   var type = e.type;

   // Change event?
   if (ed.c !== undefined && ed.c.indexOf(type) !== -1) {
      // Yes, enqueue change if not already in queue.
      var ch = this.changed.get(id);
      if (ch === undefined)
         this.changed.set(id, [ type ]);
      else if (ch.indexOf(type) === -1)
         ch.push(type);
      if (!ed.onChange(dlg, id, el, e))
            return false;
   }

   // Action event?
   if (ed.a !== undefined && ed.a.indexOf(type) !== -1) {
      // Yes.
      if (!ed.onEvent(dlg, id, el, e))
         return false;
      var pn;
      if (type === 'mousedown') {
         // 'mousedown', register pending 'click'.
         if ((pn = this.pendingClick) !== undefined) {
            if (ed.m === undefined)
               throw new Error("BUG: onEvent pending click id:" + pn.id +
                  " cnt:" + pn.cnt);
            this.stopPendingClickTimer();
            if (id !== pn.id) {
               // target different from already pending 'click' target,
               // submit previous event, ignore this event.
               eq.c.App.commitPendingClick();
            }
            return true;
         }
         var ea = { btn : e.buttons };
         if (ea.btn === undefined) {
            switch (e.button) {
               case 2: // Secondary button, usually right button.
                  ea.btn = 2;
                  break;
               case 1: // Auxiliary button, usually wheel/middle button.
                  ea.btn = 4;
                  break;
               default: // Assume main button, usually left button.
                  ea.btn = 1;
            }
         }
         this.pendingClick = {
            id : id,
            type : 'click',
            cnt : 0,
            ea : ea
         };
      }
      else if ((pn = this.pendingClick) !== undefined) {
         // 'click' pending, append event.
         if (pn.ev !== undefined)
            throw new Error("BUG: onEvent pending click id:" + pn.id +
               " and id:" + pn.ev.id + " type:" + pn.ev.type);
         pn.ev = {
            id : id,
            type : type,
            ea : e.eq
         };
      }
      else {
         // Submit event.
         this.submit(el, {
            id : id,
            type : type,
            ea : e.eq
         });
      }

      return true;
   }

   return false;
};

/*------------------------------------------------------------------------
   App.cancelUserAction()
   Cancel current user action if any.
------------------------------------------------------------------------*/

eq.c.App.prototype.cancelUserAction = function() {
   if (this.capture !== undefined) {
      // Clear pending capture.
      this.endCapture();
   }
   // Close overlay if currently present.
   this.closeOverlay();
   // Close menu if currently open.
   eq.Dom.closeMenu();
};

eq.c.App.onCancelUserAction = function(e) {
   if (e.type == 'focus') {
      if (eq.Dom.isRootChild(e.target))
         eq.app.dlgFocus = true;
   }
   else if (e.target === window) {
      var app = eq.app;
      app.cancelUserAction();
      if (e.type == 'blur')
         app.dlgFocus = false;
   }
};

/*------------------------------------------------------------------------
   App.setChanged()
   Set element changed.
------------------------------------------------------------------------*/

eq.c.App.prototype.setChanged = function(id, ty) {
   var dlg = this.currDlg, ed;
   if (!dlg)
      throw new Error("BUG: onEvent no current dialog");
   if ((ed = dlg.eds.get(id)) !== undefined) {
      if (ed.c !== undefined && ed.c.indexOf(ty) !== -1) {
         var ch = this.changed.get(id);
         if (ch === undefined)
            this.changed.set(id, [ ty ]);
         else if (ch.indexOf(ty) === -1)
            ch.push(ty);
      }
   }
};

/*------------------------------------------------------------------------
   App.elementsChanged()
   Collect and return element changes.
------------------------------------------------------------------------*/

eq.c.App.prototype.elementsChanged = function(ea) {
   var
      app = this.constructor,
      dlg = this.currDlg,
      w, r, p, m,
      changed = [{
         ty : eq.RsValType.idFocus,
         sv : this.focus !== null ? this.focus : ""
      }];
   if (!dlg)
      throw new Error("BUG: elementsChanged no current dialog");
   if (ea !== undefined) {
      // Dialog-specific event attributes.
      if (ea.key !== undefined) {
         changed.push({
            ty : eq.RsValType.ruleKey,
            iv : ea.key
         });
      }
      else {
         if (ea.btn !== undefined) {
            changed.push({
               ty : eq.RsValType.mouseBtns,
               iv : ea.btn
            });
         }
         if (ea.cnt !== undefined) {
            changed.push({
               ty : eq.RsValType.mouseCnt,
               iv : ea.cnt
            });
         }
         if (ea.mod !== undefined) {
            changed.push({
               ty : eq.RsValType.mouseMods,
               iv : ea.mod
            });
         }
      }
   }
   if (this.notified) {
      changed.push({
         ty : eq.RsValType.notified
      });
      this.notified = undefined;
   }
   if ((w = dlg.w) !== null && (p = (r = w.root).querySelector(
         '.eq-window>.eq-pane>.eq-dialog>.eq-pane'))) {
      if ((m = w.modified).maximized) {
         m.maximized = undefined;
         changed.push({
            ty : eq.RsValType.dialogMax,
            iv : r.classList.contains('eq-max') ? 1 : 0
         });
      }
      if (m.pos) {
         m.pos = undefined;
         changed.push({
            ty : eq.RsValType.dialogPos,
            x : r.offsetLeft,
            y : r.offsetTop
         });
      }
      if (m.size) {
         m.size = undefined;
         changed.push({
            ty : eq.RsValType.dialogSize,
            w : p.offsetWidth,
            h : p.offsetHeight
         });
      }
   }
   if (this.onceRule !== undefined) {
      changed.push({
         id : this.onceId,
         ty : eq.RsValType.ruleOnce,
         iv : this.onceRule
      });
      this.onceRule = undefined;
      this.onceId = undefined;
   }
   this.changed.forEach(app.cbElementsChanged, { d: dlg, c : changed });
   this.changed.clear();
   return changed;
};

eq.c.App.cbElementsChanged = function(ty, id) {
   var
      dlg = this.d,
      ed, el, v;
   if ((ed = dlg.eds.get(id)) === undefined)
      throw new Error("BUG: elementsChanged id:" + id +
         " not registered, type:" + ty);
   el = document.getElementById(id);
   if (!el.classList.contains('eq-event'))
      el = el.querySelector('.eq-event');
   if (!el)
      throw new Error("BUG: cbElementsChanged id:" + id +
         " event element not found");
   if ((v = ed.changed(dlg, id, el, ty)) !== null)
      this.c.push(v);
};

/*------------------------------------------------------------------------
   App.submit()
   Submit event.
------------------------------------------------------------------------*/

eq.c.App.prototype.submit = function(el, ev) {
   var
      app = this.constructor,
      dm = eq.Dom,
      dlg = this.currDlg,
      ddo;
   this.mouseDownTimers.dispose();
   this.dialogTimers.dispose();
   this.cancelUserAction();
   if (   this.onceRule === -1
       || (this.onceRule === undefined && el.classList.contains('eq-help'))) {
      this.api.postMessage({
         o : eq.ApiOp.onHelp,
         d : this.onceRule === -1 && this.onceId !== undefined
           ? this.onceId : el.id
      });
      return;
   }
   if ((ddo = dlg.ddo) !== undefined) {
      // Dialog.do event.
      dlg.ddo = undefined;
      var
         f = this.focus !== null ? this.focus : "",
         r;
      if (ddo.hide || ddo.show) {
         // Hide and/or show as specified.
         if (ddo.hide && (el = document.getElementById(ddo.hide)))
            el.style.display = 'none';
         if (ddo.show && (el = document.getElementById(ddo.show)))
            el.style.display = '';
      }
      else {
         // Hide Dialog.
         for (var i = dlg.root.length; --i >= 0;) {
            if ((r = dlg.root[i]).classList.contains('eq-dialog'))
               dlg.w.root.style.display = 'none';
            else
               r.style.display = 'none';
         }
      }
      this.currDlg = null;
      this.focus = null;
      this.changed.clear();
      this.setIdle(dlg);
      // Notify API thread.
      this.api.postMessage({
         o : eq.ApiOp.onDo,
         i : dlg.id,
         f : f
      });
      return;
   }
   // Special case for CheckBox, RadioButton,
   // enqueue change if not already in queue.
   if (   (   el.classList.contains('eq-checkbox')
           || el.classList.contains('eq-radiobutton'))
       && !this.changed.has(el.id))
      this.changed.set(el.id, 'input');
   // Notify API thread on next event cycle.
   this.setIdle(dlg);
   dm.onNextCycle(app.apiSubmit, ev);
};

eq.c.App.apiSubmit = function(ev) {
   var self = eq.app;
   if (self.currDlg !== null) {
      ev.changed = self.elementsChanged(ev.ea);
      self.api.postMessage({
         o : eq.ApiOp.onEvent,
         d : ev
      });
      self.startTypeAhead();
      self.currDlg = null;
      self.focus = null;
   }
};

/*------------------------------------------------------------------------
   static App.onGlobalKeyDown() .onGlobalKeyUp()
   Global keyboard event handler.
------------------------------------------------------------------------*/

eq.c.App.onGlobalKeyDown = function(e) {
   var self = eq.app, app = self.constructor, dm = eq.Dom;
   if (!self.dlgFocus) {
      // Focus is somewhere else, ignore event.
      if (self.debounceKeyTimer !== undefined) {
         window.clearTimeout(self.debounceKeyTimer);
         self.debounceKeyTimer = undefined;
         self.debounceKey = undefined;
      }
      return;
   }

   if (e.key === self.debounceKey) {
      // Continue debouncing key.
      app.debounceKey(e.key);
      dm.consumeEvent(e);
      return;
   }

   if (self.popupBox !== undefined) {
      // POPUP BOX is open.
      if (self.popupBox.onKey(e))
         dm.consumeEvent(e);
      return;
   }

   if (self.addTypeAhead(e)) {
      // Key stored in type-ahead buffer.
      dm.consumeEvent(e);
      return;
   }

   if (self.idle) {
      dm.consumeEvent(e);
      return;
   }

   // Current Dialog?
   var dlg = self.currDlg;
   if (dlg) {
      // Obtain target element, overlay or 'eq-control' or focus.
      var el, id, ed, cl, t = e.target;
      if (self.overlay !== undefined)
         el = self.overlay.el;
      else if (self.focus !== null)
         if (t === document.body || dm.eqControl(t))
            el = document.getElementById(self.focus);
      if (el) {
         if ((cl = el.classList).contains('eq-disabled')) {
            dm.consumeEvent(e);
            return;
         }
         // DLG element?
         if ((ed = dlg.eds.get(id = el.id)) === undefined)
            return;
      }

      // Navigation or function key?
      var fkey;
      switch (e.key) {
         case 'Enter':
            if (cl && cl.contains('eq-enter')) {
               // Enter key has special function.
               break;
            }
            if (dlg.fkey && dlg.fkey.length > 0 && dlg.fkey[0]) {
               // Dialog.cr function key.
               fkey = 0;
               break;
            }
            if (!cl || cl.contains('eq-noenter')) {
               // No focused element, or .tabonenter disabled.
               break;
            }
            // FALLTHROUGH
         case 'Tab':
            if (el) {
               dm.consumeEvent(e);

               if (!self.inSetFocus) {
                  // Navigate to next/previous element in tab order.
                  var elNext = dlg.nextTab(id, e.shiftKey);
                  if (elNext)
                     self.setFocus(elNext.id, elNext);
                  else {
                     // No other element found in tab order,
                     // submit 'blur' event if registered.
                     self.processEvent(dlg, {
                        type : 'blur',
                        target : ed.focusElement(dlg, el)
                     }, id, el, ed);
                  }
               }
            }
            return;

         case 'F1':
            fkey = 1;
            if (app.onHelpInstalled) {
               // IE: suppress default help action on F1 key.
               self.onHelpIgnore = true;
            }
            break;
         case 'F2':
            fkey = 2;
            break;
         case 'F3':
            fkey = 3;
            break;
         case 'F4':
            fkey = 4;
            break;
         case 'F5':
            fkey = 5;
            break;
         case 'F6':
            fkey = 6;
            break;
         case 'F7':
            fkey = 7;
            break;
         case 'F8':
            fkey = 8;
            break;
         case 'F9':
            fkey = 9;
            break;
         case 'F10':
            fkey = 10;
            break;
         case 'F11':
            fkey = 11;
            break;
         case 'F12':
            fkey = 12;
            break;
      }

      if (el && fkey === undefined && ed.onKey(dlg, id, el, e)) {
         dm.consumeEvent(e);
         return;
      }

      if (fkey !== undefined) {
         // Function key?
         if (dlg.fkey && fkey < dlg.fkey.length) {
            var r = dlg.fkey[fkey];
            if (r) {
               dm.consumeEvent(e);

               if (r > 0) {
                  // Submit function key rule.
                  self.onceRule = r;
                  if (   self.focus !== null
                      && (el = document.getElementById(self.focus))) {
                     self.onceId = self.focus;
                     self.submit(el, { id : self.onceId, type : 'key' });
                  }
                  else
                     self.submit(document.body, { type : 'key' });
               }
               else if (r === -1)
                  self.api.postMessage({
                     o : eq.ApiOp.onHelp,
                     d : el ? el.id : null
                  });

               return;
            }
         }

         self.onHelpIgnore = undefined;
      }
   }
};

eq.c.App.onGlobalKeyUp = function(e) {
   var self = eq.app;
   if (e.key === self.debounceKey) {
      window.clearTimeout(self.debounceKeyTimer);
      self.debounceKeyTimer = undefined;
      self.debounceKey = undefined;
      eq.Dom.consumeEvent(e);
   }
};

/*------------------------------------------------------------------------
   static App.debounceKey()
   Debounce key stroke.
------------------------------------------------------------------------*/

eq.c.App.debounceKey = function(key) {
   var self = eq.app;
   self.debounceKey = key;
   if (self.debounceKeyTimer !== undefined)
      window.clearTimeout(self.debounceKeyTimer);
   self.debounceKeyTimer = window.setTimeout(this.onDebounceKeyTimer, 500);
};

eq.c.App.onDebounceKeyTimer = function() {
   var self = eq.app;
   // Timer has elapsed.
   self.debounceKeyTimer = undefined;
   self.debounceKey = undefined;
};

/*------------------------------------------------------------------------
   static App.help()
   IE: Selectively filter F1 function key default action.
------------------------------------------------------------------------*/

eq.c.App.onHelp = function(e) {
   var self = eq.app;
   if (self.onHelpIgnore) {
      // IE: suppress default help action on F1 key.
      self.onHelpIgnore = undefined;
      eq.Dom.consumeEvent(e);
      return false;
   }
   return true;
};

/*------------------------------------------------------------------------
   static App.onGlobalMouseDownOrClick()
   Global 'mousedown' or 'click' event handler.
------------------------------------------------------------------------*/

eq.c.App.onGlobalMouseDownOrClick = function(e) {
   var target = e.target, cl;
   if (e.button !== 0) {
      // Not the default button.
      while (target && target !== document) {
         if (   (cl = target.classList)
             && (cl.contains('eq-root') || cl.contains('eq-window'))) {
            e.preventDefault();
            e.stopImmediatePropagation();
            break;
         }
         target = target.parentNode;
      }
      return;
   }

   var
      self = eq.app,
      dm = eq.Dom,
      md = e.type === 'mousedown';

   if (self.capture !== undefined) {
      // End capture.
      e.stopPropagation();
      self.endCapture();
   }

   // 'mousedown' on overlay?
   if (md && self.overlay !== undefined) {
      if (dm.eqOverlay(target)) {
         var
            el = dm.eqControl(target) || self.overlay.el,
            id = el.id,
            dlg = self.currDlg,
            ed;
         if (!dlg)
            throw new Error(
               "BUG: onGlobalMouseDownOrClick overlay no current dialog");
         if ((ed = dlg.eds.get(id)) !== undefined)
            self.processEvent(dlg, e, id, el, ed);
      }
      else if (dm.eqControl(target) !== self.overlay.el) {
         // Clicked outside overlay, close.
         self.closeOverlay();
      }
   }

   // Close menu if clicked outside Dialog or MenuBar.
   if (self.menuOpen) {
      var
         el = target,
         found;
      while (el && el !== document) {
         if (cl = el.classList)
            if (cl.contains('eq-dialog') || cl.contains('eq-menubar')) {
               found = true;
               break;
            }
         el = el.parentNode;
      }
      if (!found)
         dm.onNextCycle(dm.closeMenu);
   }

   // POPUP BOX.
   if (self.idle && self.popupBox !== undefined) {
      if (self.popupBox.isEvent(e)) {
         if (self.popupBox.close(e)) {
            self.popupBox = undefined;
            return;
         }
      }
      else if (!target.classList.contains('eq-idle'))
         return;

      if (md) {
         // 'mousedown', target not focusable.
         e.preventDefault();
      }
   }
   else {
      var w = eq.c.Window.fromTarget(target);
      if (w) {
         w.onMouseDownOrClick(e);
         if (md) {
            // 'mousedown', target not focusable.
            dm.consumeEvent(e);
         }
      }
   }
};

/*------------------------------------------------------------------------
   static App.onGlobalMouseUp()
   Global 'mouseup' event handler.
------------------------------------------------------------------------*/

eq.c.App.onGlobalMouseUp = function(e) {
   var self = eq.app;

   // Cancel mousedown timers if any.
   self.mouseDownTimers.cancel();

   if (self.capture !== undefined) {
      // End capture.
      e.stopPropagation();
      if (!self.endCapture()) {
         self.pendingMouseDown = undefined;
         self.pendingClick = undefined;
         return;
      }
   }

   if (self.idle) {
      // 'mouseup' while idle.
      self.pendingMouseDown = undefined;
      // 'click' must not be pending.
      if (self.pendingClick !== undefined) {
         var id = self.pendingClick.id;
         self.pendingClick = undefined;
         throw new Error("BUG: idle onGlobalMouseUp pending click id:" + id);
      }
      return;
   }

   var dm = eq.Dom, pn;
   if ((pn = self.pendingMouseDown) !== undefined) {
      // 'mouseup' while 'mousedown' pending.
      var t = pn.t;
      if (t === e.target) {
         // 'mouseup' target same as pending 'mousedown' target,
         // obtain 'eq-control' element.
         var
            id = pn.id,
            dlg = self.currDlg,
            el = dm.eqControl(t);
         if (!dlg)
            throw new Error("BUG: onGlobalMouseUp no current dialog");
         // 'click' committed.
         self.processEvent(dlg, {
            type : 'click',
            target : t
         }, id, el, dlg.eds.get(id));
         if (el.classList.contains('eq-focus')) {
            // Element focusable, set focus.
            self.setFocus(id, el);
         }
      }
      else if (self.menuOpen && dm.isMenu(t = e.target)) {
         // menu is open, 'mouseup' target is menu element,
         // obtain 'eq-control' element.
         var
            dlg = self.currDlg,
            el = dm.eqControl(t),
            id = el.id;
         if (!dlg)
            throw new Error("BUG: onGlobalMouseUp no current dialog");
         // 'click' committed.
         self.processEvent(dlg, {
            type : 'click',
            target : t
         }, id, el, dlg.eds.get(id));
      }

      self.pendingMouseDown = undefined;

      // Close menu if item click not submitted.
      if (self.menuOpen && dm.isMenu(t = e.target)) {
         var el = dm.eqControl(t);
         if (el && el.classList.contains('eq-item'))
            dm.onNextCycle(dm.closeMenu);
      }
   }

   if ((pn = self.pendingClick) !== undefined) {
      // 'mouseup' while 'click' pending.
      self.stopPendingClickTimer();
      // Obtain 'eq-control' element.
      var
         dlg = self.currDlg,
         el = dm.eqControl(e.target),
         id, ed;
      if (!dlg)
         throw new Error("BUG: onGlobalMouseUp pendingClick no current dialog");
      if (el && (id = el.id) === pn.id) {
         // 'mouseup' target same as pending 'click' target,
         // obtain event attributes btn, cnt, mod.
         if (e.eq !== undefined)
            pn.ea = e.eq;
         else {
            var
               ea = pn.ea = pn.ea || {},
               mod = 0;
            ea.cnt = pn.cnt ? pn.cnt + 1 : 1;
            if (e.shiftKey)
               mod += 1;
            if (e.ctrlKey)
               mod += 2;
            if (e.metaKey)
               mod += 4;
            if (e.altKey)
               mod += 8;
            if (mod)
               ea.mod = mod;
         }
         if ((ed = dlg.eds.get(id)) !== undefined && ed.m !== undefined)
         {
            // Check number of mouse clicks.
            if (pn.cnt === 0) {
               // First click, wait for second click.
               pn.cnt = 1;
               var app = self.constructor;
               self.commitPendingClickTimer =
                  window.setTimeout(app.commitPendingClick,
                                    app.doubleClickTimeout);
               return;
            }
         }
         // 'click' committed, submit event.
         self.submit(el, pn);
      }
      else if ((pn = pn.ev) !== undefined) {
         // 'mouseup' target different from pending 'click' target,
         // submit pending 2nd event.
         if (el = document.getElementById(pn.id))
            self.submit(el, pn);
      }

      self.pendingClick = undefined;
   }
};

eq.c.App.commitPendingClick = function() {
   var self = eq.app, pn;
   self.commitPendingClickTimer = undefined;
   if ((pn = self.pendingClick) !== undefined) {
      var id = pn.id, dlg, ed, el;
      if ((dlg = self.currDlg) && (ed = dlg.eds.get(id)) !== undefined) {
         if (ed.m === undefined)
            throw new Error(
               "BUG: commitPendingClick undefined pending click id:" + id +
               " cnt:" + pn.cnt);
         if (pn.cnt >= ed.m) {
            // 'click' committed, submit event.
            if (el = document.getElementById(id))
               self.submit(el, pn);
         }
         else if ((pn = pn.ev) !== undefined) {
            // submit pending 2nd event.
            if (el = document.getElementById(pn.id))
               self.submit(el, pn);
         }
      }
      self.pendingClick = undefined;
   }
};

eq.c.App.prototype.stopPendingClickTimer = function() {
   if (this.commitPendingClickTimer !== undefined) {
      window.clearTimeout(this.commitPendingClickTimer);
      this.commitPendingClickTimer = undefined;
   }
};

/*------------------------------------------------------------------------
   App.moveCapture()
   Begin 'mousemove' capture.
------------------------------------------------------------------------*/

eq.c.App.prototype.moveCapture = function(arg, cb, done) {
   var app = this.constructor;
   if (this.capture !== undefined)
      this.endCapture();
   this.capture = { arg : arg, cb : cb, done : done };
   window.addEventListener('mousemove', app.onCapture);
};

eq.c.App.prototype.endCapture = function() {
   var app = this.constructor, c = this.capture;
   window.removeEventListener('mousemove', app.onCapture);
   this.capture = undefined;
   if (c.done)
      return c.done(c.arg);
   return true;
};

eq.c.App.onCapture = function(e) {
   e.stopPropagation();
   var self = eq.app, app = self.constructor, c = self.capture;
   if (!c)
      window.removeEventListener('mousemove', app.onCapture);
   else if (c.cb)
      c.cb(e, c.arg);
};

/*------------------------------------------------------------------------
   static App.onGlobalContextMenu()
   Global 'contextmenu' event handler, typically invoked after
   secondary mouse button 'mousedown' event.
------------------------------------------------------------------------*/

eq.c.App.onGlobalContextMenu = function(e) {
   var t = e.target, cl, id, ic;
   while (t && t !== document) {
      if (cl = t.classList) {
         if (!id) {
            if (ic || cl.contains('eq-control'))
               id = t.id;
            else if (cl.contains('eq-container'))
               if (!(id = t.id))
                  ic = true;
         }
         if (cl.contains('eq-root') || cl.contains('eq-window')) {
            var self = eq.app;
            self.cancelUserAction();
            e.preventDefault();
            e.stopImmediatePropagation();
            if (id) {
               self.apiUpdate(id, eq.UpdateTag.mouseButton2, {
                  x : e.clientX,
                  y : e.clientY
               });
            }
            return false;
         }
      }
      t = t.parentNode;
   }
   return true;
};

/*------------------------------------------------------------------------
   static App.onGlobalPageHide()
   Global 'pagehide' event handler, fired when leaving current page.
------------------------------------------------------------------------*/

eq.c.App.onGlobalPageHide = function(e) {
   var self = eq.app;
   if (self.stopped !== undefined) {
      // Application stopped because user closed window or moved away.
      // Restart session later if possible, see eq.c.App.onStopped().
      window.clearTimeout(self.stopped);
      self.stopped = undefined;
   }
   self.updateSessionList(true);
};

/*------------------------------------------------------------------------
   static App.onStorageModified()
   Global 'storage' event handler, storage modified in other document.
   Updates session list from localStorage if necessary.
------------------------------------------------------------------------*/

eq.c.App.onStorageModified = function(e) {
   if (e.storageArea === window.localStorage) {
      switch (e.key) {
         case 'eq-webdlg.sls': {
            // Session list.
            var self = eq.app;
            if (!e.newValue) {
               // Session list deleted.
               self.sls = undefined;
            }
            else {
               // Session list modified.
               try {
                  var
                     nsls = JSON.parse(e.newValue),
                     csls = self.sls;
                  if (!csls || nsls.ts >= csls.ts) {
                     // No current session list,
                     // or modified list is newer, take modified list.
                     self.sls = nsls;
                  }
               }
               catch (e) {
                  console.log("Failed to parse session list, " + e.message);
                  self.sls = undefined;
                  window.localStorage.removeItem('eq-webdlg.sls');
               }
            }
            self.updateSessionList();
            break;
         }

         case 'eq-webdlg.act': {
            // Activate session.
            var sid;
            if (   e.newValue
                && (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
                && e.newValue === sid) {
               window.focus();
               window.localStorage.removeItem('eq-webdlg.act');
            }
            break;
         }
      }
   }
};

/*------------------------------------------------------------------------
   static App.activateSession()
   Activate other session.
------------------------------------------------------------------------*/

eq.c.App.activateSession = function(s) {
   var sid, w;
   if (   s
       && (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
       && s !== sid
       && (w = window.open('', 'eq-webdlg.' + s))) {
      // Try to activate session window.
      w.focus();
      window.localStorage.setItem('eq-webdlg.act', s);
   }
};

/*------------------------------------------------------------------------
   App.sessionTitles()
   Obtain titles of all other sessions.
------------------------------------------------------------------------*/

eq.c.App.prototype.sessionTitles = function() {
   var t = [], sls = this.sls, uat, opn, sn, s;
   if (   sls !== undefined
       && (uat = window.sessionStorage.getItem('eq-webdlg.uat'))
       && sls.ua == uat
       && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))) {
      for (var i = 0, l = (sn = sls.sn).length; i < l; i++)
         if (opn === (s = sn[i]).o && s.t)
            t.push(s.t);
   }
   return t;
};

/*------------------------------------------------------------------------
   App.sessionByIndex()
   Obtain other session by index.
------------------------------------------------------------------------*/

eq.c.App.prototype.sessionByIndex = function(si) {
   var sls = this.sls, uat, opn, sn, s;
   if (   sls !== undefined
       && (uat = window.sessionStorage.getItem('eq-webdlg.uat'))
       && sls.ua == uat
       && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))) {
      for (var i = 0, l = (sn = sls.sn).length; i < l; i++)
         if (opn === (s = sn[i]).o && s.t && si-- === 0)
            return s.s;
   }
   return null;
};

/*------------------------------------------------------------------------
   App.sessionByTitle()
   Obtain other session by title.
------------------------------------------------------------------------*/

eq.c.App.prototype.sessionByTitle = function(t) {
   var sls = this.sls, uat, sid, opn, sn, s;
   if (   sls !== undefined
       && (uat = window.sessionStorage.getItem('eq-webdlg.uat'))
       && sls.ua == uat
       && (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
       && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))) {
      for (var i = 0, l = (sn = sls.sn).length; i < l; i++)
         if (opn === (s = sn[i]).o && sid !== s.s && t === s.t)
            return s.s;
   }
   return null;
};

/*------------------------------------------------------------------------
   App.updateSessionList()
   Update session list.
------------------------------------------------------------------------*/

eq.c.App.prototype.updateSessionList = function(closing) {
   var uat, sid, opn, sls, sn;
   if (   !closing
       && (uat = window.sessionStorage.getItem('eq-webdlg.uat'))
       && (sid = window.sessionStorage.getItem('eq-webdlg.sid'))
       && (opn = window.sessionStorage.getItem('eq-webdlg.opn'))) {
      if ((sls = this.sls) === undefined) {
         try {
            if (sls = JSON.parse(window.localStorage.getItem('eq-webdlg.sls')))
               this.sls = sls;
         } catch (e) {}
      }
      if (!sls || sls.ua !== uat) {
         // No session list or different user agent tag.
         this.sls = sls = { ua: uat, sn : [] };
      }
      var i, l = (sn = sls.sn).length, t = document.title, s, modified;
      for (i = 0; i < l; i++) {
         if (sid === (s = sn[i]).s) {
            if (s.o !== opn || s.t !== t) {
               s.o = opn;
               s.t = t;
               modified = true;
            }
            break;
         }
      }
      if (i === l) {
         sn.push({
            s : sid,
            o : opn,
            t : t
         });
         modified = true;
      }
      if (modified) {
         sls.ts = Date.now();
         window.localStorage.setItem('eq-webdlg.sls', JSON.stringify(sls));
      }
   }
   else {
      this.sls = undefined;
      if (   (uat = window.sessionStorage.getItem('eq-webdlg.uat'))
          && (sid = window.sessionStorage.getItem('eq-webdlg.sid'))) {
         try {
            sls = JSON.parse(window.localStorage.getItem('eq-webdlg.sls'));
            if (sls.ua === uat)
               for (var i = 0, l = (sn = sls.sn).length; i < l; i++)
                  if (sid === sn[i].s) {
                     sn.splice(i, 1);
                     sls.ts = Date.now();
                     window.localStorage.setItem(
                        'eq-webdlg.sls', JSON.stringify(sls));
                     break;
                  }
         } catch (e) {}
      }
   }
};

/*------------------------------------------------------------------------
   static App.addRootListeners() .removeRootListeners()
   Add/remove root element event listeners.
------------------------------------------------------------------------*/

eq.c.App.rootEventTypes = [
   'change',
   'focusin',
   'input',
   'mousedown'
];

eq.c.App.addRootListeners = function(el) {
   var ln = this.onEvent, types = this.rootEventTypes;
   for (var i = 0, l = types.length; i < l; i++)
      el.addEventListener(types[i], ln, true);
};

eq.c.App.removeRootListeners = function(el) {
   var ln = this.onEvent, types = this.rootEventTypes;
   for (var i = 0, l = types.length; i < l; i++)
      el.removeEventListener(types[i], ln, true);
};

/*------------------------------------------------------------------------
   App.setFocus()
   Set focus by element id.
------------------------------------------------------------------------*/

eq.c.App.prototype.setFocus = function(id, el) {
   // Focus changed?
   if (id !== this.focus) {
      var
         app = this.constructor,
         dlg = this.currDlg,
         n, f, ed;
      if (!dlg)
         throw new Error("BUG: setFocus no current dialog");

      // Remove current focus indicator.
      n = document.body.getElementsByClassName('eq-focused');
      for (var i = 0, l = n.length; i < l; i++)
         n[i].classList.remove('eq-focused');

      if ((f = this.focus) !== null) {
         // Focus change, check 'blur' event on current focus element.
         if ((ed = dlg.eds.get(f)) === undefined)
            throw new Error("setFocus: from '" + f + "' not registered");
         var e, from = document.getElementById(f);
         if (!from)
            throw new Error("setFocus: from '" + f + "' does not exist");
         ed.focusLost(dlg, from);
         if (this.processEvent(dlg, {
            type : 'blur',
            target : ed.focusElement(dlg, from)
         }, f, from, ed)) {
            // Action event submitted, done.
            this.focus = id;
            return true;
         }
      }

      // Set new focus.
      if ((ed = dlg.eds.get(id)) === undefined)
         throw new Error("setFocus: to '" + id + "' not registered");
      if (el === undefined && !(el = document.getElementById(id)))
         throw new Error("setFocus: to '" + id + "' does not exist");
      el.classList.add('eq-focused');
      ed.focusGained(dlg, el = ed.focusElement(dlg, el));
      this.focus = id;
      // Set element focus on next event cycle.
      this.inSetFocus = true;
      eq.Dom.onNextCycle(app.setElementFocus, el);
   }
   return false;
};

eq.c.App.setElementFocus = function(el) {
   el.focus();
   // Verify element focus on next event cycle.
   eq.Dom.onNextCycle(eq.c.App.verifyElementFocus);
};

eq.c.App.verifyElementFocus = function() {
   var
      self = eq.app,
      f = self.focus,
      typeAhead = self.typeAhead,
      el;
   if (f && (!(el = eq.Dom.eqControl(document.activeElement)) || el.id !== f))
      // Focus could not be set, set focus to body element.
      if ((el = document.activeElement) && el !== document.body)
         el.blur();

   // Element focus verified.
   self.inSetFocus = undefined;

   // Process type-ahead key strokes if necessary.
   if (typeAhead !== undefined && typeAhead.active) {
      var app = self.constructor, d = self.currDlg, ed, key;
      if (d !== null && d === typeAhead.dlg && f === typeAhead.focus) {
         el = document.getElementById(f);
         for (var i = 0, k = typeAhead.buf, l = k.length; i < l;) {
            if (!(ed = d.eds.get(f)))
               break;
            if ((key = k[i++]) === 'Enter' || key === 'ShiftEnter')
               if (d.fkey && d.fkey.length > 0 && d.fkey[0]) {
                  // Invalid: Enter key, Dialog.cr function key is configured.
                  break;
               }
            if (!ed.onTypeAhead(d, f, el, key))
               break;
            if (self.focus !== f) {
               // Focus has changed,
               // continue when new focus has been verified.
               if ((typeAhead.focus = self.focus) !== null) {
                  k.splice(0, i);
                  return;
               }
               break;
            }
         }
      }
      app.cancelTypeAhead();
   }
};

/*------------------------------------------------------------------------
   App.setIdle()
   Set application idle or active.
------------------------------------------------------------------------*/

eq.c.App.prototype.setIdle = function(dlg) {
   var app = this.constructor;
   if (dlg === undefined) {
      // Delay setting active state until next event cycle.
      eq.Dom.onNextCycle(app.setActive)
   }
   else {
      // Idle.
      this.idle = true;
      // Add idle pane(s) if necessary.
      var pane;
      for (var i = 0, l = dlg.root.length; i < l; i++) {
         var el = dlg.root[i];
         if (!el.querySelector('.eq-root>.eq-idle')) {
            pane = document.createElement('DIV');
            pane.className = 'eq-idle';
            pane.style.zIndex = eq.Dom.zIndex(el) + 2;
            el.appendChild(pane);
         }
         // After 2 seconds, visualize idle state on newly added panes.
         if (pane) {
            if (this.idleTimer !== undefined)
               window.clearTimeout(this.idleTimer);
            this.idleTimer = window.setTimeout(app.onIdleTimer, 2000);
         }
      }
   }
};

// Set active state.
eq.c.App.setActive = function() {
   var
      self = eq.app,
      dlg = self.currDlg,
      el;
   if (!dlg)
      throw new Error("BUG: setActive no current dialog");
   self.idle = false;
   // Remove idle pane(s) if any.
   for (var i = 0, l = dlg.root.length; i < l; i++)
      if (el = dlg.root[i].querySelector('.eq-root>.eq-idle'))
         el.parentNode.removeChild(el);
};

// Visualze idle state.
eq.c.App.onIdleTimer = function() {
   // Timer has elapsed.
   eq.app.idleTimer = undefined;
   // Add 'eq-fog' class to idle pane(s).
   var n = document.body.getElementsByClassName('eq-idle');
   for (var i = 0, l = n.length; i < l; i++)
      n[i].classList.add('eq-fog');
};

/*------------------------------------------------------------------------
   App.makeOverlay() .closeOverlay()
   Overlay management.
------------------------------------------------------------------------*/

eq.c.App.prototype.makeOverlay = function(el, overlay, onClose) {
   if (this.overlay !== undefined)
      throw new Error("makeOverlay BUG: pending overlay id:" +
         this.overlay.el.id);
   if (document.body.querySelector('body>.eq-overlay'))
      throw new Error("makeOverlay BUG: orphaned overlay");
   this.overlay = {
      el : el,
      overlay : overlay,
      onClose : onClose
   };
   var b = document.body, n = b.children, z, maxZ = 0;
   for (var i = n.length; --i >= 0;)
      if ((z = eq.Dom.zIndex(n[i])) > maxZ)
         maxZ = z;
   overlay.classList.add('eq-overlay');
   overlay.style.zIndex = maxZ + 1;
   b.appendChild(overlay);
};

eq.c.App.prototype.closeOverlay = function(el, arg) {
   var od = this.overlay;
   if (od === undefined) {
      if (document.body.querySelector('body>.eq-overlay'))
         throw new Error("closeOverlay BUG: orphaned overlay");
      return;
   }
   var ov = od.overlay;
   if (el !== undefined && el !== od.el)
      throw new Error("closeOverlay BUG: wrong id:" + el.id +
         ", expected:" + od.el.id);
   if (!ov.classList.contains('eq-overlay'))
      throw new Error("closeOverlay BUG: no overlay id:" + od.el.id);
   this.overlay = undefined;
   od.onClose(od.el, ov, arg);
   ov.style.zIndex = '';
   ov.classList.remove('eq-overlay');
};

/*------------------------------------------------------------------------
   App.cleanupStyles() .removeStyles()
   Cleanup application styles, remove styles by id.
------------------------------------------------------------------------*/

eq.c.App.prototype.cleanupStyles = function() {
   var
      stl = this.stl,
      sheet = stl.sheet,
      rules = sheet.cssRules,
      l;
   if (l = rules.length) {
      if (l !== stl.owner.length)
         throw new Error("cleanupStyles BUG: " + l +
            " stale rules, inconsistent owner.length:" + stl.owner.length);
      else
         console.warn("cleanupStyles: " + l + " stale rules");
      do
         sheet.deleteRule(l - 1);
      while (l = rules.length);
      stl.owner.length = 0;
   }
   else if (stl.owner.length)
      throw new Error("cleanupStyles BUG: inconsistent owner.length:" +
         stl.owner.length);
};

eq.c.App.prototype.removeStyles = function(id) {
   var
      stl = this.stl,
      sheet = stl.sheet,
      owner = stl.owner,
      rl = owner.length,
      i = 0, ri, ci;
   if ((ri = ci = owner.indexOf(id)) !== -1) {
      do {
         sheet.deleteRule(ci);
         i++;
      } while (++ri < rl && owner[ri] === id);
      owner.splice(ci, i);
   }
};

/*------------------------------------------------------------------------
   App.newDialog()
   Create new dialog.
------------------------------------------------------------------------*/

eq.c.App.prototype.newDialog = function(id) {
   var m = eq.LayoutMode, lm = this.layoutMode, dlg;
   if (lm !== undefined && (lm & m.inline) && !(lm & m.dialog)) {
      // Single inline Dialog, forced by layout mode,
      // invalidate and close existing dialogs.
      this.dlg.forEach(eq.c.App.cbInvalidateCloseDlg, this);
      this.dlg.clear();
   }

   dlg = new eq.c.Dlg(id);
   this.dlg.set(id, dlg);
   return dlg;
};

eq.c.App.cbInvalidateCloseDlg = function(d) {
   this.api.postMessage({
      o : eq.ApiOp.invalidate,
      d : d.id
   });
   d.close();
};

/*------------------------------------------------------------------------
   App.closeDialog()
   Close dialog.
------------------------------------------------------------------------*/

eq.c.App.prototype.closeDialog = function(dlg) {
   this.dlg.delete(dlg.id);
   dlg.close();
};

/*------------------------------------------------------------------------
   App.topDialog()
   Get topmost dialog.
------------------------------------------------------------------------*/

eq.c.App.prototype.topDialog = function() {
   var wl, i, d;
   if ((wl = this.wl) !== undefined)
      for (i = wl.length; --i >= 0;)
         if ((d = wl[i].dlg) !== undefined)
            return d;

   return null;
};

/*------------------------------------------------------------------------
   App.setCurrDialog()
   Set current dialog, make it top and visible.
------------------------------------------------------------------------*/

eq.c.App.prototype.setCurrDialog = function(dlg) {
   if (dlg.w !== null)
      dlg.w.activate();

   dlg.visible();
   this.currDlg = dlg;

   // Set Dialog title.
   var
      app = this.constructor,
      t = dlg.title || app.title;
   if (t !== document.title) {
      document.title = t;
      this.updateSessionList();
   }
};

/*------------------------------------------------------------------------
   App.dialogsVisible()
   Make dialogs visible.
------------------------------------------------------------------------*/

eq.c.App.prototype.dialogsVisible = function() {
   var app = this.constructor;
   this.dlg.forEach(app.cbDialogsVisible);
};

eq.c.App.cbDialogsVisible = function(d) {
   d.visible();
};

/*------------------------------------------------------------------------
   Type-ahead management.
------------------------------------------------------------------------*/

eq.c.App.prototype.startTypeAhead = function() {
   var typeAhead = this.typeAhead, f, d;
   if (typeAhead !== undefined && !typeAhead.active) {
      if ((f = this.focus) !== null) {
         if ((d = this.currDlg) === null)
            throw new Error("App.startTypeAhead BUG: no current dialog");
         typeAhead.active = true;
         typeAhead.dlg = d;
         typeAhead.focus = f;
      }
   }
};

eq.c.App.prototype.addTypeAhead = function(e) {
   var app = this.constructor, typeAhead = this.typeAhead, key;
   if (typeAhead !== undefined && typeAhead.active) {
      switch (key = e.key) {
         case 'Tab':
         case 'Enter':
            if (e.shiftKey)
               key = 'Shift' + key;
            break;

         case ' ':
         case 'Spacebar': // IE, Firefox <= 37
            key = ' ';
            break;

         case 'Shift':
         case 'CapsLock':
            // Allowed modifier keys, ignore.
            return false;

         default:
            if (key.length !== 1 || e.ctrlKey || e.altKey || e.metaKey) {
               // Invalid key.
               app.cancelTypeAhead();
               return false;
            }
      }

      typeAhead.buf.push(key);
      if (typeAhead.timer !== 0)
         window.clearTimeout(typeAhead.timer);
      typeAhead.timer = window.setTimeout(app.cancelTypeAhead, 10000);
      return true;
   }

   return false;
};

eq.c.App.cancelTypeAhead = function() {
   var typeAhead = eq.app.typeAhead;
   if (typeAhead !== undefined) {
      typeAhead.active = false;
      typeAhead.dlg = null;
      typeAhead.focus = null;
      typeAhead.buf.length = 0;
      if (typeAhead.timer !== 0) {
         window.clearTimeout(typeAhead.timer);
         typeAhead.timer = 0;
      }
   }
};

/*========================================================================
   Start main thread
========================================================================*/

eq.app = new eq.c.App();
if (!eq.redirected) {
   // Install global event listeners.
   window.addEventListener('keydown', eq.c.App.onGlobalKeyDown, true);
   window.addEventListener('keyup', eq.c.App.onGlobalKeyUp, true);
   window.addEventListener('click', eq.c.App.onGlobalMouseDownOrClick, true);
   window.addEventListener('mousedown', eq.c.App.onGlobalMouseDownOrClick, true);
   window.addEventListener('mouseup', eq.c.App.onGlobalMouseUp, true);
   window.addEventListener('contextmenu', eq.c.App.onGlobalContextMenu, true);
   window.addEventListener('pagehide', eq.c.App.onGlobalPageHide, true);
   window.addEventListener('storage', eq.c.App.onStorageModified);
   window.addEventListener('resize', eq.c.App.onCancelUserAction, true);
   window.addEventListener('focus', eq.c.App.onCancelUserAction, true);
   window.addEventListener('blur', eq.c.App.onCancelUserAction, true);

   // Start API thread.
   eq.c.App.apiStart();

   // Continue initialization when DOM is ready.
   if (document.readyState === 'loading')
      window.addEventListener('DOMContentLoaded', eq.c.App.domReady);
   else
      eq.c.App.domReady();
}

